/*
   md.c : Multiple Devices driver for Linux
	  Copyright (C) 1998, 1999, 2000 Ingo Molnar

     completely rewritten, based on the MD driver code from Marc Zyngier

   Changes:

   - RAID-1/RAID-5 extensions by Miguel de Icaza, Gadi Oxman, Ingo Molnar
   - RAID-6 extensions by H. Peter Anvin <hpa@zytor.com>
   - boot support for linear and striped mode by Harald Hoyer <HarryH@Royal.Net>
   - kerneld support by Boris Tobotras <boris@xtalk.msk.su>
   - kmod support by: Cyrus Durgin
   - RAID0 bugfixes: Mark Anthony Lisher <markal@iname.com>
   - Devfs support by Richard Gooch <rgooch@atnf.csiro.au>

   - lots of fixes and improvements to the RAID1/RAID5 and generic
     RAID code (such as request based resynchronization):

     Neil Brown <neilb@cse.unsw.edu.au>.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   You should have received a copy of the GNU General Public License
   (for example /usr/src/linux/COPYING); if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/module.h>
#include <linux/config.h>
#include <linux/linkage.h>
#include <linux/raid/md.h>
#include <linux/sysctl.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/buffer_head.h> /* for invalidate_bdev */
#include <linux/suspend.h>
#include <linux/fs.h>

#include <linux/init.h>

#ifdef CONFIG_KMOD
#include <linux/kmod.h>
#endif

#include <asm/unaligned.h>

static int mdp_major = 0;
#define MAJOR_NR MD_MAJOR

/* 63 partitions with the alternate major number (mdp) */
#define MDP_MINOR_SHIFT 6
#define mdpminor(mddev) ((mddev)->unit << MDP_MINOR_SHIFT)

#define DEBUG 0
#define dprintk(x...) ((void)(DEBUG && printk(x)))

static int md_open(struct inode *, struct file *);
static int md_release(struct inode *inode, struct file *file);
static int md_ioctl(struct inode *, struct file *, u_int, unsigned long arg);
static int md_media_changed(struct gendisk *);
static int md_revalidate(struct gendisk *);

static struct block_device_operations md_fops =
{
	.owner		= THIS_MODULE,
	.open		= md_open,
	.release	= md_release,
	.ioctl		= md_ioctl,
	.media_changed	= md_media_changed,
	.revalidate_disk= md_revalidate,
};

static void autostart_arrays (void);
static mdk_personality_t *pers[MAX_PERSONALITY];
static spinlock_t pers_lock = SPIN_LOCK_UNLOCKED;

struct block_device *md_get_blkdev(dev_t dev)
{
	struct block_device *bdev;
	int error;

	bdev = bdget(dev);
	if (bdev == NULL)
		return (NULL);
	if ((error = blkdev_get(bdev, FMODE_READ, 0)) < 0) {
		printk("blkdev_get failed with errno %d\n", -error);
		return (NULL);
	}

	return (bdev);
}

void md_release_blkdev(struct block_device *blkdev)
{
	blkdev_put(blkdev);
}

int md_read_blocks(struct block_device *bdev, sector_t start_sector,
		   int num_sectors, uint8_t *dest_buf)
{
	Sector sect;
	u_char *data;
	int i;

	for (i = 0; i < num_sectors; i++) {
		data = read_dev_sector(bdev, start_sector + i, &sect);
		if (data == NULL) {
			printk("md_read_blocks: read_dev_sector failed\n");
			return (-EIO);
		}
		memcpy(dest_buf + (i * 512), data, 512);
		put_dev_sector(sect);
	}
	return (0);
}

static int md_write_blocks_complete(struct bio *bio, u_int bytes_done,
				    int error)
{
	if (bio->bi_size)
		return 1;

	complete((struct completion*)bio->bi_private);
	return 0;
}

int md_write_blocks(struct block_device *blkdev, sector_t sector_offset,
		    int num_sectors, uint8_t *src_buf)
{
	struct bio bio;
	struct bio_vec vec;
	struct completion event;
	uint8_t *vaddr;
	int cur_size;
	int cur_sectors;
	int error;

	bio_init(&bio);
	bio.bi_io_vec = &vec;
	vec.bv_page = alloc_page(GFP_KERNEL);
	if (vec.bv_page == NULL)
		return -ENOMEM;
	vaddr = page_address(vec.bv_page);
	vec.bv_offset = 0;
	bio.bi_vcnt = 1;
	bio.bi_idx = 0;
	bio.bi_bdev = blkdev;
	init_completion(&event);
	bio.bi_private = &event;
	bio.bi_end_io = md_write_blocks_complete;
	error = 0;

	while (num_sectors > 0) {
		
		if ((num_sectors << 9) > PAGE_SIZE)
			cur_size = PAGE_SIZE;
		else
			cur_size = num_sectors << 9;
		cur_sectors = cur_size >> 9;
		memcpy(vaddr, src_buf, cur_size);
		vec.bv_len = cur_size;
		bio.bi_size = cur_size;
		bio.bi_sector = sector_offset;
		submit_bio(WRITE, &bio);
		/* XXX Really necessary?? */
		blk_run_queues();
		wait_for_completion(&event);
		if (!test_bit(BIO_UPTODATE, &bio.bi_flags)) {
			error = -EIO;
			break;
		}
		num_sectors -= cur_sectors;
		sector_offset += cur_sectors;
		src_buf += cur_size;
	}
	__free_page(vec.bv_page);
	return (error);
}

/*
 * Current RAID-1,4,5 parallel reconstruction 'guaranteed speed limit'
 * is 1000 KB/sec, so the extra system load does not show up that much.
 * Increase it if you want to have more _guaranteed_ speed. Note that
 * the RAID driver will use the maximum available bandwith if the IO
 * subsystem is idle. There is also an 'absolute maximum' reconstruction
 * speed limit - in case reconstruction slows down your system despite
 * idle IO detection.
 *
 * you can change it via /proc/sys/dev/raid/speed_limit_min and _max.
 */

static int sysctl_speed_limit_min = 1000;
static int sysctl_speed_limit_max = 200000;

void md_member_add_child(mdk_member_t *member, mdk_member_t *child)
{
	struct list_head *list_item;
	u_long flags;

	child->parent = member;
	member_get(member);
	spin_lock_irqsave(&member->member_lock, flags);
	list_item = member->child_set.next;
	while (list_item != &member->child_set) {
		mdk_member_t *list_member;

		list_member = list_entry(list_item, mdk_member_t, peer_links);
		if (list_member->priority > child->priority)
			break;
		list_item = list_item->next;
	}
	if (list_item != &member->child_set)
		list_add_tail(&child->peer_links, list_item);
	else
		list_add_tail(&child->peer_links, &member->child_set);
	member->child_set_gencount++;
	spin_unlock_irqrestore(&member->member_lock, flags);
}

void md_member_orphan(mdk_member_t *member)
{
	u_long flags;

	if (member->parent != NULL) {
		spin_lock_irqsave(&member->parent->member_lock, flags);
		list_del(&member->peer_links);
		member->parent->child_set_gencount++;
		spin_unlock_irqrestore(&member->parent->member_lock, flags);
		member_put(member->parent);
	}
}

static void member_init(mdk_member_t *member, mdk_member_t *parent,
			mdk_mtype_t type, mdk_msubtype_t subtype,
			int priority, void (*destructor)(mdk_member_t *))
{
	spin_lock_init(&member->member_lock);
	member->parent = parent;
	member->type = type;
	member->subtype = subtype;
	INIT_LIST_HEAD(&member->child_set);
	INIT_LIST_HEAD(&member->peer_links);
	member->priority = priority;
	member->destructor = destructor;
	atomic_set(&member->refcount, 1);

	member->state = MDK_MSTATE_UNKNOWN;

	/*
	 * Sort this new member into the
	 * child list of the parent.
	 */
	if (parent != NULL)
		md_member_add_child(parent, member);
}

static void member_destructor(mdk_member_t *member)
{
	md_member_orphan(member);
}

enum md_bcast_e
{
	MD_BCAST_PERS,
	MD_BCAST_METAPERS
};
typedef enum md_bcast_e md_bcast_t;

/*
 * Perform a depth first traversal of the member
 * tree rooted in "member" calling the member personality
 * specified by "who" with the opcode "op" and the argument
 * "arg".
 */
static void md_member_bcast(mdk_member_t *member, int op, void *arg,
			    md_bcast_t who)
{
	struct list_head *tmp;
	mdk_member_t *child;
	u_long flags;

	ITERATE_CHILDREN(member, child, tmp, flags) {
		md_member_bcast(child, op, arg, who);
	}
	switch (who) {
	case MD_BCAST_PERS:
		if (member->pers != NULL)
			member->pers->raid_op(member, op, arg);
		break;
	case MD_BCAST_METAPERS:
		if (member->metapers != NULL)
			member->metapers->member_op(member, op, arg);
		break;
	default:
		break;
	}
}

static struct ctl_table_header *raid_table_header;
static LIST_HEAD(md_metapersonality_list);
static spinlock_t metapersonality_list_lock = SPIN_LOCK_UNLOCKED;
static int metapers_change_count;

/*
 * Iterates through all registered metadata personalities.
 * We take care to grab the personality list lock whenever
 * navigating the list, and to always hold a refcount when
 * unlocked.  Any code which breaks out of this loop will
 * own a reference to the personality pointed to by metapers
 * that must evenutally be released.
 */
#define ITERATE_METAPERS(metapers,tmp)					    \
									    \
for (({ spin_lock(&metapersonality_list_lock);				    \
	tmp = md_metapersonality_list.next;				    \
	metapers = NULL;});						    \
     ({ if (tmp != &md_metapersonality_list)				    \
		metapers_get(list_entry(tmp, mdk_metapersonality_t, links));\
	spin_unlock(&metapersonality_list_lock);			    \
	if (metapers) metapers_put(metapers);				    \
	metapers = list_entry(tmp, mdk_metapersonality_t, links);	    \
	tmp != &md_metapersonality_list;});				    \
     ({ spin_lock(&metapersonality_list_lock);				    \
	tmp = tmp->next;})						    \
	)

static inline mdk_metapersonality_t *
metapers_get(mdk_metapersonality_t *metapers)
{
	atomic_inc(&metapers->refcount);
	return metapers;
}

static void metapers_put(mdk_metapersonality_t *metapers)
{
	if (!atomic_dec_and_lock(&metapers->refcount,
				 &metapersonality_list_lock)) 
		return;
	list_del(&metapers->links);
	/* XXX Handle metapers release... Decrement module count? */
	spin_unlock(&metapersonality_list_lock);
}

static ctl_table raid_table[] = {
	{
		.ctl_name	= DEV_RAID_SPEED_LIMIT_MIN,
		.procname	= "speed_limit_min",
		.data		= &sysctl_speed_limit_min,
		.maxlen		= sizeof(int),
		.mode		= 0644,
		.proc_handler	= &proc_dointvec,
	},
	{
		.ctl_name	= DEV_RAID_SPEED_LIMIT_MAX,
		.procname	= "speed_limit_max",
		.data		= &sysctl_speed_limit_max,
		.maxlen		= sizeof(int),
		.mode		= 0644,
		.proc_handler	= &proc_dointvec,
	},
	{ .ctl_name = 0 }
};

static ctl_table raid_dir_table[] = {
	{
		.ctl_name	= DEV_RAID,
		.procname	= "raid",
		.maxlen		= 0,
		.mode		= 0555,
		.child		= raid_table,
	},
	{ .ctl_name = 0 }
};

static ctl_table raid_root_table[] = {
	{
		.ctl_name	= CTL_DEV,
		.procname	= "dev",
		.maxlen		= 0,
		.mode		= 0555,
		.child		= raid_dir_table,
	},
	{ .ctl_name = 0 }
};

/*
 * All top-level arrays are children of the
 * "toplevel_arrays".  This simplifies code since
 * all "members" have a parent and we can use
 * the same "member" rules and semantics to
 * manage not only the top level array list,
 * but array children too.
 */
mdk_member_t toplevel_arrays = {
	.type		= MDK_MTYPE_MASTER,
	.subtype	= MDK_MSUBTYPE_NONE,
	.member_lock	= SPIN_LOCK_UNLOCKED,
	.child_set	= LIST_HEAD_INIT(toplevel_arrays.child_set),
	.refcount	= ATOMIC_INIT(1)
};

static int md_fail_request (request_queue_t *q, struct bio *bio)
{
	bio_io_error(bio, bio->bi_size);
	return 0;
}

static void mddev_destructor(mdk_member_t *member)
{
	mddev_t *mddev;

	member_destructor(member);
	mddev = to_mddev(member);
	blk_put_queue(mddev->queue);
	kfree(mddev);
}

static mddev_t *mddev_find(int unit)
{
	struct list_head *tmp;
	mdk_member_t *child;
	mddev_t *mddev;
	u_long flags;

	ITERATE_CHILDREN(&toplevel_arrays, child, tmp, flags) {
		mddev = to_mddev(child);
		if (mddev->unit == unit)
			return (mddev);
	}

	return NULL;
}

/* 
 * md_get_unit returns the lowest free index of mddev_map array, if
 * the requested_unit is -1. For requested_unit is >= 0, it returns
 * the same requested_unit, if the index is free else returns any 
 * other free index. 
 */
static int md_get_unit(int requested_unit)
{
	static int unit;
	
/* XXX Need to scan top-level arrays for allocated unit numbers. */
	return (unit++);
}

mddev_t *md_mddev_alloc(mdk_member_t *parent, mdk_msubtype_t mddev_subtype,
			int priority)
{
	mddev_t *mddev;

	mddev = (mddev_t *) kmalloc(sizeof(*mddev), GFP_KERNEL);
	if (!mddev)
		return NULL;

	memset(mddev, 0, sizeof(*mddev));

	init_MUTEX(&mddev->reconfig_sem);
	init_timer(&mddev->safemode_timer);

	mddev->queue = blk_alloc_queue(GFP_KERNEL);
	if (!mddev->queue) {
		kfree(mddev);
		return NULL;
	}

	blk_queue_make_request(mddev->queue, md_fail_request);

	member_init(&mddev->member, parent, MDK_MTYPE_MDDEV,
		    mddev_subtype, priority, mddev_destructor);

	return mddev;
}

static inline int mddev_lock(mddev_t * mddev)
{
	return down_interruptible(&mddev->reconfig_sem);
}

static inline void mddev_lock_uninterruptible(mddev_t * mddev)
{
	down(&mddev->reconfig_sem);
}

static inline int mddev_trylock(mddev_t * mddev)
{
	return down_trylock(&mddev->reconfig_sem);
}

static inline void mddev_unlock(mddev_t * mddev)
{
	up(&mddev->reconfig_sem);

	if (mddev->thread)
		md_wakeup_thread(mddev->thread);
}

/*
 * prevent the device from being mounted, repartitioned or
 * otherwise reused by a RAID array (or any other kernel
 * subsystem), by opening the device. [simply getting an
 * inode is not enough, the SCSI module usage code needs
 * an explicit open() on the device]
 */
static int lock_rdev(mdk_rdev_t *rdev)
{
	int err = 0;

	err = bd_claim(rdev->bdev, rdev);
	sync_blockdev(rdev->bdev);
	invalidate_bdev(rdev->bdev, 0);
	return err;
}

static void unlock_rdev(mdk_rdev_t *rdev)
{
	if (rdev->bdev)
		bd_release(rdev->bdev);
}

void md_autodetect_dev(dev_t dev);

static void export_array(mddev_t *mddev)
{
#if 0
	struct list_head *tmp;
	mdk_rdev_t *rdev;

	ITERATE_RDEV(mddev,rdev,tmp) {
		if (!rdev->member.parent) {
			MD_BUG();
			continue;
		}
		kick_rdev_from_array(rdev);
	}
	if (!list_empty(&mddev->member.child_set))
		MD_BUG();
	mddev->raid_disks = 0;
	mddev->major_version = 0;
#endif
}

static void print_rdev(mdk_rdev_t *rdev)
{
	mdk_member_t *member;

	member = to_member(rdev);
	printk("<%s>", mdname(member));
	printk(KERN_INFO "md: rdev %s, SZ:%08llu S:%x DN:%u\n",
		mdname(member), (unsigned long long)member->data_size,
	       	member->state, member->priority);
#if 0
	if (rdev->sb_loaded) {
		printk(KERN_INFO "md: rdev superblock:\n");
		print_sb((mdp_super_t*)page_address(rdev->sb_page));
	} else
		printk(KERN_INFO "md: no rdev superblock!\n");
#endif
}

static void md_print_member(mdk_member_t *member)
{
	struct list_head *tmp;
	mdk_member_t *child;
	u_long flags;

	if (member->type == MDK_MTYPE_RDEV) {
		/* Leaf node */
		print_rdev(to_rdev(member));
		return;
	}

	/* Print out the array and all children of the array. */
	printk("%s: ", mdname(member));
	ITERATE_CHILDREN(member, child, tmp, flags) {
		md_print_member(child);
	}
	printk("\n");
	
}

void md_print_devices(void)
{
	struct list_head *tmp;
	mdk_member_t *child;
	u_long flags;

	printk("\n");
	printk("md:	**********************************\n");
	printk("md:	* <COMPLETE RAID STATE PRINTOUT> *\n");
	printk("md:	**********************************\n");
	
	ITERATE_CHILDREN(&toplevel_arrays, child, tmp, flags) {
		md_print_member(child);
	}
	printk("md:	**********************************\n");
	printk("\n");
}

static void rdev_destructor(mdk_member_t *member)
{
	mdk_rdev_t *rdev;

	member_destructor(member);
	rdev = to_rdev(member);
	unlock_rdev(rdev);
	if (rdev->bdev)
		blkdev_put(rdev->bdev);
	kfree(rdev);
}

mdk_rdev_t *md_rdev_alloc(mdk_member_t *parent, mdk_msubtype_t subtype,
			  struct block_device *blkdev, int priority)
{
	char b[BDEVNAME_SIZE];
	mdk_rdev_t *rdev;
	sector_t size;
	int err;

	rdev = (mdk_rdev_t *) kmalloc(sizeof(*rdev), GFP_KERNEL);
	if (!rdev) {
		printk(KERN_ERR "md: could not alloc mem for %s!\n", 
		       bdevname(blkdev, b));
		return NULL;
	}
	memset(rdev, 0, sizeof(*rdev));

	rdev->member.data_offset = 0;
	blkdev = bdget(blkdev->bd_dev);
	if ((err = blkdev_get(blkdev, FMODE_READ|FMODE_WRITE, 0)) < 0) {
		printk(KERN_ERR "md: could not open %s RW.\n",
		       bdevname(blkdev, b));
		bdput(blkdev);
		goto abort_free_nolock;
	}
	rdev->bdev = blkdev;

	err = lock_rdev(rdev);
	if (err) {
		printk(KERN_ERR "md: could not lock %s.\n",
		       bdevname(blkdev, b));
		goto abort_free_nolock;
	}

	size = rdev->bdev->bd_inode->i_size >> BLOCK_SIZE_BITS;
	if (!size) {
		printk(KERN_WARNING 
			"md: %s has zero or unknown size, marking faulty!\n",
			bdevname(rdev->bdev,b));
		goto abort_free;
	}

	memcpy(rdev->member.name, bdevname(rdev->bdev, b), BDEVNAME_SIZE);
	member_init(&rdev->member, parent, MDK_MTYPE_RDEV,
		    subtype, priority, rdev_destructor);

	return rdev;

abort_free:
	unlock_rdev(rdev);
abort_free_nolock:
	if (rdev->bdev != NULL)
		blkdev_put(rdev->bdev);
	kfree(rdev);
	return NULL;
}

void md_wakeup_thread(mdk_thread_t *thread);

/*
 * Write idle timer.  Queue a sync handler workitem to
 * to do the possibly blocking work of updating the
 * meta-data to place us in safemode.
 */
static void md_safemode_timeout(unsigned long data)
{
	mdk_sync_support_t *syncinfo;

	syncinfo = (mdk_sync_support_t *)data;
	spin_lock_irq(&syncinfo->sync_lock);
	if (syncinfo->safemode_workitem.work_flags == MD_WF_NONE) {
		md_queue_workitem(&syncinfo->work_state,
				  &syncinfo->safemode_workitem,
				  /*head_of_queue*/0);
	}
	spin_unlock_irq(&syncinfo->sync_lock);
}

int md_check_chunksize(mddev_t * mddev)
{
	struct list_head *tmp;
	mdk_member_t *child;
	mdk_member_t *member;
	mdk_rdev_t *rdev;
	u_long flags;
	int pnum, chunk_size;

	chunk_size = mddev->chunk_size;
	pnum = level_to_pers(mddev->level);

	if ((pnum != MULTIPATH) && (pnum != RAID1)) {
		if (!chunk_size) {
			/*
			 * 'default chunksize' in the old md code used to
			 * be PAGE_SIZE, baaad.
			 * we abort here to be on the safe side. We don't
			 * want to continue the bad practice.
			 */
			printk(KERN_ERR 
				"no chunksize specified, see 'man raidtab'\n");
			return -EINVAL;
		}
		if (chunk_size > MAX_CHUNK_SIZE) {
			printk(KERN_ERR "too big chunk_size: %d > %d\n",
				chunk_size, MAX_CHUNK_SIZE);
			return -EINVAL;
		}
		/*
		 * chunk-size has to be a power of 2 and multiples of PAGE_SIZE
		 */
		if ( (1 << ffz(~chunk_size)) != chunk_size) {
			MD_BUG();
			return -EINVAL;
		}
		if (chunk_size < PAGE_SIZE) {
			printk(KERN_ERR "too small chunk_size: %d < %ld\n",
				chunk_size, PAGE_SIZE);
			return -EINVAL;
		}

		member = (mdk_member_t *)mddev;
		/* devices must have minimum size of one chunk */
		ITERATE_CHILDREN(member, child, tmp, flags) {
			if (child->type == MDK_MTYPE_RDEV) {
				rdev = (mdk_rdev_t *) child;
				if (rdev->member.state != MDK_MSTATE_OPTIMAL)
					continue;
				if (rdev->member.data_size < chunk_size) {
					printk(KERN_WARNING
					"md: Dev %s smaller than chunk_size:"
					" %lluk < %dk\n",
					mdname(to_member(rdev)),
					(unsigned long long)rdev->member.data_size,
					chunk_size / 1024);
					return -EINVAL;
				}
			}
		}
	}
	return 0;
}

static int runarrays_allowed;

int md_runnable(mddev_t *mddev)
{
	static DECLARE_MUTEX(disks_sem);
	struct gendisk *disk;
	mdk_personality_t *pers;
	int unit;

	pers = mddev->member.pers;

	/* Check if it is already runned once or if runarrays not allowed*/
	if ((mddev->member.name[0] != 0) || (runarrays_allowed != 1))
		return 0;

	if ((unit = md_get_unit(mddev->preferred_unit)) == -1){
		printk("md_runnable: no unit available, requested %d\n",
		       mddev->preferred_unit);
		return -1;
	}
	if ((mddev->preferred_unit != -1) && (unit != mddev->preferred_unit))
		printk("md_runnable: preferred unit %d unavailable. "
		       "Assigned %d instead\n", mddev->preferred_unit, unit);
	mddev->unit = unit;

	down(&disks_sem);
	disk = alloc_disk(1 << MDP_MINOR_SHIFT);
	if (!disk) {
		up(&disks_sem);
		return 0;
	}
	disk->major = mdp_major;
	disk->first_minor = unit << MDP_MINOR_SHIFT;
	sprintf(disk->disk_name, "md%c", 'a' + unit);
	sprintf(to_member(mddev)->name, "md%c", 'a' + unit);
	disk->fops = &md_fops;
	disk->private_data = mddev;
	disk->queue = mddev->queue;
	add_disk(disk);
	mddev->gendisk = disk;

 	atomic_set(&mddev->writes_pending,0);
	mddev->safemode = 0;
	mddev->safemode_timer.function = md_safemode_timeout;
	mddev->safemode_timer.data = (unsigned long) mddev;
	mddev->safemode_delay = (20 * HZ)/1000 +1; /* 20 msec delay */
	mddev->member.state = MDK_MSTATE_OPTIMAL;

	if (pers->raid_op(to_member(mddev),MDK_RAID_OP_RUN, 0) != 0) {
		printk("MDK_RAID_OP_RUN failed in %s\n", 
			mdname(to_member(mddev)));
/* XXX Release unit number.  Mark array as not already run. */
		up(&disks_sem);
		return -ENODEV;
	}

	/* If we call blk_queue_make_request here, it will
	 * re-initialise max_sectors etc which may have been
	 * refined inside -> run.  So just set the bits we need to set.
	 * Most initialisation happended when we called
	 * blk_queue_make_request(..., md_fail_request)
	 * earlier.
	 */
	mddev->queue->queuedata = mddev;
	mddev->queue->make_request_fn = mddev->member.pers->make_request;
	set_capacity(disk, mddev->member.data_size);
	mddev->changed = 1;
	up(&disks_sem);
	return 0;
}

static int restart_array(mddev_t *mddev)
{
	struct gendisk *disk = mddev->gendisk;
	int err;

	/*
	 * Complain if it has no devices
	 */
	err = -ENXIO;
	if (list_empty(&mddev->member.child_set))
		goto out;

	if (mddev->member.pers) {
		err = -EBUSY;
		if (!mddev->ro)
			goto out;

		mddev->safemode = 0;
		mddev->ro = 0;
		set_disk_ro(disk, 0);

		printk(KERN_INFO "%s: switched to read-write mode.\n",
		       mdname(to_member(mddev)));
		/*
		 * Kick recovery or resync if necessary
		 */
		set_bit(MD_RECOVERY_NEEDED, &mddev->recovery);
		md_wakeup_thread(mddev->thread);
		err = 0;
	} else {
		printk(KERN_ERR "%s: has no personality assigned.\n",
		       mdname(to_member(mddev)));
		err = -EINVAL;
	}

out:
	return err;
}

int do_md_ro(mddev_t *mddev, int ro)
{
	set_disk_ro(mddev->gendisk, ro);
	mddev->ro = ro;
	return 0;
}

static int do_md_stop(mddev_t * mddev, int ro)
{
	int err = 0;
	mdk_member_t *member;
	struct gendisk *disk = mddev->gendisk;

	member = to_member(mddev);
	if (mddev->member.pers) {

		invalidate_partition(disk, 0);

		if (ro) {
			err  = -ENXIO;
			if (mddev->ro)
				goto out;
			do_md_ro(mddev, 1);
			member->pers->raid_op(member, MDK_RAID_OP_STOP,
					      &mddev->ro);
		} else {
			if (mddev->ro)
				set_disk_ro(disk, 0);
			blk_queue_make_request(mddev->queue, md_fail_request);
			mddev->ro = 0;
			member->pers->raid_op(member, MDK_RAID_OP_STOP,
					      &mddev->ro);
		}
	}
	/*
	 * Free resources if final stop
	 */
	if (!ro) {
		struct gendisk *disk;

		printk(KERN_INFO "%s: stopped.\n", mdname(member));

		export_array(mddev);

		mddev->member.data_size = 0;
		disk = mddev->gendisk;
		if (disk)
			set_capacity(disk, 0);
		mddev->changed = 1;
	} else
		printk(KERN_INFO "%s: switched to read-only mode.\n",
		       mdname(member));
	err = 0;
out:
	return err;
}

static int get_version(void * arg)
{
	mdu_version_t ver;

	ver.major = MD_MAJOR_VERSION;
	ver.minor = MD_MINOR_VERSION;
	ver.patchlevel = MD_PATCHLEVEL_VERSION;

	if (copy_to_user(arg, &ver, sizeof(ver)))
		return -EFAULT;

	return 0;
}

static int get_array_info(mddev_t * mddev, void * arg)
{
	mdu_array_info_t info;
	int nr,working,active,failed,spare;
	mdk_rdev_t *rdev;
	struct list_head *tmp;
	mdk_member_t *member;
	mdk_member_t *child;
	u_long flags;

	nr=working=active=failed=spare=0;

	member = (mdk_member_t *)mddev;
	ITERATE_CHILDREN(member, child, tmp, flags) {
		if (child->type == MDK_MTYPE_RDEV) {
			nr++;
			rdev = (mdk_rdev_t *) child;
			if (rdev->member.state == MDK_MSTATE_FAILED){
				failed++;
			} else {
				working++;
				if (rdev->member.state == MDK_MSTATE_OPTIMAL)
					active++;	
				else
					spare++;
			}
		}
	}

	info.major_version = mddev->major_version;
	info.minor_version = mddev->minor_version;
	info.patch_version = 1;
	info.ctime         = mddev->ctime;
	info.level         = mddev->level;
	info.size          = mddev->member.data_size;
	info.nr_disks      = nr;
	info.raid_disks    = mddev->max_disks;
	info.not_persistent= !mddev->persistent;

	info.utime         = mddev->utime;
	info.state         = 0;
	if (mddev->member.state == MDK_MSTATE_OPTIMAL)
		info.state = (1<<MD_SB_CLEAN);
	info.active_disks  = active;
	info.working_disks = working;
	info.failed_disks  = failed;
	info.spare_disks   = spare;

	info.layout        = mddev->layout;
	info.chunk_size    = mddev->chunk_size;
	strcpy( info.metaname, mddev->member.metapers->name );
	info.array_level = mddev->level;
	info.parent_cookie = (unsigned long) mddev->member.parent;;
	info.cookie = (unsigned long) mddev;

	if (copy_to_user(arg, &info, sizeof(info)))
		return -EFAULT;

	return 0;
}

static int get_disk_info(mddev_t * mddev, void * arg)
{
	mdu_disk_info_t info;
	mdk_rdev_t *rdev;

	if ( mddev->member.type != MDK_MTYPE_RDEV ) {
		printk( KERN_WARNING "ERROR: get_disk_info() Struct is not an RDEV\n" );
		return -EFAULT;
	}

	if (copy_from_user(&info, arg, sizeof(info)))
		return -EFAULT;

	rdev = (mdk_rdev_t *) mddev;

	if (rdev) {
		info.major = MAJOR(rdev->bdev->bd_dev);
		info.minor = MINOR(rdev->bdev->bd_dev);
		info.raid_disk = 0;//rdev->raid_disk;
		info.state = 0;
		info.number = rdev->member.priority;
		if (rdev->member.state != MDK_MSTATE_OPTIMAL)
			info.state |= (1<<MD_DISK_FAULTY);
		else if (rdev->member.state == MDK_MSTATE_OPTIMAL) {
			info.state |= (1<<MD_DISK_ACTIVE);
			info.state |= (1<<MD_DISK_SYNC);
		}
	} else {
		info.major = info.minor = 0;
		info.raid_disk = -1;
		info.state = (1<<MD_DISK_REMOVED);
	}
	
	info.device_type = mddev->member.type;
	info.parent_cookie = (unsigned long) mddev->member.parent;;
	info.cookie = (unsigned long) mddev;

	if (copy_to_user(arg, &info, sizeof(info)))
		return -EFAULT;

	return 0;
}

static int hot_generate_error(mddev_t * mddev, dev_t dev)
{
	return 0;
}

static int hot_remove_disk(mddev_t * mddev, dev_t dev)
{
	return -EBUSY;
}

static int hot_add_disk(mddev_t * mddev, dev_t dev)
{
	return -ENODEV;
}

/*
 * set_array_info is used two different ways
 * The original usage is when creating a new array.
 * In this usage, raid_disks is > 0 and it together with
 *  level, size, not_persistent,layout,chunksize determine the
 *  shape of the array.
 *  This will always create an array with a type-0.90.0 superblock.
 * The newer usage is when assembling an array.
 *  In this case raid_disks will be 0, and the major_version field is
 *  use to determine which style super-blocks are to be found on the devices.
 *  The minor and patch _version numbers are also kept incase the
 *  super_block handler wishes to interpret them.
 */
static int set_array_info(mddev_t * mddev, mdu_array_info_t *info)
{

	if (info->raid_disks == 0) {
		/* just setting version number for superblock loading */
		/* XXX XXX XXX sizeof(super_types) is not known! Hack! */
#if 0
		if (info->major_version < 0 ||
		    info->major_version >= 2 /* sizeof(super_types)/sizeof(super_types[0]) */ ||
		    super_types[info->major_version].name == NULL) {
			/* maybe try to auto-load a module? */
			printk(KERN_INFO 
				"md: superblock version %d not known\n",
				info->major_version);
			return -EINVAL;
		}
#endif
		mddev->major_version = info->major_version;
		mddev->minor_version = info->minor_version;
		mddev->patch_version = info->patch_version;
		return 0;
	}
	mddev->major_version = MD_MAJOR_VERSION;
	mddev->minor_version = MD_MINOR_VERSION;
	mddev->patch_version = MD_PATCHLEVEL_VERSION;
	mddev->ctime         = get_seconds();

	mddev->level         = info->level;
	mddev->size          = info->size;
	mddev->raid_disks    = info->raid_disks;
	/* don't set md_minor, it is determined by which /dev/md* was
	 * openned
	 */
	if (info->state & (1<<MD_SB_CLEAN))
		mddev->recovery_cp = MaxSector;
	else
		mddev->recovery_cp = 0;
	mddev->persistent    = ! info->not_persistent;

	mddev->layout        = info->layout;
	mddev->chunk_size    = info->chunk_size;

	mddev->max_disks     = MD_SB_DISKS;

	mddev->sb_dirty      = 1;

	/*
	 * Generate a 128 bit UUID
	 */
	get_random_bytes(mddev->uuid, 16);

	return 0;
}

static mdk_member_t *get_member_from_member(mdk_member_t *member, mdk_member_t *cookie )
{
	struct list_head *tmp;
	mdk_member_t *child;
	mdk_member_t *grandchild;
	u_long flags;
	
	/* This function is called to verify that the member   */
	/* is valid and adds a refcount to the member if found.*/
//	printk( KERN_WARNING "get_member_from_member()\n" );

	ITERATE_CHILDREN(member, child, tmp, flags) {
		if (child == cookie)
			return (child);
					
		grandchild = get_member_from_member(child, cookie);
		if (grandchild != NULL) {
			member_put(child);
			return (grandchild);
		}
	}
	
	return (NULL);
}

static mdk_member_t *get_child_from_index(mdk_member_t *parent, int index )
{
	struct list_head *tmp;
	mdk_member_t *child;
	u_long flags;
	
	parent = get_member_from_member(&toplevel_arrays, parent);
	if (parent == NULL)
		return (NULL);
		
	member_put(parent);

	ITERATE_CHILDREN(parent, child, tmp, flags) {
		if (child->priority == index) 
			return (child);
	}
	
	return (NULL);
}

static mdk_member_t *get_member_from_device(mdk_member_t *member, dev_t device )
{
	struct list_head *tmp;
	mdk_member_t *child;
	mdk_member_t *grandchild;
	u_long flags;
	mdk_rdev_t *rdev;
	mddev_t *mddev;
	dev_t  tempdev;

	ITERATE_CHILDREN(member, child, tmp, flags) {
		if (child->type == MDK_MTYPE_MDDEV) {
			mddev = (mddev_t *) child;
			tempdev = MKDEV(mdp_major, mdpminor(mddev));
			if (tempdev == device)
				return (child);
		}
	
		if (child->type == MDK_MTYPE_RDEV) {
			rdev = (mdk_rdev_t *) child;
			if (rdev->bdev->bd_dev == device)
				return (child); 
		}
		
		grandchild = get_member_from_device(child, device);
		if (grandchild != NULL) {
			member_put(child);
			return (grandchild);
		}
	}
	
	return (NULL);
}

static int get_cookie_info(mdk_member_t * member, void *arg)
{
	mdu_cookie_info_t *cookie_info;
	mdk_rdev_t *rdev;

	cookie_info = (mdu_cookie_info_t *) arg;
	
	switch (member->type)
	{
	case MDK_MTYPE_RDEV:
		rdev = to_rdev(member);
		cookie_info->major_minor.major = MAJOR(rdev->bdev->bd_dev);
		cookie_info->major_minor.minor = MINOR(rdev->bdev->bd_dev);
			
		break;
	
	case MDK_MTYPE_MDDEV: 
		cookie_info->major_minor.major = mdp_major;
		cookie_info->major_minor.minor = mdpminor(to_mddev(member));
		break;
		
	default:
		cookie_info->major_minor.major = 0;
		cookie_info->major_minor.minor = 0;
		break;
	}	
	
   	cookie_info->parent_cookie = (unsigned long) member->parent;            
	cookie_info->cookie = (unsigned long) member;                  
	cookie_info->type = member->type;
	cookie_info->subtype = member->subtype; 
	cookie_info->gencount = member->child_set_gencount;
	
	return 0;
}

static int add_new_disk(mddev_t *mddev, mdu_disk_info_t *disk_info)
{
// TBD ... Implement this function
	
	/* 
	 * SET_ARRAY_INFO
	 * set_array_info - check if metadata type is supported,
	 * check if the minor doesn't conflict with another array,
	 * then depending upon metadata type call a metadata
	 * specific mddev allocation routine & add it to toplevel array
	 *
	 * ADD_NEW_DISK
	 * subsequent call to add_new_disk will pass a cookie that will refer 
	 * to the allocated mddev and also specify the k_devt of the device.
	 * probe through the unclaimed dev_list & remove it from the list
	 * if it already exists, else just check for its existence and
	 * stamp metadata on it, 
	 *
	 * RUN_ARRAY- TBD
	 * check if all members exist or  arrived and then run the array 
	 */
	
	 return 0;
}

static int md_ioctl(struct inode *inode, struct file *file,
			unsigned int cmd, unsigned long arg)
{
	mddev_t *mddev;
	mdk_member_t *member;
	struct hd_geometry *loc;
	u_int heads, sectors;
	sector_t cylinders;
	dev_t dev;
	int err;

	if (!capable(CAP_SYS_ADMIN))
		return -EACCES;

	if (cmd != HDIO_GETGEO || arg == 0)
		return -EINVAL;

	dev = inode->i_rdev;
	member = get_member_from_device(&toplevel_arrays, dev);
	
	if (member == NULL)
		return -ENODEV;

	mddev = to_mddev(member);
	loc = (struct hd_geometry *)arg;
	heads = 255;
	sectors = 63;
	err = put_user (heads, (char *) &loc->heads);
	if (!err)
		err = put_user (sectors, (char *) &loc->sectors);
	if (!err) {
		cylinders = get_capacity(mddev->gendisk);
		sector_div(cylinders, (heads * sectors));
		err = put_user(cylinders, (short *) &loc->cylinders);
	}
	if (!err)
		err = put_user(get_start_sect(inode->i_bdev),
			       (long *)&loc->start);
	member_put(member);
	return err;
}

static int md_open(struct inode *inode, struct file *file)
{
	/*
	 * Succeed if we can lock the mddev, which confirms that
	 * it isn't being stopped right now.
	 */
	mddev_t *mddev = inode->i_bdev->bd_disk->private_data;
	int err;

	if (mddev == NULL) {
		return -ENODEV;
	}

	if ((err = mddev_lock(mddev)))
		goto out;

	err = 0;
	mddev_unlock(mddev);

	check_disk_change(inode->i_bdev);
 out:
	return err;
}

static int md_release(struct inode *inode, struct file *file)
{
	return 0;
}

static int md_media_changed(struct gendisk *disk)
{
	mddev_t *mddev = disk->private_data;

	return mddev->changed;
}

static int md_revalidate(struct gendisk *disk)
{
	mddev_t *mddev = disk->private_data;

	mddev->changed = 0;
	return 0;
}

void md_update_thread_string(mdk_thread_t *thread, const char *fmt)
{
	if (thread->mddev != NULL) {
		sprintf(thread->tsk->comm, fmt,
			mdname(to_member(thread->mddev)));
	} else {
		sprintf(thread->tsk->comm, fmt);
	}
	thread->name = thread->tsk->comm;
}

int md_thread(void * arg)
{
	mdk_thread_t *thread = arg;

	lock_kernel();

	/*
	 * Detach thread
	 */

	if (thread->mddev != NULL)
		daemonize(thread->name, mdname(to_member(thread->mddev)));
	else
		daemonize(thread->name, NULL);

	current->exit_signal = SIGCHLD;
	allow_signal(SIGKILL);
	thread->tsk = current;

	md_update_thread_string(thread, thread->name);

	/*
	 * md_thread is a 'system-thread', it's priority should be very
	 * high. We avoid resource deadlocks individually in each
	 * raid personality. (RAID5 does preallocation) We also use RR and
	 * the very same RT priority as kswapd, thus we will never get
	 * into a priority inversion deadlock.
	 *
	 * we definitely have to have equal or higher priority than
	 * bdflush, otherwise bdflush will deadlock if there are too
	 * many dirty RAID5 blocks.
	 */
	unlock_kernel();

	complete(thread->event);
	while (thread->run) {
		mdk_thread_handler_t *run;

		wait_event_interruptible(thread->wqueue,
					 test_bit(THREAD_WAKEUP, &thread->flags));
		if (current->flags & PF_FREEZE)
			refrigerator(PF_IOTHREAD);

		clear_bit(THREAD_WAKEUP, &thread->flags);

		run = thread->run;
		if (run) {
			run(thread->mddev, thread->arg);
			blk_run_queues();
		}
		if (signal_pending(current))
			flush_signals(current);
	}
	complete(thread->event);
	return 0;
}

void md_wakeup_thread(mdk_thread_t *thread)
{
	if (thread) {
		dprintk("md: waking up MD thread %s.\n", thread->tsk->comm);
		set_bit(THREAD_WAKEUP, &thread->flags);
		wake_up(&thread->wqueue);
	}
}

mdk_thread_t *md_register_thread(mdk_thread_handler_t *run, mddev_t *mddev,
				 void *arg, const char *name)
{
	mdk_thread_t *thread;
	int ret;
	struct completion event;

	thread = (mdk_thread_t *) kmalloc
				(sizeof(mdk_thread_t), GFP_KERNEL);
	if (!thread)
		return NULL;

	memset(thread, 0, sizeof(mdk_thread_t));
	init_waitqueue_head(&thread->wqueue);

	init_completion(&event);
	thread->event = &event;
	thread->run = run;
	thread->arg = arg;
	thread->mddev = mddev;
	thread->name = name;
	ret = kernel_thread(md_thread, thread, 0);
	if (ret < 0) {
		kfree(thread);
		return NULL;
	}
	wait_for_completion(&event);
	return thread;
}

void md_interrupt_thread(mdk_thread_t *thread)
{
	if (!thread->tsk) {
		MD_BUG();
		return;
	}
	dprintk("interrupting MD-thread pid %d\n", thread->tsk->pid);
	send_sig(SIGKILL, thread->tsk, 1);
}

void md_unregister_thread(mdk_thread_t *thread)
{
	struct completion event;

	init_completion(&event);

	thread->event = &event;
	thread->run = NULL;
	thread->arg = NULL;
	thread->name = NULL;
	md_interrupt_thread(thread);
	wait_for_completion(&event);
	kfree(thread);
}

/* seq_file implementation /proc/mdstat */

static void status_unused(struct seq_file *seq)
{
	seq_printf(seq, "unused devices: ");

	seq_printf(seq, "<none>");

	seq_printf(seq, "\n");
}


static void status_resync(struct seq_file *seq, mddev_t *mddev)
{
	mdk_checkpoint_arg_t checkpoint_arg;
	mdk_syncstatsop_arg_t syncstats;
	mdk_member_t *member;
	mdk_metapersonality_t *metapers;
	mdk_personality_t *pers;
	u_long resync, max_blocks, res, rt;
	const char *noun;
	int error;

	member = to_member(mddev);
	metapers = member->metapers;
	pers = member->pers;

	error = metapers->member_op(member,
				    MDK_MEMBER_OP_GET_CHECKPOINT,
				    &checkpoint_arg);
	if (error != 0) {
		return;
	}
	error = pers->raid_op(member, MDK_RAID_OP_SYNC_STATS, &syncstats);
	if (error != 0) {
		return;
	}
	
	resync = checkpoint_arg.checkpoint>>1;
	max_blocks = mddev->size>>1;

	/*
	 * Should not happen.
	 */
	if (!max_blocks) {
		MD_BUG();
		return;
	}
	res = (resync/1024)*1000/(max_blocks/1024 + 1);
	{
		int i, x = res/50, y = 20-x;
		seq_printf(seq, "[");
		for (i = 0; i < x; i++)
			seq_printf(seq, "=");
		seq_printf(seq, ">");
		for (i = 0; i < y; i++)
			seq_printf(seq, ".");
		seq_printf(seq, "] ");
	}
	switch (syncstats.syncop) {
	case MD_SYNCOP_REBUILD:
		noun = "rebuild";
		break;
	case MD_SYNCOP_INIT:
		noun = "init";
		break;
	case MD_SYNCOP_VERIFY:
	case MD_SYNCOP_VERIFY_FIX:
		noun = "verify/fix";
		break;
	default:
		/* silence warnings */
		noun = NULL;
		MD_BUG();
		break;
	}
	
	seq_printf(seq, " %s =%3lu.%lu%% (%lu/%lu)",
		      (test_bit(MD_RECOVERY_SYNC, &mddev->recovery) ?
		       "resync" : "recovery"),
		      res/10, res % 10, resync, max_blocks);

	/*
	 * We do not want to overflow, so the order of operands and
	 * the * 100 / 100 trick are important. We do a +1 to be
	 * safe against division by zero. We only estimate anyway.
	 *
	 * rt: remaining time
	 */
	if (syncstats.stat_sectors == 0)
		return;
	
	/* sectors to 1K blocks */
	syncstats.stat_sectors >>= 1;
	rt = (syncstats.stat_time
	   * ((max_blocks-resync) / (syncstats.stat_sectors/100+1)))/100;

	seq_printf(seq, " finish=%lu.%lumin", rt / 60, (rt % 60)/6);

	seq_printf(seq, " speed=%ldK/sec",
		   syncstats.stat_sectors/syncstats.stat_time);
}

static void *md_seq_start(struct seq_file *seq, loff_t *pos)
{
	struct list_head *tmp;
	loff_t l = *pos;
	mdk_member_t *member;
	u_long flags;

	if (l >= 0x10000)
		return NULL;
	if (!l--)
		/* header */
		return (void*)1;

	spin_lock_irqsave(&toplevel_arrays.member_lock, flags);
	list_for_each(tmp,&toplevel_arrays.child_set)
		if (!l--) {
			member = list_entry(tmp, mdk_member_t, peer_links);
			member_get(member);
			spin_unlock_irqrestore(&toplevel_arrays.member_lock,
					       flags);
			return to_mddev(member);
		}
	spin_unlock_irqrestore(&toplevel_arrays.member_lock, flags);
	if (!l--)
		return (void*)2;/* tail */
	return NULL;
}

static void *md_seq_next(struct seq_file *seq, void *v, loff_t *pos)
{
	struct list_head *tmp;
	mddev_t *next_mddev, *mddev = v;
	mdk_member_t *next_member;
	u_long flags;
	
	++*pos;
	if (v == (void*)2)
		return NULL;

	spin_lock_irqsave(&toplevel_arrays.member_lock, flags);
	if (v == (void*)1)
		tmp = toplevel_arrays.child_set.next;
	else
		tmp = mddev->member.peer_links.next;
	if (tmp != &toplevel_arrays.child_set) {
		next_member = list_entry(tmp,mdk_member_t,peer_links);
		next_mddev = to_mddev(member_get(next_member));
	} else {
		next_mddev = (void*)2;
		*pos = 0x10000;
	}		
	spin_unlock_irqrestore(&toplevel_arrays.member_lock, flags);

	if (v != (void*)1)
		member_put(to_member(mddev));
	return next_mddev;
}

static void md_seq_stop(struct seq_file *seq, void *v)
{
	mddev_t *mddev = v;

	if (mddev && v != (void*)1 && v != (void*)2)
		member_put(to_member(mddev));
}

static int md_seq_show(struct seq_file *seq, void *v)
{
	mddev_t *mddev = v;
	sector_t size;
#if 0
	struct list_head *tmp2;
	mdk_rdev_t *rdev;
#endif
	int i;

	if (v == (void*)1) {
		seq_printf(seq, "Personalities : ");
		spin_lock(&pers_lock);
		for (i = 0; i < MAX_PERSONALITY; i++)
			if (pers[i])
				seq_printf(seq, "[%s] ", pers[i]->name);

		spin_unlock(&pers_lock);
		seq_printf(seq, "\n");
		return 0;
	}
	if (v == (void*)2) {
		status_unused(seq);
		return 0;
	}

	if (mddev_lock(mddev)!=0) 
		return -EINTR;
	if (mddev->member.pers || mddev->raid_disks || !list_empty(&mddev->member.child_set)) {
		seq_printf(seq, "%s : %sactive", mdname(to_member(mddev)),
						mddev->member.pers ? "" : "in");
		if (mddev->member.pers) {
			if (mddev->ro)
				seq_printf(seq, " (read-only)");
			seq_printf(seq, " %s", mddev->member.pers->name);
		}

		size = 0;
#if 0
		ITERATE_RDEV(mddev,rdev,tmp2) {
			char b[BDEVNAME_SIZE];
			seq_printf(seq, " %s[%d]",
				bdevname(rdev->bdev,b), rdev->desc_nr);
			if (rdev->faulty) {
				seq_printf(seq, "(F)");
				continue;
			}
			size += rdev->member.data_size;
		}
#endif

		if (!list_empty(&mddev->member.child_set)) {
			if (mddev->member.pers)
				seq_printf(seq, "\n      %llu blocks",
					(unsigned long long)mddev->member.data_size);
			else
				seq_printf(seq, "\n      %llu blocks",
					(unsigned long long)size);
		}

		if (mddev->member.pers) {
#if XXX
			mddev->member.pers->status (seq, mddev);
	 		seq_printf(seq, "\n      ");
			if (mddev->curr_resync > 2)
				status_resync (seq, mddev);
			else if (mddev->curr_resync == 1 || mddev->curr_resync == 2)
				seq_printf(seq, "	resync=DELAYED");
#endif
		}

		seq_printf(seq, "\n");
	}
	mddev_unlock(mddev);
	
	return 0;
}

static struct seq_operations md_seq_ops = {
	.start  = md_seq_start,
	.next   = md_seq_next,
	.stop   = md_seq_stop,
	.show   = md_seq_show,
};

static int md_seq_open(struct inode *inode, struct file *file)
{
	int error;

	error = seq_open(file, &md_seq_ops);
	return error;
}

static struct file_operations md_seq_fops = {
	.open           = md_seq_open,
	.read           = seq_read,
	.llseek         = seq_lseek,
	.release	= seq_release,
};

mdk_personality_t *md_find_pers(int level)
{
	mdk_personality_t *mdpers;
	int pnum;

	pnum = level_to_pers(level);
	if (pnum >= MAX_PERSONALITY) {
		MD_BUG();
		return NULL;
	}

	spin_lock(&pers_lock);
#ifdef CONFIG_KMOD
	if (!pers[pnum])
	{
		spin_unlock(&pers_lock);
		request_module("md-personality-%d", pnum);
		spin_lock(&pers_lock);
	}
#endif

	if (!pers[pnum] || !try_module_get(pers[pnum]->owner)) {
		spin_unlock(&pers_lock);
		printk(KERN_ERR "md: personality %d is not loaded!\n",
		       pnum);
		return NULL;
	}

	mdpers = pers[pnum];
	spin_unlock(&pers_lock);

	return (mdpers);
}

int register_md_personality(int pnum, mdk_personality_t *p)
{
	if (pnum >= MAX_PERSONALITY) {
		printk(KERN_ERR
		       "md: tried to install personality %s as nr %d, but max is %lu\n",
		       p->name, pnum, MAX_PERSONALITY-1);
		return -EINVAL;
	}

	spin_lock(&pers_lock);
	if (pers[pnum]) {
		spin_unlock(&pers_lock);
		MD_BUG();
		return -EBUSY;
	}

	pers[pnum] = p;
	printk(KERN_INFO "md: %s personality registered as nr %d\n", p->name, pnum);
	spin_unlock(&pers_lock);
	return 0;
}

int unregister_md_personality(int pnum)
{
	if (pnum >= MAX_PERSONALITY) {
		MD_BUG();
		return -EINVAL;
	}

	printk(KERN_INFO "md: %s personality unregistered\n", pers[pnum]->name);
	spin_lock(&pers_lock);
	pers[pnum] = NULL;
	spin_unlock(&pers_lock);
	return 0;
}

int register_md_metapersonality(mdk_metapersonality_t *p)
{
	struct list_head *ph;
	mdk_metapersonality_t *pers;

	printk("md: registering metapers %s\n", p->name);
	spin_lock(&metapersonality_list_lock);
	list_for_each(ph, &md_metapersonality_list) {
		pers = list_entry(ph, mdk_metapersonality_t, links);
		if (strncmp(p->name, pers->name, MD_METAPERS_NAME_LENGTH) == 0){
			spin_unlock(&metapersonality_list_lock);
			printk("md: metapers already registered, ignoring\n");
			return -1;
		}
	}
	list_add(&p->links, &md_metapersonality_list);
	metapers_change_count++;
	spin_unlock(&metapersonality_list_lock);

	/*
	 * See if any unclaimed devices can be
	 * claimed by this personality.
	 */
	autostart_arrays();
	return 0;
}

int unregister_md_metapersonality(mdk_metapersonality_t *p)
{
	printk("md: unregistering metapers %s\n", p->name);

	list_del(&p->links);
	return 0;
}

void md_sync_acct(mdk_member_t *member, unsigned long nr_sectors)
{
#if XXX
	rdev->bdev->bd_contains->bd_disk->sync_io += nr_sectors;
#endif
}

#if 0
static int is_mddev_idle(mddev_t *mddev)
{
	mdk_rdev_t * rdev;
	struct list_head *tmp;
	int idle;
	unsigned long curr_events;

	idle = 1;
	ITERATE_RDEV(mddev,rdev,tmp) {
		struct gendisk *disk = rdev->bdev->bd_contains->bd_disk;
		curr_events = disk_stat_read(disk, read_sectors) + 
				disk_stat_read(disk, write_sectors) - 
				disk->sync_io;
		if ((curr_events - rdev->last_events) > 32) {
			rdev->last_events = curr_events;
			idle = 0;
		}
	}
	return idle;
}
#endif

void md_done_sync(mddev_t *mddev, int blocks, int ok)
{
	/* another "blocks" (512byte) blocks have been synced */
	atomic_sub(blocks, &mddev->recovery_active);
	wake_up(&mddev->recovery_wait);
	if (!ok) {
		set_bit(MD_RECOVERY_ERR, &mddev->recovery);
		md_wakeup_thread(mddev->thread);
		// stop recovery, signal do_sync ....
	}
}

void md_write_start(mddev_t *mddev)
{
#if 0
	if (!atomic_read(&mddev->writes_pending)) {
		mddev_lock_uninterruptible(mddev);
		if (mddev->in_sync) {
			mddev->in_sync = 0;
 			del_timer(&mddev->safemode_timer);
			md_update_sb(mddev);
		}
		atomic_inc(&mddev->writes_pending);
		mddev_unlock(mddev);
	} else
		atomic_inc(&mddev->writes_pending);
#endif
}

void md_write_end(mddev_t *mddev)
{
	if (atomic_dec_and_test(&mddev->writes_pending)) {
		if (mddev->safemode == 2)
			md_wakeup_thread(mddev->thread);
		else
			mod_timer(&mddev->safemode_timer, jiffies + mddev->safemode_delay);
	}
}

DECLARE_WAIT_QUEUE_HEAD(resync_wait);

#define SYNC_MARKS	32
#define	SYNC_MARK_STEP	(HZ/2)

int md_notify_reboot(struct notifier_block *this,
		     unsigned long code, void *x)
{
	struct list_head *tmp;
	mdk_member_t *member;
	mddev_t *mddev;

	if ((code == SYS_DOWN) || (code == SYS_HALT) || (code == SYS_POWER_OFF)) {
		u_long flags;

		printk(KERN_INFO "md: stopping all md devices.\n");

		ITERATE_CHILDREN(&toplevel_arrays, member, tmp, flags) {
			mddev = to_mddev(member);
			if (mddev_trylock(mddev)==0)
				do_md_stop (mddev, 1);
		}
		/*
		 * certain more exotic SCSI devices are known to be
		 * volatile wrt too early system reboots. While the
		 * right place to handle this issue is the given
		 * driver, we do want to have a safe RAID driver ...
		 */
		mdelay(1000*1);
	}
	return NOTIFY_DONE;
}

struct notifier_block md_notifier = {
	.notifier_call	= md_notify_reboot,
	.next		= NULL,
	.priority	= INT_MAX, /* before any real devices */
};

static void md_geninit(void)
{
	struct proc_dir_entry *p;

	dprintk("md: sizeof(mdp_super_t) = %d\n", (int)sizeof(mdp_super_t));

	p = create_proc_entry("mdstat", S_IRUGO, NULL);
	if (p)
		p->proc_fops = &md_seq_fops;
}


/*
 * Searches all registered partitions for autorun RAID arrays
 * at boot time.
 */
struct unclaimed_dev_s
{
	struct	list_head links;
	dev_t	dev;
};

typedef struct unclaimed_dev_s unclaimed_dev_t;

static LIST_HEAD(unclaimed_dev_list);
static spinlock_t unclaimed_dev_list_lock = SPIN_LOCK_UNLOCKED;
static int autodetect_running;
static int autodetect_allowed;

void md_autodetect_dev(dev_t dev)
{
	char b[BDEVNAME_SIZE];
	unclaimed_dev_t *new_dev;

	new_dev = (unclaimed_dev_t *)kmalloc(sizeof(*new_dev), GFP_KERNEL);
	if (new_dev == NULL) {
		printk(KERN_WARNING
		       "md: Failed to create setup item for dev %s.\n",
			__bdevname(dev, b));
		return;
	}
	new_dev->dev = dev;
	spin_lock(&unclaimed_dev_list_lock);
	list_add_tail(&new_dev->links, &unclaimed_dev_list);
	spin_unlock(&unclaimed_dev_list_lock);

	/*
	 * Whenever a new device arrives, probe for metadata.
	 */
	autostart_arrays();
}

static void autostart_arrays(void)
{
	LIST_HEAD(still_unclaimed_list);
	struct list_head *tmp;
	unclaimed_dev_t *cur_dev;
	int last_metapers_change_count;

	/*
	 * Only allow one scan process at a time.
	 * Also honor the core's request to defer
	 * the autodetection process until we are
	 * ready to handle it.
	 */
	spin_lock(&unclaimed_dev_list_lock);
	if (autodetect_running || !autodetect_allowed) {
		spin_unlock(&unclaimed_dev_list_lock);
		return;
	}

	printk(KERN_INFO "md: Autodetecting RAID arrays.\n");
	autodetect_running = 1;

	/*
	 * Record the metadata personality change count so
	 * that we can rescan should the list change mid-scan.
	 */
	spin_lock(&metapersonality_list_lock);
rescan:
	last_metapers_change_count = metapers_change_count;
	spin_unlock(&metapersonality_list_lock);
	while ((tmp = unclaimed_dev_list.next) != &unclaimed_dev_list) {
		mdk_metapersonality_t *pers;
		mdk_metapersonality_t *high_bidder;
		int		       high_bid;
		int		       bid;
		int		       error;

		list_del(tmp);
		spin_unlock(&unclaimed_dev_list_lock);
		cur_dev = list_entry(tmp, unclaimed_dev_t, links);

		/*
		 * Auction off the device to all
		 * registered meta-data modules.
		 */
		high_bidder = NULL;
		high_bid = 0;
		ITERATE_METAPERS(pers, tmp) {

			bid = 0;
			error = pers->dev_op(cur_dev->dev,
					     MDK_DEVOP_CLAIM, &bid);
			if (error != 0) {
				/*
				 * XXX need to bail out early from bidding
				 *     process for some errors?
				 */
				continue;
			}
			if (bid > high_bid) {
				/*
				 * Keep a reference count on the
				 * highest bidder so that this
				 * personality is still available
				 * when we exit our loop.
				 */
				if (high_bidder != NULL)
					metapers_put(high_bidder);
				high_bidder = pers;
				high_bid = bid;
				metapers_get(high_bidder);
			}
		}
		if (high_bidder != NULL) {
			error = high_bidder->dev_op(cur_dev->dev,
						    MDK_DEVOP_ATTACH, NULL);
			metapers_put(high_bidder);
			if (error == 0) {
				kfree(cur_dev);
				spin_lock(&unclaimed_dev_list_lock);
				continue;
			}
			/* FALLTHROUGH */
		}
		list_add_tail(&cur_dev->links, &still_unclaimed_list);
		spin_lock(&unclaimed_dev_list_lock);
	}
	/*
	 * Return any devices that we were unable to
	 * associate into arrays to the unclaimed_dev_list.
	 * A new metadata personality may be registered later
	 * that will claim them.
	 */
	list_splice(&still_unclaimed_list, &unclaimed_dev_list);

	spin_lock(&metapersonality_list_lock);
	if (last_metapers_change_count != metapers_change_count)
		goto rescan;
	spin_unlock(&metapersonality_list_lock);
	autodetect_running = 0;
	spin_unlock(&unclaimed_dev_list_lock);
}

int md_init_workitem(mdk_workitem_t *work, u_int opcode,
		     mdk_member_callback_t *control,
		     mdk_work_handler_t *handler, void *data)
{
	memset(work, 0, sizeof(*work));
	work->work_opcode = opcode;
	work->work_control = control;
	work->work_handler = handler;
	work->work_data = data;
	INIT_LIST_HEAD(&work->links);

	return 0;
}

void
md_cancel_workitems(mdk_workitem_support_t *workinfo)
{
	wait_queue_t wait;

	init_waitqueue_entry(&wait, current);

	spin_lock_irq(&workinfo->work_state_lock);
	add_wait_queue(&workinfo->work_wait, &wait);
	set_current_state(TASK_UNINTERRUPTIBLE);
	while (!list_empty(&workinfo->work_items)) {

		md_interrupt_thread(workinfo->work_thread);
		spin_unlock_irq(&workinfo->work_state_lock);
		blk_run_queues();
		schedule();
		spin_lock_irq(&workinfo->work_state_lock);
	}
	current->state = TASK_RUNNING;
	remove_wait_queue(&workinfo->work_wait, &wait);
	spin_unlock_irq(&workinfo->work_state_lock);
}

/*
 * Wait for a workitem to complete.  Passing in a NULL workitem
 * indicates a desire to wait until the work queue is empty.
 */
void
md_wait_workitem(mdk_workitem_support_t *workinfo, mdk_workitem_t *searchitem)
{
	wait_queue_t wait;

	init_waitqueue_entry(&wait, current);

	spin_lock_irq(&workinfo->work_state_lock);
	add_wait_queue(&workinfo->work_wait, &wait);
	for (;;) {
		struct list_head *list_item;
		int found;
		
		set_current_state(TASK_UNINTERRUPTIBLE);
		found = 0;
		list_item = workinfo->work_items.next;
		while (list_item != &workinfo->work_items) {
			mdk_workitem_t *work;

			work = list_entry(list_item, mdk_workitem_t, links);
			if (work == searchitem || searchitem == NULL) {
				found++;
				break;
			}
			list_item = list_item->next;
		}

		if (!found)
			break;

		spin_unlock_irq(&workinfo->work_state_lock);
		blk_run_queues();
		schedule();
		spin_lock_irq(&workinfo->work_state_lock);
	}
	current->state = TASK_RUNNING;
	remove_wait_queue(&workinfo->work_wait, &wait);
	spin_unlock_irq(&workinfo->work_state_lock);
}

mdk_workitem_flags_t md_find_workitem(mdk_workitem_support_t *workinfo,
				      u_int opcode, int *pfound)
{
	struct list_head *list_item;
	mdk_workitem_flags_t flags;
	int found;

	found = 0;
	flags = MD_WF_NONE;
	spin_lock_irq(&workinfo->work_state_lock);
	list_item = workinfo->work_items.next;
	while (list_item != &workinfo->work_items) {
		mdk_workitem_t *work;

		work = list_entry(list_item, mdk_workitem_t, links);
		if (work->work_opcode == opcode) {
			found++;
			if ((work->work_flags & MD_WF_RUNNING) != 0)
				flags |= MD_WF_RUNNING;
			else if ((work->work_flags & MD_WF_QUEUED) != 0
			      && (flags & MD_WF_RUNNING) == 0)
				flags |= MD_WF_QUEUED;
		}
		list_item = list_item->next;
	}
	spin_unlock_irq(&workinfo->work_state_lock);
	if (pfound != NULL)
		*pfound = found;
	return flags;
}

int md_queue_workitem(mdk_workitem_support_t *workinfo,
		      mdk_workitem_t *work, int head_of_queue)
{
	int error;
	mdk_workitem_t *first_work;

	error = 0;

	if (workinfo->work_thread == NULL) {
		error = -EINVAL;
		printk("md_queue_workitem: no work thread, exiting\n");
		goto bailout;
	}

	work->work_flags = MD_WF_QUEUED;
	spin_lock_irq(&workinfo->work_state_lock);

	/*
	 * If the user wants this item to go to the head of the queue,
	 * we insert this on the list *after* currently running item, if
	 * any.  Then we interrupt the thread to kill the current task.
	 * The new work state will run once the current task has been
	 * terminated.
	 */
	if (head_of_queue != 0
	 &&!list_empty(&workinfo->work_items)) {
		first_work = list_entry(workinfo->work_items.next,
					mdk_workitem_t, links);
		if ((first_work->work_flags & MD_WF_RUNNING) != 0) {
			list_add_tail(&work->links, &first_work->links);
			md_interrupt_thread(workinfo->work_thread);
		} else {
			list_add(&work->links, &workinfo->work_items);
			md_wakeup_thread(workinfo->work_thread);
		}
	} else {
		list_add(&work->links, &workinfo->work_items);

		/*
		 * If a task is running, don't wake up the thread, since
		 * this task will get run after the current one finishes.
		 */

		first_work = list_entry(workinfo->work_items.next,
					mdk_workitem_t, links);
		if ((first_work->work_flags & MD_WF_RUNNING) == 0)
			md_wakeup_thread(workinfo->work_thread);
	}

	spin_unlock_irq(&workinfo->work_state_lock);

bailout:
	return (error);
}

/*
 * Thread handler for running array workitems.
 */
static void
md_handle_workitems(mddev_t *mddev, void *arg)
{
	mdk_workitem_support_t *workinfo;
	mdk_workitem_t *workitem;
	mdk_member_t *member;
	int error;

	if (mddev != NULL)
		member = to_member(mddev);
	else
		member = NULL;
	workinfo = (mdk_workitem_support_t *)arg;

	printk("md_handle_workitems(): called for %p!\n", mddev);

	if (workinfo == NULL) {
		printk("md: md_handle_workitems(): NULL workitem_suppott!\n");
		return;
	}

	printk(KERN_INFO "%s: work thread woken up\n", current->comm);

	spin_lock_irq(&workinfo->work_state_lock);

	while (!list_empty(&workinfo->work_items)) {

		/*
		 * Indicate that we are actively servicing
		 * this workitem.
		 */
		workitem = list_entry(workinfo->work_items.next,
				      mdk_workitem_t, links);
		workitem->work_flags |= MD_WF_RUNNING;
		spin_unlock_irq(&workinfo->work_state_lock);
		error = workitem->work_control(member, MDK_WORKITEM_OP_OPEN,
					       workitem);
		if (error != 0) {
			printk(KERN_ERR
			       "%s: cannot open work item\n", current->comm);
			goto dequeue;
		}

		error = workitem->work_handler(mddev, workitem);

		/*
		 * Dequeue and free the workitem.
		 */
dequeue:
		spin_lock_irq(&workinfo->work_state_lock);
		list_del(&workitem->links);
		workitem->work_control(member, MDK_WORKITEM_OP_CLOSE, workitem);
		spin_unlock_irq(&workinfo->work_state_lock);
		workitem->work_control(member, MDK_WORKITEM_OP_FREE, workitem);

		/*
		 * Indicate to waiters that our queue has changed.
		 */
		wake_up(&workinfo->work_wait);

		/*
		 * Lock for next loop iteration.
		 */
		spin_lock_irq(&workinfo->work_state_lock);

		if (error == -EINTR || error == -ENOMEM)
			break;
	}
	spin_unlock_irq(&workinfo->work_state_lock);
	printk(KERN_INFO "%s: work thread done\n", current->comm);
}

/*
 * Initialize a workitem support structure.  These
 * are usually embedded in the private data area of
 * a RAID personality.
 */
int md_init_workitem_support(mddev_t *mddev, mdk_workitem_support_t *workinfo,
			     const char *thread_string)
{
	memset(workinfo, 0, sizeof(*workinfo));
	spin_lock_init(&workinfo->work_state_lock);
	INIT_LIST_HEAD(&workinfo->work_items);
	init_MUTEX(&workinfo->work_config_sem);
	init_waitqueue_head(&workinfo->work_wait);

	if (thread_string == NULL)
		thread_string = "%s_work";
	workinfo->work_thread = md_register_thread(md_handle_workitems,
						   mddev, workinfo,
						   thread_string);
	if (workinfo->work_thread == NULL)
		return -ENOMEM;
	return 0;
}

void md_teardown_workitem_support(mdk_workitem_support_t *workinfo)
{
	md_unregister_thread(workinfo->work_thread);
}

static int md_sync_safemode_control(mdk_member_t *member, int op, void *arg)
{
	mdk_workitem_t *work;
	mdk_sync_support_t *syncinfo;

	work = (mdk_workitem_t *)arg;
	syncinfo = (mdk_sync_support_t *)work->work_data;
	switch (op) {
	case MDK_WORKITEM_OP_OPEN:
	{
		mddev_t *mddev;

		/*
		 * If it is safe to enter safe mode, mark us
		 * as in safemode and allow the workitem to
		 * put us into safemode.
		 */
		spin_lock_irq(&syncinfo->sync_lock);
		if ((syncinfo->safemode != MDK_SAFEMODE_IDLE_WAIT
		  && syncinfo->safemode != MDK_SAFEMODE_OFF)
		 || syncinfo->writes_pending[0] != 0
		 || syncinfo->writes_pending[1] != 0
		 || syncinfo->sync_pending[0] != 0
		 || syncinfo->sync_pending[1] != 0) {
			spin_unlock_irq(&syncinfo->sync_lock);
			return -EBUSY;
		}
		syncinfo->safemode = MDK_SAFEMODE_ACTIVE;
		mddev = syncinfo->work_state.work_thread->mddev;
		syncinfo->lock_range.start = 0;
		syncinfo->lock_range.end = to_member(mddev)->data_size;
		spin_unlock_irq(&syncinfo->sync_lock);
		break;
	}
	case MDK_WORKITEM_OP_CLOSE:
		/* NOOP */
		break;
	case MDK_WORKITEM_OP_FREE:
		work->work_flags = MD_WF_NONE;
		break;
	default:
		return (-EINVAL);
	}
	return 0;
}

int
md_sync_safemode_handler(mddev_t *mddev, mdk_workitem_t *workitem)
{
	mdk_member_t *member;

	member = to_member(mddev);
	member->pers->raid_op(member, MDK_RAID_OP_ENTER_SAFEMODE, NULL);
	return 0;
}

/*
 * Initialize a sync support structure.  These
 * are usually embedded in the private data area of
 * a RAID personality.
 */
int md_init_sync_support(mddev_t *mddev, mdk_sync_support_t *syncinfo,
			 const char *thread_string)
{
	int error;

	memset(syncinfo, 0, sizeof(*syncinfo));
	error = md_init_workitem_support(mddev, &syncinfo->work_state,
					 thread_string);
	if (error != 0)
		return error;

	atomic_set(&syncinfo->reads_pending, 0);
	spin_lock_init(&syncinfo->sync_lock);

	/*
	 * The lock range is 0 length at the start of
	 * the array when no sync operation is in progress.
	 * This ensures that no newly added sync target
	 * will be trusted until the window opens or is
	 * moved up the volume.  The only exception to
	 * this rule is in safemode.  There we open the
	 * lock range to cover the entire disk so that
	 * all writes block until we exit safemode.  Start
	 * the disk out in safemode.
	 */
	syncinfo->lock_range.start = 0;
	syncinfo->lock_range.end = to_member(mddev)->data_size;
	syncinfo->safemode = MDK_SAFEMODE_ACTIVE;
	init_waitqueue_head(&syncinfo->io_wait);
	init_waitqueue_head(&syncinfo->sync_wait);

	md_init_workitem(&syncinfo->safemode_workitem, MD_SYNCOP_ENTER_SAFEMODE,
			 md_sync_safemode_control, md_sync_safemode_handler, 
			 syncinfo);
	init_timer(&syncinfo->safemode_timer);
	syncinfo->safemode_timer.function = md_safemode_timeout;
	syncinfo->safemode_timer.data = (unsigned long)syncinfo;
	syncinfo->safemode_delay = 20 * HZ;

	/*
	 * A zero window size, io_limit_min and max
	 * indicate use the defaults.
	 */
	return 0;
}

void md_teardown_sync_support(mddev_t *mddev, mdk_sync_support_t *syncinfo)
{
	md_teardown_workitem_support(&syncinfo->work_state);
}

/*
 * Generic workitem handler that "walks" a volume
 * issuing sync (resync, initialize, verify, verify/fix)
 * operations.  The actual nature of the sync operation
 * is unknown to this routine but it does guarantee that
 * oustanding sync operations have exclusive write access
 * to the range on which they operate.
 */
int md_sync_handler(mddev_t *mddev, mdk_workitem_t *workitem)
{
	mdk_checkpoint_arg_t checkpoint_arg;
	mdk_sync_support_t *syncinfo;
	mdk_member_t *member;
	mdk_metapersonality_t *metapers;
	mdk_personality_t *pers;
	char *verb, *noun;
	u_long stat_sectors[SYNC_MARKS];
	u_long stat_time[SYNC_MARKS];
	u_long currspeed;
	sector_t sync_start;
	int error;
	int stat_bucket;

	syncinfo = (mdk_sync_support_t *)workitem->work_data;
	member = to_member(mddev);
	metapers = member->metapers;
	pers = member->pers;
 
	switch (workitem->work_opcode) {
	case MD_SYNCOP_REBUILD:
		verb = "reconstructing";
		noun = "reconstruction";
		break;
	case MD_SYNCOP_INIT:
		verb = "initializing";
		noun = "initialization";
		break;
	case MD_SYNCOP_VERIFY:
	case MD_SYNCOP_VERIFY_FIX:
		verb = "verifying";
		noun = "verification";
		break;
	default:
		/* silence warnings */
		verb = NULL;
		noun = NULL;
		MD_BUG();
		break;
	}

	printk(KERN_INFO "md: %s RAID array %s\n", verb, mdname(member));
	printk(KERN_INFO "md: minimum _guaranteed_ %s speed: %d KB/sec/disc.\n",
	       noun, sysctl_speed_limit_min);
	printk(KERN_INFO "md: using maximum available idle IO bandwith\n");
	printk(KERN_INFO "md: (but not more than %d KB/sec) for %s.\n",
	       sysctl_speed_limit_max, noun);

	/*
	 * Initialize our statistics counters.
	 */
	for (stat_bucket = 0; stat_bucket < SYNC_MARKS; stat_bucket++) {
		stat_sectors[stat_bucket] = 0;
		stat_time[stat_bucket] = jiffies;
	}
	stat_bucket = 0;
	syncinfo->stat_sectors = 0;

	/*
	 * We start the sync operation at the last checkpointed
	 * location that is applicable to this type of workitem.
	 */
	checkpoint_arg.checkpoint_type = workitem->work_opcode;
	error = metapers->member_op(member,
				    MDK_MEMBER_OP_GET_CHECKPOINT,
				    &checkpoint_arg);
	if (error != 0) {
		/*
		 * XXX record workitem failures somewhere?
		 */
		return error;
	}

	/*
	 * Round the starting sector down to a window boundary.  The
	 * RAID personalities should specify a window that is an even
	 * multiple of any minimum I/O size required to do a sync.
	 * We assume that window_size is a power of 2.
	 */
	sync_start = checkpoint_arg.checkpoint & ~(syncinfo->window_size - 1);
	syncinfo->sync_pending[0] = 0;
	syncinfo->sync_pending[1] = 0;
	syncinfo->sync_range[syncinfo->phase].start = sync_start;
	syncinfo->sync_range[syncinfo->phase].end = sync_start;

	spin_lock_irq(&syncinfo->sync_lock);
	/*
	 * Initialize our lock_range.  The main loop opens
	 * the range, so don't do that here.
	 */
	syncinfo->lock_range.start = sync_start;
	syncinfo->lock_range.end = sync_start;

	/*
	 * Main sync loop;
	 * - if more syncs to issue (e.g. lock_range.end < array-size), loop.
	 * - else, wait for lock_range to close
	 * - return.
	 */
	while (syncinfo->lock_range.start < member->data_size) {
		int old_phase;
		int new_phase;

		old_phase = syncinfo->phase;
		new_phase = syncinfo->phase ^ 1;

		/*
		 * Wait for sync I/O in oldest phase to cease.  Once
		 * we toggle the phase, this oldest sync window will
		 * become the newest one, so we must be careful not
		 * to collide with any pending operations.
		 */
		wait_event_lock_irq(syncinfo->io_wait,
				    (syncinfo->sync_pending[new_phase] == 0),
				    syncinfo->sync_lock);
		/*
		 * Update the checkpoint of this operation.
		 */
		checkpoint_arg.checkpoint = syncinfo->sync_range[old_phase].end;
		spin_unlock_irq(&syncinfo->sync_lock);
		error = metapers->member_op(member,
					    MDK_MEMBER_OP_SET_CHECKPOINT,
					    &checkpoint_arg);
		if (error != 0) {
			/*
			 * XXX record workitem failures somewhere?
			 *     close sync??
			 */
			return error;
		}
		spin_lock_irq(&syncinfo->sync_lock);

		/*
		 * Toggle the phase, extend the lock range, and create a
		 * new sync range to cover the new extention of the lock
		 * range.
		 */
		if (syncinfo->sync_range[old_phase].end
		 != syncinfo->lock_range.end)
			MD_BUG();
		syncinfo->phase = new_phase;
		syncinfo->sync_range[new_phase].start =
		    syncinfo->lock_range.end;
		syncinfo->lock_range.end += syncinfo->window_size;
		if (syncinfo->lock_range.end > member->data_size)
			syncinfo->lock_range.end = member->data_size;
		syncinfo->sync_range[new_phase].end = syncinfo->lock_range.end;

		/*
		 * Before starting any I/O in the new sync range, wait
		 * for outstanding I/O in the old phase to complete.
		 * This guarantees that only I/O operations that have
		 * seen the new lock range are outstanding.
		 */
		wait_event_lock_irq(syncinfo->sync_wait,
				    (syncinfo->writes_pending[old_phase] == 0),
				    syncinfo->sync_lock);
		spin_unlock_irq(&syncinfo->sync_lock);

		/*
		 * Issue syncs in the new range closing the new sync
		 * range as we go.
		 */
		while (syncinfo->sync_range[new_phase].start
		     < syncinfo->sync_range[new_phase].end) {
			sector_t issued;
	
			issued = syncinfo->sync_range[new_phase].start;
			error = pers->raid_op(member, MDK_RAID_OP_SYNC_REQUEST,
					      workitem);
			if (error) {
				printk(KERN_INFO "%s: md_sync_handler() "
				       "error on sync I/O... exiting\n",
				       mdname(member));
				spin_lock_irq(&syncinfo->sync_lock);
				goto out;
			}
			issued = syncinfo->sync_range[new_phase].start - issued;
			stat_sectors[stat_bucket] += issued;
test_throttle:
			if (jiffies >= stat_time[stat_bucket]) {
				int next_bucket;

				/*
				 * Rotate statistics and update
				 * utility visible statistics counters.
				 * Record actual time of switch
				 * for proper accounting.
				 */
				stat_time[stat_bucket] = jiffies;
				next_bucket = stat_bucket % SYNC_MARKS;
				syncinfo->stat_sectors +=
				    stat_sectors[stat_bucket];
				syncinfo->stat_sectors -=
				    stat_sectors[next_bucket];
				syncinfo->stat_time =
				    jiffies - stat_time[next_bucket];
				stat_sectors[next_bucket] = 0;
				stat_time[next_bucket] =
				    jiffies + SYNC_MARK_STEP;
				stat_bucket = next_bucket;
			}

			if (signal_pending(current)) {
				/*
				 * got a signal, exit.
				 */
				printk(KERN_INFO "%s: md_sync_handler() "
				       "got signal ... exiting\n",
				       mdname(member));
				flush_signals(current);
				error = -EINTR;
				spin_lock_irq(&syncinfo->sync_lock);
				goto out;
			}

			/*
			 * Throttle our resync based on calculated I/O
			 * load so long as no normal I/O is blocked waiting
			 * on our resync window.  Throttling kicks in any
			 * time we exceed the maximum resync rate as well
			 * as anytime normal I/O is in progress as we are
			 * exceeding the guaranteed resync rate.
			 */
			cond_resched();
			if (syncinfo->stat_sectors == 0)
				currspeed = 0;
			else
				currspeed = (syncinfo->stat_sectors >> 1)
				    / (syncinfo->stat_time/HZ);
			if (!waitqueue_active(&syncinfo->io_wait)
			 && (currspeed > sysctl_speed_limit_max
			  || (currspeed > sysctl_speed_limit_min
			   && (syncinfo->writes_pending[syncinfo->phase]
			    || atomic_read(&syncinfo->reads_pending))))) {
				current->state = TASK_INTERRUPTIBLE;
				schedule_timeout(HZ/4);
				goto test_throttle;
			}
		}
		spin_lock_irq(&syncinfo->sync_lock);
	}

out:
	/*
	 * Now that all sync requests have been issued, wait for
	 * the syncs to complete, write a final checkpoint, release
	 * the lock_range, and exit.  We forst the end of the sync
	 * ranges to be the same as the start position just in
	 * case we exited due to a signal or error.  Otherwise
	 * the completion handlers will not know to signal us
	 * when the pending sync count drops to 0.
	 */
 	syncinfo->sync_range[0].end = syncinfo->sync_range[0].start;
 	syncinfo->sync_range[1].end = syncinfo->sync_range[1].start;
	wait_event_lock_irq(syncinfo->sync_wait,
			    (syncinfo->sync_pending[0] == 0
			  && syncinfo->sync_pending[1] == 0),
			    syncinfo->sync_lock);
	syncinfo->lock_range.start = 0;
	syncinfo->lock_range.end = 0;
	spin_unlock_irq(&syncinfo->sync_lock);
	checkpoint_arg.checkpoint = syncinfo->sync_range[syncinfo->phase].end;
	metapers->member_op(member, MDK_MEMBER_OP_SET_CHECKPOINT,
			    &checkpoint_arg);
	return error;
}

static mdk_metapersonality_t *md_check_metadata_support(char *metaname)
{
	struct list_head *ph;
	mdk_metapersonality_t *pers;

	printk("md: checking metatype support for < %s > :", metaname);
	spin_lock(&metapersonality_list_lock);
	list_for_each(ph, &md_metapersonality_list) {
		pers = list_entry(ph, mdk_metapersonality_t, links);
		if (strncmp(metaname, pers->name, MD_METAPERS_NAME_LENGTH) == 0){
			spin_unlock(&metapersonality_list_lock);
			printk("Supported\n");
			return pers;
		}
	}
	spin_unlock(&metapersonality_list_lock);
	printk("Unsupported\n");
	
	return NULL;
}

/*
 * Mark meta-data as open, release sync_lock range, and update
 * safemode state to reflect that safemode is off.  Called with
 * the syncinfo lock held.
 */
void md_exit_safemode(mddev_t *mddev, mdk_sync_support_t *syncinfo)
{
	mdk_member_t *member;

	/*
	 * Prevent other I/Os from attempting to end
	 * safe-mode by indicating that we are leaving
	 * safemode.  Future writes will still end up
	 * blocking on the lock-range that still covers
	 * the entire volume.
	 */
	syncinfo->safemode = MDK_SAFEMODE_EXIT;
	spin_unlock_irq(&syncinfo->sync_lock);

	member = to_member(mddev);

	member->pers->raid_op(member, MDK_RAID_OP_EXIT_SAFEMODE, NULL);

	spin_lock_irq(&syncinfo->sync_lock);
	syncinfo->safemode = MDK_SAFEMODE_OFF;
	syncinfo->lock_range.start = 0;
	syncinfo->lock_range.end = 0;

	/*
	 * Wakeup any I/O blocked on the exit from safemode.
	 */
	wake_up(&syncinfo->io_wait);
}

static int set_disk_faulty(mddev_t *mddev, dev_t dev)
{
	mdk_member_t *member;

	member = get_member_from_device(&toplevel_arrays, dev);
	if (member == NULL
	 || member->type != MDK_MTYPE_RDEV)
		return -ENODEV;

	md_set_member_state(member, MDK_MSTATE_FAILED);

	member_put(member);

	return 0;
}

static int md_ctl_open(struct inode *inode, struct file *file)
{
	/* XXX Should do a refcount here and limit the number of openers. */
	return (0);
}

static int md_ctl_ioctl(struct inode *inode, struct file *file,
			unsigned int cmd, unsigned long arg)
{
	int err = 0;
	mddev_t *mddev = NULL;
	mdu_disk_info_t *disk;
	mdu_cookie_info_t *cookie_info;
	mdk_member_t *member = NULL;
	dev_t dev;
	
	if (!capable(CAP_SYS_ADMIN))
		return -EACCES;

	dev = inode->i_rdev;

	/*
	 * Commands dealing with the RAID driver but not any
	 * particular array:
	 */
	switch (cmd) {
	case RAID_VERSION:
		err = get_version((void *)arg);
		goto done;
#if 0
	case PRINT_RAID_DEBUG:
		err = 0;
		md_print_devices();
		goto done;
	case RAID_AUTORUN:
		err = 0;
		autostart_arrays();
		goto done;
#endif
	default:;
	}
	
/* The ioctls below are working */	
	switch (cmd) {
	case EMD_GET_DEVICE_INFO:
	{
		unsigned int major, minor;
		mdu_cookie_info_t *pcookie = (mdu_cookie_info_t *)arg;
		mdu_cookie_info_t cookie;
		mdk_member_t *member;

		copy_from_user(&cookie, pcookie, sizeof(cookie));
		major = cookie.major_minor.major;
		minor = cookie.major_minor.minor;
		member = get_member_from_device
			(&toplevel_arrays, MKDEV(major, minor));
		if (member == NULL) {
			return -ENODEV;
		}
		get_cookie_info(member, &cookie);
		if (copy_to_user(pcookie, &cookie, sizeof(cookie))) {
			return -EFAULT;
		}
		//break;
		return 0;
	}
	/* This IOCTL passes an array structure as the arg */
	case EMD_GET_ARRAY_INFO:
	{
		mdu_array_info_t array_info;
		mddev_t * pmddev = NULL;
		mdk_member_t *member = NULL;
		
		copy_from_user(&array_info, (void *)arg, sizeof(array_info));
		member = get_member_from_member( &toplevel_arrays, 
				( mdk_member_t * ) array_info.cookie);
		
		/* If this is an rdev, you really want */
		/* then get the array info from the parent. */
		pmddev = to_mddev(member);
		if (member && (member->type == MDK_MTYPE_RDEV)) {
			member_put((mdk_member_t *) member);
			member = get_member_from_member(&toplevel_arrays, 
						pmddev->member.parent);
		}
		if (member == NULL) {
			return -EFAULT;
		}
		pmddev = to_mddev(member);
		err = get_array_info(pmddev, (void *)arg);
		//break;
		return 0;
	}
	case EMD_SET_ARRAY_INFO:
	{
		mdu_array_info_t array_k;
		mdu_array_info_t *parray_info = (mdu_array_info_t *)arg;
		mddev_t *pmddev;
		
		copy_from_user(&array_k, parray_info, sizeof(array_k));
		dev = MKDEV(mdp_major, array_k.md_minor);
		
		member = get_member_from_device(&toplevel_arrays, dev);
		pmddev = to_mddev(member);
		
		if (pmddev == NULL) {
			/* minor not in use already - OK
			 * validate metatype and call app mddev alloc routine 
			 * using member_op for that meta type
			 */
			mdk_metapersonality_t *metapers;
			if ((metapers = md_check_metadata_support(array_k.metaname)) == NULL) {
				return -EFAULT;
			}
			metapers->member_op(member, MDK_MEMBER_OP_ALLOC_ARRAY_INFO, &array_k);

		}else{
			printk("**Array already exists***\n");
			return -EFAULT;
		}

#if 0		
		err = mddev_lock(mddev);
		if (err) {
			printk(KERN_INFO 
			       "md: ioctl lock interrupted, reason %d, cmd %d\n",
				err, cmd);
			goto abort;
		}

		if (!list_empty(&mddev->member.child_set)) {
			printk(KERN_WARNING 
				"md: array %s already has disks!\n",
				mdname(to_member(mddev)));
			err = -EBUSY;
			goto abort_unlock;
		}
		if (mddev->raid_disks) {
			printk(KERN_WARNING 
				"md: array %s already initialised!\n",
				mdname(to_member(mddev)));
			err = -EBUSY;
			goto abort_unlock;
		}
		{
			mdu_array_info_t info;
			if (!arg)
				memset(&info, 0, sizeof(info));
			else if (copy_from_user(&info, (void*)arg, sizeof(info))) {
				err = -EFAULT;
				goto abort_unlock;
			}
			err = set_array_info(mddev, &info);
			if (err) {
				printk(KERN_WARNING "md: couldn't set"
					" array info. %d\n", err);
				goto abort_unlock;
			}
		}
		goto done_unlock;
#endif		
		return 0;
	}
	default:
		break;
	}

	switch (cmd) {
	case EMD_STOP_ARRAY_RO:
	case EMD_RESTART_ARRAY_RW:
	{
		dev_t dev;

		mdu_major_minor_t major_minor;
		
		err = copy_from_user(&major_minor, (void *)arg,
			      	     sizeof(mdu_major_minor_t));
		if (err)
			return err;
		if (major_minor.minor % (1 << MDP_MINOR_SHIFT))
			return -EPERM;
		dev = MKDEV(major_minor.major,
			    major_minor.minor);
		member = get_member_from_device(&toplevel_arrays, dev);
		if (!member)
			return -EINVAL;
		mddev = to_mddev(member);
		if (cmd == EMD_STOP_ARRAY_RO)
			err = do_md_ro(mddev, 1);
		else
			err = do_md_ro(mddev, 0);
		member_put(member);
		return err;
	}
	case EMD_KILL_META_DATA:
	{
		mdu_major_minor_t major_minor;
		dev_t kdev;
		mdk_member_t *member;
		mdk_metapersonality_t *pers;
		struct list_head *tmp;
			
		err = copy_from_user(&major_minor, (void *)arg,
			      	     sizeof(mdu_major_minor_t));
		if (err)
			return err;
		if (major_minor.minor % (1 << MDP_MINOR_SHIFT))
			return -EPERM;
		kdev = MKDEV(major_minor.major, major_minor.minor);
		member = get_member_from_device(&toplevel_arrays, kdev);
		if (member != NULL) {
			/*
			 * kdev is part of an array
			 */
			member_put(member);
			return -EBUSY;
		}

		/*
		 * Checking if valid metadata is present before zeroing.
		 */
		err = -EINVAL;
		ITERATE_METAPERS(pers, tmp) {
			int bid;
			int error;

			bid = 0;
			error = pers->dev_op(kdev, MDK_DEVOP_CLAIM, &bid);
			if (error != 0)
				continue;
			if (bid > 0) {
				pers->dev_op(kdev, MDK_DEVOP_WIPE_METADATA,
					     NULL);
				err = 0;
			}
		}
		return err;
	}
	case EMD_GET_FREESIZE:
	{
		mdu_device_metapers_t mpers_dev;
		dev_t kdev;
		mdk_metapersonality_t *pers;
		struct list_head *tmp;
		int found = 0;
		sector_t freesize;
		
		err = copy_from_user(&mpers_dev, (void *)arg, 
				     sizeof(mdu_device_metapers_t));
		if(err)
			return err;
		
		kdev = MKDEV(mpers_dev.major_minor.major, 
			     mpers_dev.major_minor.minor);
		
		ITERATE_METAPERS(pers, tmp) {
			if(strcmp(pers->name, mpers_dev.metapers_name) != 0)
			       	continue;
			else{
				found = 1;
				break;
			}
		}
		if(!found){
			printk("metapers %s not registered\n", 
				mpers_dev.metapers_name);
			return (-EINVAL);
		}
		err = pers->dev_op(kdev, MDK_DEVOP_GET_FREESIZE, &freesize);
		copy_to_user((void *)arg, &freesize, sizeof(sector_t));
		return err;
	}
	default:
		break;
	}

	/*
	 * Commands creating/starting a new array:
	 */
#if 0        
	switch (cmd) {
// TBD .. This need work. The call to get_member_from_device() needs to be replaced
//        with something that creates an "array" type member ...  		
	case EMD_SET_ARRAY_INFO:
		array = (mdu_array_info_t *) arg;
		device = MKDEV(mdp_major, array->md_minor);
		member = get_member_from_device(&toplevel_arrays, device);
		mddev = (mddev_t *) member;
		if (!mddev) {
			err = -ENODEV;
			printk(KERN_WARNING "md: NULL member for EMD_SET_ARRAY IOCTL\n" );
			goto abort;
		}

		err = mddev_lock(mddev);
		if (err) {
			printk(KERN_INFO 
				"md: ioctl lock interrupted, reason %d, cmd %d\n",
				err, cmd);
			goto abort;
		}

		if (!list_empty(&mddev->member.child_set)) {
			printk(KERN_WARNING 
				"md: array %s already has disks!\n",
				mdname(to_member(mddev)));
			err = -EBUSY;
			goto abort_unlock;
		}
		if (mddev->raid_disks) {
			printk(KERN_WARNING 
				"md: array %s already initialised!\n",
				mdname(to_member(mddev)));
			err = -EBUSY;
			goto abort_unlock;
		}
		{
			mdu_array_info_t info;
			if (!arg)
				memset(&info, 0, sizeof(info));
			else if (copy_from_user(&info, (void*)arg, sizeof(info))) {
				err = -EFAULT;
				goto abort_unlock;
			}
			err = set_array_info(mddev, &info);
			if (err) {
				printk(KERN_WARNING "md: couldn't set"
					" array info. %d\n", err);
				goto abort_unlock;
			}
		}
		goto done_unlock;

	default:;
	}
#endif	

	/* We need to get the appropriate mddev */
	switch (cmd) {
	/* These IOCTLs pass the "cookie" value as the arg */
	case EMD_RESTART_ARRAY_RW:
	case EMD_STOP_ARRAY:
	case EMD_STOP_ARRAY_RO:
		member = get_member_from_member( &toplevel_arrays, 
					(mdk_member_t *) arg);
		break;
		
	/* These IOCTLs pass a "cookie" structure as the arg */
	case EMD_GET_MEMBER_INFO:
		cookie_info = ( mdu_cookie_info_t *) arg;
		member = get_member_from_member(&toplevel_arrays, 
					(mdk_member_t *) cookie_info->cookie);
		break;
	
	/* These IOCTLs pass a disk structure as the arg */	
	case EMD_GET_DISK_INFO:
	case EMD_ADD_NEW_DISK:
		disk = (mdu_disk_info_t *) arg;
		dev = MKDEV(disk->major, disk->minor);
		member = get_member_from_device(&toplevel_arrays, dev);
		break;
	case EMD_GET_DISK_INFO_FROM_PARENT:
		disk = (mdu_disk_info_t *) arg;
		member = get_child_from_index( (mdk_member_t *) disk->parent_cookie,
					 disk->number );
		break;
		
	/* These IOCTLs pass a device number as the arg */	
	case HOT_GENERATE_ERROR:
	case EMD_HOT_REMOVE_DISK:
	case EMD_HOT_ADD_DISK:
	case EMD_SET_DISK_FAULTY:
	case EMD_RUN_ARRAY:
		member = get_member_from_device(&toplevel_arrays, (int) arg);
		break;
	default:
		goto done;
		break;

	}
	
	mddev = (mddev_t *) member;

	/* Error check so we don't crash and burn if the mddev is NULL */
	if (!mddev) {
		err = -ENODEV;
		printk(KERN_WARNING "md: NULL member for IOCTL\n" );
		goto abort;
	}

	if (mddev->member.type == MDK_MTYPE_MDDEV ) {
 		err = mddev_lock(mddev);
		if (err) {
			printk(KERN_INFO 
				"md: ioctl lock interrupted, reason %d, cmd %d\n",
				err, cmd);
			goto abort;
		}
	}
		
	/*
	 * Commands querying/configuring an existing array:
	 */
	 
	/* if we are initialised yet, only ADD_NEW_DISK or STOP_ARRAY is allowed */
	if ((mddev) && (mddev->member.type == MDK_MTYPE_MDDEV )){
		if (!mddev->raid_disks &&
		     cmd != EMD_ADD_NEW_DISK &&
		     cmd != EMD_STOP_ARRAY &&
		     cmd != EMD_RUN_ARRAY) {
			err = -ENODEV;
			goto abort_unlock;
		}
	}
	
	/*
	 * Commands even a read-only array can execute:
	 */
	switch (cmd) {
	case EMD_GET_MEMBER_INFO:
		err = get_cookie_info(( mdk_member_t *) mddev, (void *) arg);		
		goto done_unlock;

	case EMD_GET_DISK_INFO:
	case EMD_GET_DISK_INFO_FROM_PARENT:
		err = get_disk_info(mddev, (void *)arg);
		goto done_unlock;

	case EMD_RESTART_ARRAY_RW:
		err = restart_array(mddev);
		goto done_unlock;

	case EMD_STOP_ARRAY:
		err = do_md_stop (mddev, 0);
		goto done_unlock;

	default:
		break;
	}

	/*
	 * The remaining ioctls are changing the state of the
	 * superblock, so we do not allow read-only arrays
	 * here:
	 */
		
	if (mddev->ro) {
		err = -EROFS;
		goto abort_unlock;
	}

	switch (cmd) {
	case EMD_ADD_NEW_DISK:
	{
		mdu_disk_info_t info;
		if (copy_from_user(&info, (void *) arg, sizeof(info)))
			err = -EFAULT;
		else
			err = add_new_disk(mddev, &info );
		goto done_unlock;
	}	
	case HOT_GENERATE_ERROR:
		err = hot_generate_error(mddev, new_decode_dev(arg));
		goto done_unlock;
		
	case EMD_HOT_REMOVE_DISK:
		err = hot_remove_disk(mddev, new_decode_dev(arg));
		goto done_unlock;

	case EMD_HOT_ADD_DISK:
		err = hot_add_disk(mddev, new_decode_dev(arg));
		goto done_unlock;

	case EMD_SET_DISK_FAULTY:
		err = set_disk_faulty(mddev, new_decode_dev(arg));
		goto done_unlock;

	case EMD_RUN_ARRAY:
		err = md_runnable(mddev);
		goto done_unlock;

	default:
		if (_IOC_TYPE(cmd) == MD_MAJOR)
			printk(KERN_WARNING "md: %s(pid %d) used"
				" obsolete MD ioctl, upgrade your"
				" software to use new ictls.\n",
				current->comm, current->pid);
		err = -EINVAL;
		goto abort_unlock;
	}

done_unlock:
abort_unlock:
	if (mddev->member.type == MDK_MTYPE_MDDEV ) 
		mddev_unlock(mddev);
done:
	if (err)
		MD_BUG();
abort:
	if (mddev)
		member_put( (mdk_member_t *) mddev);
	return err;
}

static int md_ctl_major;
static struct file_operations md_ctl_fops = {
	.owner	= THIS_MODULE,
	.open	= md_ctl_open,
	.ioctl	= md_ctl_ioctl,
};

static struct kobject *md_probe(dev_t dev, int *part, void *data)
{
	return NULL;
}

int __init md_init(void)
{
	mddev_t *mddev;
	mdk_member_t *member;
	struct list_head *tmp;
	u_long flags;
	int minor;

	printk(KERN_INFO "md: md driver %d.%d.%d MAX_MD_DEVS=%d,"
			" MD_SB_DISKS=%d\n",
			MD_MAJOR_VERSION, MD_MINOR_VERSION,
			MD_PATCHLEVEL_VERSION, MAX_MD_DEVS, MD_SB_DISKS);

	if (register_blkdev(MAJOR_NR, "md"))
		return -1;
	if ((mdp_major=register_blkdev(0, "mdp"))<=0) {
		unregister_blkdev(MAJOR_NR, "md");
		return -1;
	}
	devfs_mk_dir("md");
	blk_register_region(MKDEV(MAJOR_NR, 0), MAX_MD_DEVS, THIS_MODULE,
			    md_probe, NULL, NULL);
	blk_register_region(MKDEV(mdp_major, 0), MAX_MD_DEVS<<MDP_MINOR_SHIFT,
			    THIS_MODULE, md_probe, NULL, NULL);

	for (minor=0; minor < MAX_MD_DEVS; ++minor)
		devfs_mk_bdev(MKDEV(MAJOR_NR, minor),
				S_IFBLK|S_IRUSR|S_IWUSR,
				"md/%d", minor);

	for (minor=0; minor < MAX_MD_DEVS; ++minor)
		devfs_mk_bdev(MKDEV(mdp_major, minor<<MDP_MINOR_SHIFT),
			      S_IFBLK|S_IRUSR|S_IWUSR,
			      "md/d%d", minor);

	md_ctl_major = register_chrdev(0, "mdctl", &md_ctl_fops);
	if (md_ctl_major < 0)
		printk(KERN_WARNING
		       "md_init: unable to register mdctl device\n");

	register_reboot_notifier(&md_notifier);
	raid_table_header = register_sysctl_table(raid_root_table, 1);

	md_geninit();
	autodetect_allowed = 1;
	autostart_arrays();

	runarrays_allowed = 1;
	ITERATE_CHILDREN(&toplevel_arrays, member, tmp, flags) {
		mdk_personality_t *pers = member->pers;
		
		mddev = to_mddev(member);
		printk("md_init: calling MDK_RAID_OP_RUNNABLE for %p\n",
			mddev);
		if (!pers->raid_op(to_member(mddev), MDK_RAID_OP_RUNNABLE, 0))
			md_runnable(mddev);
	}
	return (0);
}

static __exit void md_exit(void)
{
	mdk_member_t *member;
	mddev_t *mddev;
	struct list_head *tmp;
	u_long flags;
	int i;

	blk_unregister_region(MKDEV(MAJOR_NR,0), MAX_MD_DEVS);
	blk_unregister_region(MKDEV(mdp_major,0),
			      MAX_MD_DEVS << MDP_MINOR_SHIFT);
	for (i=0; i < MAX_MD_DEVS; i++)
		devfs_remove("md/%d", i);
	for (i=0; i < MAX_MD_DEVS; i++)
		devfs_remove("md/d%d", i);

	devfs_remove("md");

	if (md_ctl_major >= 0)
		unregister_chrdev(md_ctl_major, "mdctl");

	unregister_blkdev(MAJOR_NR,"md");
	unregister_blkdev(mdp_major, "mdp");
	unregister_reboot_notifier(&md_notifier);
	unregister_sysctl_table(raid_table_header);
	remove_proc_entry("mdstat", NULL);
	ITERATE_CHILDREN(&toplevel_arrays, member, tmp, flags) {
		struct gendisk *disk;

		mddev = to_mddev(member);
		disk = mddev->gendisk;
		if (!disk)
			continue;
		export_array(mddev);
		del_gendisk(disk);
		put_disk(disk);
		mddev->gendisk = NULL;
		member_put(member);
	}
}

module_init(md_init)
module_exit(md_exit)

EXPORT_SYMBOL(register_md_personality);
EXPORT_SYMBOL(unregister_md_personality);
EXPORT_SYMBOL(register_md_metapersonality);
EXPORT_SYMBOL(unregister_md_metapersonality);
EXPORT_SYMBOL(md_sync_acct);
EXPORT_SYMBOL(md_done_sync);
EXPORT_SYMBOL(md_write_start);
EXPORT_SYMBOL(md_write_end);
EXPORT_SYMBOL(md_register_thread);
EXPORT_SYMBOL(md_unregister_thread);
EXPORT_SYMBOL(md_update_thread_string);
EXPORT_SYMBOL(md_wakeup_thread);
EXPORT_SYMBOL(md_print_devices);
EXPORT_SYMBOL(md_interrupt_thread);
EXPORT_SYMBOL(md_read_blocks);
EXPORT_SYMBOL(md_write_blocks);
EXPORT_SYMBOL(md_member_add_child);
EXPORT_SYMBOL(md_member_orphan);
EXPORT_SYMBOL(md_rdev_alloc);
EXPORT_SYMBOL(md_mddev_alloc);
EXPORT_SYMBOL(md_find_pers);
EXPORT_SYMBOL(md_check_chunksize);
EXPORT_SYMBOL(toplevel_arrays);
EXPORT_SYMBOL(md_init_sync_support);
EXPORT_SYMBOL(md_teardown_sync_support);
EXPORT_SYMBOL(md_cancel_workitems);
EXPORT_SYMBOL(md_wait_workitem);
EXPORT_SYMBOL(md_teardown_workitem_support);
EXPORT_SYMBOL(md_sync_handler);
EXPORT_SYMBOL(md_exit_safemode);
EXPORT_SYMBOL(md_init_workitem);
EXPORT_SYMBOL(md_find_workitem);
EXPORT_SYMBOL(md_queue_workitem);
MODULE_LICENSE("GPL");
