/*
 * Copyright (c) 2001, 2002, 2004 Adaptec Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification.
 * 2. Redistributions in binary form must reproduce at minimum a disclaimer
 *    substantially similar to the "NO WARRANTY" disclaimer below
 *    ("Disclaimer") and any redistribution must be conditioned upon
 *    including a substantially similar Disclaimer requirement for further
 *    binary redistribution.
 * 3. Neither the names of the above-listed copyright holders nor the names
 *    of any contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * NO WARRANTY
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/config.h>
#include <linux/raid/md.h>
#include "md_metadata_adaptec.h"

/*
 * DPT RAID table is 8K, and the reserved block is 512 bytes.  These two
 * blocks are at the very end of the disk -- they take up the last 17 512
 * byte sectors of the disk.  The reserved block is the last block of the
 * disk, and contains a pointer to the start of the raid table, which
 * generally takes up the 8KB before the reserved block.
 */
#define ADAPTEC_ASR_RT_SIZE	8192
#define ADAPTEC_ASR_RB_SIZE	512

/*
 * The priority this module returns
 * when claiming a device.
 */
#define	MD_ADAPTEC_CLAIM_PRIORITY	1

/*
 * Modular arithmetic macros to deal with ASR RAID table sequence numbers,
 * which are 16 bit unsigned integers that can wrap.
 */
#define	ASR_SEQ_LT(a,b)		((int16_t)((a)-(b)) < 0)
#define	ASR_SEQ_LEQ(a,b)	((int16_t)((a)-(b)) <= 0)
#define	ASR_SEQ_GT(a,b)		((int16_t)((a)-(b)) > 0)
#define	ASR_SEQ_GEQ(a,b)	((int16_t)((a)-(b)) >= 0)

#define MD_DEBUG 1

#ifdef MD_DEBUG
#define MD_ASR_DEBUG(A)	printk A
#else
#define MD_ASR_DEBUG(A)
#endif

struct md_asr_meta_info {
	struct asr_reservedblock asr_rb;
	struct asr_raidtable asr_rt;
	struct block_device *blkdev;
};

typedef enum {
	MD_ASR_MD_DIRTY		= 0x1,	/*
					 * Meta-data should be written
					 * at shutdown.  Used for things
					 * like checkpoints that we don't
					 * bother to commit to media on
					 * every change.
					 */
	MD_ASR_MD_NEED_SYNC	= 0x2,	/*
					 * Meta-data should be written
					 * immediately.
					 */
	MD_ASR_MD_SYNC_INPROG	= 0x4,	/*
					 * A Meta-data update request is
					 * being processed.
					 */
	MD_ASR_MD_SYNC_ALLOWED	= 0x8	/*
					 * We defer performing any meta-data
					 * sync operations until the array
					 * has been opened for write operation
					 * or we are degraded and start a
					 * rebuild operation or allocate a
					 * spare.  This avoids incrementing
					 * the sequence number and invalidating
					 * optimal members until our meta-data
					 * state changes.  This flag is the
					 * control for this behavior.
					 */
} md_asr_md_flags;

#define MD_ASR_WORKOP_SYNC 0x1

struct md_asr_array_info {
	spinlock_t				rt_lock;
	uint32_t				raidmagic;
	uint16_t				raidseq;
	uint32_t				sparedrivemagic;
	mdk_member_t				*spare;
	struct asr_raidtable			asr_rt;
	struct asr_raid_configline		*asr_rcl;
	md_asr_md_flags				flags;
	int					num_levels;
	int					toplevel;
	int					spare_configured;
	mddev_t					*spare_pool;
	mdk_workitem_t				*sync_md_workitem;
	struct md_asr_meta_info			*sync_minfo;
};

#define	mddev_to_array_info(mddev) \
	((struct md_asr_array_info*)mddev->member.metadata_private)

#define	get_rb(rdev) \
	((struct asr_reservedblock *)rdev->member.metadata_private)

static mdk_member_t spare_arrays = {
	.type		= MDK_MTYPE_MASTER,
	.subtype	= MDK_MSUBTYPE_NONE,
	.member_lock	= SPIN_LOCK_UNLOCKED,
	.child_set	= LIST_HEAD_INIT(spare_arrays.child_set),
	.refcount	= ATOMIC_INIT(1)
};

mdk_workitem_support_t	md_asr_syncer;

static int md_asr_work_control(mdk_member_t *member, int op, void *arg);
static int md_asr_sync_minfo(struct md_asr_meta_info *minfo);
static int md_asr_sync_handler(mddev_t *mddev, mdk_workitem_t *workitem);
static void md_asr_schedule_sync(struct md_asr_array_info *array_info,
				 int wait);
static mddev_t *md_asr_find_mddev(mdk_member_t *parent, uint32_t raidmagic);
static int md_asr_check_for_metadata(struct md_asr_meta_info *);
static int md_asr_dev_op_attach(dev_t dev);
static int md_asr_add_member(mddev_t *array, uint32_t raidmagic,
			     struct md_asr_meta_info *minfo);
static int md_asr_add_spare(mddev_t *array, uint32_t raidmagic,
			    struct md_asr_meta_info *minfo);
static int md_asr_find_toplevel(struct asr_raidtable *asr_rt);
static int md_asr_find_member(struct asr_raidtable *asr_rt, uint32_t magic,
			      int type, int *parent_index,
			      int *component_index);
static mdk_rdev_t *md_asr_find_spare_by_magic(uint32_t sparemagic);
static mddev_t *md_asr_mddev_alloc(mdk_member_t *parent,
				   struct asr_raidtable *asr_rt,
				   struct asr_raid_configline *asr_rcl,
				   mdk_msubtype_t subtype, int priority);
static mdk_rdev_t *md_asr_rdev_alloc(mddev_t *array,
				     struct asr_reservedblock *asr_rb,
				     struct asr_raid_configline *asr_rcl,
				     struct block_device *bdev,
				     mdk_msubtype_t subtype,
				     int priority);
static int md_asr_update_array_state(mddev_t *array);
static void md_asr_apply_rt(mddev_t *toplevel_array,
			  struct asr_raidtable *asr_rt);
static int md_asr_member_to_array_state(mddev_t *mddev, mdk_mstate_t md_state,
					int *adaptec_state);
static int md_asr_array_to_member_state(int adaptec_state,
					mdk_mstate_t *md_state);
static int md_asr_member_to_disk_state(mdk_mstate_t md_state,
				       int *adaptec_state);
static int md_asr_disk_to_member_state(mddev_t *mddev, int adaptec_state,
				       mdk_mstate_t *md_state);
static int md_asr_to_md_raidtype(uint8_t asr_raidtype);
static const char *md_asr_raidtype_string(int raidtype);
static void md_asr_dump_rb(struct asr_reservedblock *asr_rb);
static void md_asr_dump_rcl(struct asr_raid_configline *rcl, int index);
static void md_asr_dump_rt(struct asr_raidtable *asr_rt);
static void md_asr_dump_minfo( struct md_asr_meta_info *minfo);

static int md_asr_dev_op(dev_t, mdk_dev_op_t, void *);
static int md_asr_member_op(mdk_member_t *, int, void *);

static int md_asr_member_op_enter_safemode(mddev_t *mddev);
static int md_asr_member_op_exit_safemode(mddev_t *mddev);
static int md_asr_member_op_get_checkpoint(mddev_t *mddev,
					   mdk_checkpoint_arg_t *);
static int md_asr_member_op_set_checkpoint(mddev_t *mddev,
					   mdk_checkpoint_arg_t *);
static int md_asr_member_op_set_state(mdk_member_t *member,
				      mdk_set_state_arg_t *sstate);
static int md_asr_member_op_alloc_array_info(mdk_member_t *member, 
					mdu_array_info_t *array_k);


static mdk_metapersonality_t md_asr_pers = {
 	.name		= ADAPTEC_METADATA_NAME,
	.owner		= THIS_MODULE,
	.dev_op		= md_asr_dev_op,
	.member_op	= md_asr_member_op,
	.refcount	= ATOMIC_INIT(1)
};

static int
md_asr_minfo_alloc(dev_t dev, struct md_asr_meta_info **pminfo)
{
	struct md_asr_meta_info *minfo;
	struct block_device *blkdev;

	blkdev = md_get_blkdev(dev);
	if (blkdev == NULL)
		return (-ENODEV);

	minfo = kmalloc(sizeof(*minfo), GFP_KERNEL);
	if (minfo == NULL)
		return (-ENOMEM);
	memset(minfo, 0, sizeof(*minfo));
	minfo->blkdev = blkdev;
	*pminfo = minfo;
	return (0);
}

static void
md_asr_minfo_free(struct md_asr_meta_info *minfo)
{
	if (minfo->blkdev != NULL)
		md_release_blkdev(minfo->blkdev);
	kfree(minfo);
}

static int
md_asr_dev_op(dev_t dev, mdk_dev_op_t op, void *arg)
{
	MD_ASR_DEBUG(("md_asr_dev_op called\n"));
	switch (op) {
	case MDK_DEVOP_CLAIM:
	{
		struct md_asr_meta_info *minfo;
		int retval;

		MD_ASR_DEBUG(("md_asr_dev_op: MDK_DEVOP_CLAIM\n"));
		retval = md_asr_minfo_alloc(dev, &minfo);
		if (retval != 0)
			return (retval);
		retval = md_asr_check_for_metadata(minfo);
		if (retval == 0){
			*(int *)arg = MD_ADAPTEC_CLAIM_PRIORITY;
		} else {
			*(int *)arg = -1;
		}
		md_asr_minfo_free(minfo);
		return (retval);
	}
	case MDK_DEVOP_ATTACH:
		return (md_asr_dev_op_attach(dev));
	case MDK_DEVOP_WIPE_METADATA:
	{
		int retval;
		struct md_asr_meta_info *minfo;

		MD_ASR_DEBUG(("md_asr_dev_op: MDK_DEVOP_WIPE_METADATA\n"));
		retval = md_asr_minfo_alloc(dev, &minfo);
		if (retval != 0)
			return (retval);
		retval = md_asr_sync_minfo(minfo);
		md_asr_minfo_free(minfo);
		return retval;
	}
 	case MDK_DEVOP_GET_FREESIZE:
 	{
 		sector_t freesize;
		struct block_device *blkdev;
 
		blkdev = md_get_blkdev(dev);
		if (blkdev == NULL) {
 			MD_ASR_DEBUG(("md_asr_dev_op: cannot get blkdev\n"));
 			return (-ENODEV);
 		}
 		
 		/* last 17 sectors are reserved for Adaptec ASR metadata */
 		freesize = get_capacity(blkdev->bd_disk) - 17;
 		*((sector_t *)arg) = freesize;
		md_release_blkdev(blkdev);
 		return 0;
 	}
	case MDK_DEVOP_INITIALIZE:
	default:
		return (-EOPNOTSUPP);
	}
}

static int
md_asr_member_op(mdk_member_t *member, int op, void *arg)
{
	switch (op) {
	case MDK_MEMBER_OP_ENTER_SAFEMODE:
	case MDK_MEMBER_OP_EXIT_SAFEMODE:
	{
		mddev_t *mddev;

		/*
		 * We only care about safemode events on
		 * array objects.
		 */
		if (member->type != MDK_MTYPE_MDDEV)
			return (0);

		mddev = to_mddev(member);
		if (op == MDK_MEMBER_OP_ENTER_SAFEMODE)
			return md_asr_member_op_enter_safemode(mddev);
		else
			return md_asr_member_op_exit_safemode(mddev);
		/* NOTREACHED */
	}
	case MDK_MEMBER_OP_GET_CHECKPOINT:
	case MDK_MEMBER_OP_SET_CHECKPOINT:
	{
		mdk_checkpoint_arg_t *checkpoint;
		mddev_t *mddev;

		mddev = to_mddev(member);
		checkpoint = (mdk_checkpoint_arg_t *)arg;
		if (op == MDK_MEMBER_OP_GET_CHECKPOINT)
			return md_asr_member_op_get_checkpoint(mddev,
								checkpoint);
		else
			return md_asr_member_op_set_checkpoint(mddev,
								checkpoint);
	}
	case MDK_MEMBER_OP_SET_STATE:
		return md_asr_member_op_set_state(member,
						  (mdk_set_state_arg_t *)arg);
	case MDK_MEMBER_OP_ALLOC_ARRAY_INFO:
		return md_asr_member_op_alloc_array_info(member, arg);
	case MDK_MEMBER_OP_SET_ARRAY_ACTION:
	case MDK_MEMBER_OP_SET_USER_ACTION:
	case MDK_MEMBER_OP_GET_ARRAY_ACTION:
	case MDK_MEMBER_OP_GET_USER_ACTION:
	case MDK_MEMBER_OP_ALLOC_SPARE:
	default:
		return (-EOPNOTSUPP);
	}
}

static int
md_asr_work_control(mdk_member_t *member, int op, void *arg)
{
	mdk_workitem_t *work;
	mddev_t *mddev;
	struct md_asr_array_info *array_info;
	int error;

	work = (mdk_workitem_t *)arg;
	mddev = (mddev_t *)work->work_data;
	member = to_member(mddev);
	array_info = mddev_to_array_info(mddev);
	error = 0;

	switch (op) {
	case MDK_WORKITEM_OP_OPEN:
	{
		struct asr_raidtable *asr_rt;
		uint8_t *pbyte;
		int i;
		uint16_t elmcnt;
		uint16_t new_seq;
		uint16_t new_cksum;

		spin_lock_irq(&array_info->rt_lock);
		if ((array_info->flags & MD_ASR_MD_NEED_SYNC) == 0
		 || (array_info->flags & MD_ASR_MD_SYNC_ALLOWED) == 0) {
			spin_unlock_irq(&array_info->rt_lock);
			error = -EINVAL;
			break;
		}
		spin_unlock_irq(&array_info->rt_lock);

		array_info->sync_minfo =
		    kmalloc(sizeof(struct md_asr_meta_info), GFP_KERNEL);
		spin_lock_irq(&array_info->rt_lock);
		if (array_info->sync_minfo == NULL) {
			array_info->flags &= ~MD_ASR_MD_SYNC_INPROG;
			spin_unlock_irq(&array_info->rt_lock);
			printk("md_asr_syncer: Unable to allocate buffer\n");
			error = -ENOMEM;
			break;
		}
		new_seq = htons(ntohs(array_info->asr_rcl->raidseq) + 1);
		elmcnt = ntohs(array_info->asr_rt.elmcnt);
		for (i = 0; i < elmcnt; i++)
			array_info->asr_rt.ent[i].raidseq = new_seq;
		array_info->flags &= ~(MD_ASR_MD_DIRTY|MD_ASR_MD_NEED_SYNC);
		asr_rt = &array_info->sync_minfo->asr_rt;
		elmcnt = ntohs(array_info->asr_rt.elmcnt);
		memcpy(asr_rt, &array_info->asr_rt, sizeof(*asr_rt));
		spin_unlock_irq(&array_info->rt_lock);

		/*
		 * Update checksum.  It only covers the configured elements.
		 */
		pbyte = (uint8_t *)asr_rt->ent;
		new_cksum = 0;
		for (i = 0; i < (sizeof(asr_rt->ent[0]) * elmcnt); i++)
			new_cksum += pbyte[i];
		/* 16-bit checksum */
		new_cksum &= 0xffff;
		asr_rt->rchksum = ntohs(new_cksum);
		break;
	}
	case MDK_WORKITEM_OP_CLOSE:
	{
		/* NO-OP */
		break;
	}
	case MDK_WORKITEM_OP_FREE:
	{
		int requeue;

		kfree(array_info->sync_minfo);
		array_info->sync_minfo = NULL;
		spin_lock_irq(&array_info->rt_lock);
		if ((array_info->flags & MD_ASR_MD_NEED_SYNC) != 0
		 && (array_info->flags & MD_ASR_MD_SYNC_ALLOWED) != 0) {
			requeue = 1;
		} else {
			array_info->flags &= ~MD_ASR_MD_SYNC_INPROG;
			requeue = 0;
		}
		spin_unlock_irq(&array_info->rt_lock);
		if (requeue)
			error = md_queue_workitem(&md_asr_syncer,
						  work, /*head_of_queue*/0);
		break;
	}
	default:
		error = -EINVAL;
	}
	return (error);
}

static int
md_asr_sync_minfo(struct md_asr_meta_info *minfo)
{
	sector_t rb_location;
	int retval;

	rb_location = get_capacity(minfo->blkdev->bd_disk) - 1;
	retval = md_write_blocks(minfo->blkdev, rb_location,
				 ADAPTEC_ASR_RB_SIZE / 512,
				 (uint8_t *)&minfo->asr_rb);
	if (retval != 0)
		return (retval);

	retval = md_write_blocks(minfo->blkdev, ntohl(minfo->asr_rb.raidtbl),
			       	 ADAPTEC_ASR_RT_SIZE / 512,
				 (uint8_t *)&minfo->asr_rt);
	return (retval);
}

static int
md_asr_sync_handler(mddev_t *mddev, mdk_workitem_t *workitem)
{
	mdk_member_t *array;
	struct md_asr_array_info *array_info;
	struct md_asr_meta_info *minfo;
	mdk_member_t *child;
	mdk_member_t *gchild;
	struct list_head *tmp;
	struct list_head *tmp2;
	u_long flags;
	u_long flags2;

	mddev = (mddev_t *)workitem->work_data;
	array_info = mddev_to_array_info(mddev);
	minfo = array_info->sync_minfo;
	array = to_member(mddev);

	ITERATE_CHILDREN(array, child, tmp, flags) {
		mdk_rdev_t *rdev;

		if (child->type == MDK_MTYPE_RDEV) {

			rdev = to_rdev(child);
			minfo->blkdev = rdev->bdev;
			memcpy(&minfo->asr_rb, get_rb(rdev),
			       sizeof(minfo->asr_rb));
			md_asr_sync_minfo(minfo);
			continue;
		}

		ITERATE_CHILDREN(child, gchild, tmp2, flags2) {

			if (gchild->type != MDK_MTYPE_RDEV)
				BUG();

			rdev = to_rdev(gchild);
			minfo->blkdev = rdev->bdev;
			memcpy(&minfo->asr_rb, get_rb(rdev),
			       sizeof(minfo->asr_rb));
			md_asr_sync_minfo(minfo);
		}
	}
	return (0);
}

static void
md_asr_schedule_sync(struct md_asr_array_info *array_info, int wait)
{
	u_long flags;
	int sched_work;

	sched_work = 0;
	spin_lock_irqsave(&array_info->rt_lock, flags);
	array_info->flags |= MD_ASR_MD_DIRTY|MD_ASR_MD_NEED_SYNC;
	if ((array_info->flags & MD_ASR_MD_SYNC_INPROG) == 0) {
		array_info->flags |= MD_ASR_MD_SYNC_INPROG;
		sched_work = 1;
	}
	spin_unlock_irqrestore(&array_info->rt_lock, flags);
	if (sched_work) {
		md_queue_workitem(&md_asr_syncer,
				  array_info->sync_md_workitem,
				  /*head_of_queue*/0);
	}
	if (!wait)
		return;

	for (;;) {
		int syncing;

		md_wait_workitem(&md_asr_syncer, array_info->sync_md_workitem);
		spin_lock_irqsave(&array_info->rt_lock, flags);
		syncing = (array_info->flags & MD_ASR_MD_SYNC_INPROG);
		spin_unlock_irqrestore(&array_info->rt_lock, flags);
		if (!syncing)
			break;
	}
}

static mddev_t *
md_asr_find_mddev(mdk_member_t *parent, uint32_t raidmagic)
{
	mddev_t *mddev;
	mdk_member_t *child;
	struct list_head *tmp;
	u_long flags;

	mddev = NULL;
	ITERATE_CHILDREN(parent, child, tmp, flags) {

		if (child->metapers != &md_asr_pers) 
			continue;

		mddev = to_mddev(child);
		if (mddev_to_array_info(mddev)->raidmagic == raidmagic)
			return (mddev);
	}
	return (NULL);
}

/*
 * Read the reserved block and the RAID table off of the disk and check
 * both to make sure the metadata is okay.  Return 1 for success, 0 for
 * failure.
 */
static int
md_asr_check_for_metadata(struct md_asr_meta_info *minfo)
{
#if MD_DEBUG
	char b[BDEVNAME_SIZE];
#endif
	struct block_device *blkdev;
	struct asr_reservedblock *asr_rb;
	struct asr_raidtable *asr_rt;
	unsigned int rt_cksum;
	uint8_t *rclptr;
	sector_t rb_location;
	u_int elmcnt;
	int retval, i;

	retval = 0;
	blkdev = minfo->blkdev;

	MD_ASR_DEBUG(("md_check_for_sb_adaptec called for %s %d/%d\n",
		      bdevname(blkdev, b), MAJOR(blkdev->bd_dev),
		      MINOR(blkdev->bd_dev)));

	asr_rb = &minfo->asr_rb;
	asr_rt = &minfo->asr_rt;

	/*
	 * The reserved block, if it is around, will be on the last sector
	 * of the disk.
	 */
	rb_location = get_capacity(blkdev->bd_disk) - 1;

	retval = md_read_blocks(blkdev, rb_location, ADAPTEC_ASR_RB_SIZE / 512,
				(uint8_t *)asr_rb);
	if (retval != 0) {		   
		/* print an error here? */
		MD_ASR_DEBUG(("md_asr_check_for_metadata: "
			      "md_read_blocks failed %d\n", retval));
		return (retval);
	}

	if ((ntohl(asr_rb->b0idcode) != B0RESRVD)
	 || (ntohl(asr_rb->smagic) != SVALID)) {
		MD_ASR_DEBUG(("md_asr_check_for_metadata: bailing out, "
			      "b0idcode= %d, smagic= %d\n",
			      asr_rb->b0idcode, asr_rb->smagic));
		return (-ENODEV);
	}

	retval = md_read_blocks(blkdev, ntohl(asr_rb->raidtbl),
			       ADAPTEC_ASR_RT_SIZE / 512, (uint8_t *)asr_rt);
	if (retval != 0) {
		/* print error? */
		MD_ASR_DEBUG(("md_check_for_sb_adaptec: "
			      "2nd md_read_blocks failed\n"));
		return (retval);
	}

	if (ntohl(asr_rt->ridcode) != RVALID2) {
		MD_ASR_DEBUG(("md_asr_check_for_metadata: bailing out, "
			      "ridcode= %d\n", asr_rt->ridcode));
		return (-ENODEV);
	}

	rclptr = (uint8_t *)asr_rt->ent;
	rt_cksum = 0;
	elmcnt = ntohs(asr_rt->elmcnt);

	if (elmcnt > RCTBL_MAX_ENTRIES) {
		MD_ASR_DEBUG(("md_asr_check_for_metadata: RAID table "
			      "invalid entry count %d\n", elmcnt));
		return (-ENODEV);
	}

	/*
	 * The checksum isn't calculated over the entire RAID table, only
	 * over the filled-in config line elements.
	 */
	for (i = 0; i < (sizeof(asr_rt->ent[0]) * elmcnt); i++)
		rt_cksum += rclptr[i];
	/* 16-bit checksum */
	rt_cksum &= 0xffff;

	if (ntohs(asr_rt->rchksum) != rt_cksum) {
		MD_ASR_DEBUG(("md_asr_check_for_metadata: RAID table "
			      "checksum %#x doesn't match calculated "
			      "checksum %#x\n",
			      ntohs(asr_rt->rchksum), rt_cksum));
		return (-ENODEV);
	}

	return (0);
}

/*
 * Add a disk that we have successfully bid for to our
 * array configuration.
 */
static int
md_asr_dev_op_attach(dev_t dev)
{
	struct md_asr_meta_info *minfo;
	struct md_asr_array_info *array_info;
	struct asr_raidtable *asr_rt;
	struct asr_raid_configline *toplevel_rcl;
	mddev_t *array;
	mdk_member_t *parent;
	uint32_t raidmagic;
	uint16_t raidseq;
	int toplevel;
	int retval;

	MD_ASR_DEBUG(("md_add_to_array_adaptec called\n"));

	retval = md_asr_minfo_alloc(dev, &minfo);
	if (retval != 0)
		return (retval);
	retval = md_asr_check_for_metadata(minfo);
	if (retval != 0) {
		md_asr_minfo_free(minfo);
		return (retval);
	}

	md_asr_dump_minfo(minfo);

	asr_rt = &minfo->asr_rt;

	/*
	 * Determine the toplevel array to which this device,
	 * or the logical array parent of this device, belongs.
	 */
	toplevel = md_asr_find_toplevel(asr_rt);
	toplevel_rcl = &asr_rt->ent[toplevel];
	raidmagic = ntohl(toplevel_rcl->raidmagic);
	raidseq = ntohs(toplevel_rcl->raidseq);

	/*
	 * Spare handling differs between HostRAID and processor
	 * RAID controller generated meta-data.
	 * 
	 * HostRAID only supports dedicated spares.  The RAID array members
	 * indicate their dedicated spare via a non-zero sparedrivemagic 
	 * in the raidtable of all of their members.  The dedicated spare
	 * also points back to its associated raid array by setting its
	 * raidmagic field in the raidtable to the magic of the raid array.
	 *
	 * Processor based cards do not have the concept of dedicated spares.
	 * The controller instead combines all drives into a single spare
	 * pool array.  In this configuration, the sparedrivemagic field and
	 * the raidmagic field in the raidtable will always be 0.
	 *
	 * In both cases, the spare drives are "pooled" in that a logical
	 * spare array object is listed in the metadata as the parent of
	 * the spare.  To simplify the handling of both cases, even dedicated
	 * spares are made part of a "pool" with the parent array object
	 * pointing to that pool.
	 * 
	 * RAID arrays are parented by the md core's toplevel_array
	 * object.  Spare pools are parented by the spare_arrays object.
	 * Modify the parent used to search for an existing array accordingly.
	 */
	if (toplevel_rcl->raidtype == ASR_RAIDSPR)
		parent = &spare_arrays;
	else
		parent = &toplevel_arrays;
	array = md_asr_find_mddev(parent, raidmagic);
	if (array == NULL) {
		mdk_msubtype_t subtype;
		int priority;

		/*
		 * This is the first member to arrive for an array.
		 * Create an array object to house the new array
		 * configuration.
		 */
		if (toplevel_rcl->raidtype == ASR_RAIDSPR) {
			priority = MDK_PRIORITY_UNKNOWN;
			subtype = MDK_MSUBTYPE_POOL_SPARE;
			if (asr_rt->raidmagic != 0
			 || asr_rt->raidmagic != 0xFFFFFFFFUL)
				subtype = MDK_MSUBTYPE_DEDICATED_SPARE;
		} else {
			subtype = MDK_MSUBTYPE_TOPLEVEL;
			priority = toplevel_rcl->raidhba << 24
				 | toplevel_rcl->raidchnl << 16
				 | ntohl(toplevel_rcl->raidid) << 8
				 | toplevel_rcl->raidlun;
		}
		array = md_asr_mddev_alloc(parent, asr_rt, toplevel_rcl,
					   subtype, priority);
		if (array == NULL) {
			md_asr_minfo_free(minfo);
			return (-ENOMEM);
		}

		array_info = mddev_to_array_info(array);

		/*
		 * Hold an extra ref-count on the mddev so the
		 * refcounting scheme is the code can treat it
		 * the same as a pre-exisiting array.
		 */
		member_get(to_member(array));
	} else {
		array_info = mddev_to_array_info(array);

		if (ASR_SEQ_GT(raidseq, array_info->raidseq)
			&& toplevel_rcl->raidtype != ASR_RAIDSPR) {
			/* 
			 * If the newly arrived drive has a newer sequence
			 * number than that reported in an existing array, vet
			 * the already attached members against the newer
			 * meta-data.  Members that no longer belong in the
			 * array will be removed accordingly.
			 */
			md_asr_apply_rt(array, asr_rt);

		} else if (ASR_SEQ_LT(raidseq, array_info->raidseq)
			&& toplevel_rcl->raidtype != ASR_RAIDSPR) {
			uint32_t drivemagic;
			int index;

			/* 
			 * If this member has an older sequencer number,
			 * only allow it into the array if it still exists
			 * there.  Otherwise, we fail the attach so this
			 * drive can be accessed as a "single".
			 */
			drivemagic = ntohl(minfo->asr_rb.drivemagic);
			index = md_asr_find_member(&array_info->asr_rt,
						   drivemagic, FWP,
						   NULL, NULL);
			if (index < 0) {
				md_asr_minfo_free(minfo);
				member_put(&array->member);
				return (-EINVAL);
			}

			/* XXX mark as a rebuild target??? */
		}
	}

	/*
	 * The "add" methods take over ownership of all
	 * objects and will free them if necessary on
	 * error.
	 */
	if (toplevel_rcl->raidtype == ASR_RAIDSPR)
		retval = md_asr_add_spare(array, raidmagic, minfo);
	else
		retval = md_asr_add_member(array, raidmagic, minfo);

	member_put(&array->member);

	return (retval);
}

/*
 * Handle notification of our array transitioning into
 * Read/Write mode or, in MD parlance, "exiting safemode".
 */
static int
md_asr_member_op_exit_safemode(mddev_t *mddev)
{
	struct md_asr_array_info *array_info;
	u_long flags;
	int sched_write;

	/*
	 * Update the sequence number and mark the
	 * array meta-data as open.  We use this
	 * status to detect improper shutdown of
	 * the array.  Incrementing the sequence number
	 * automaticallly prevents "missing" members
	 * that arrive late from being accepted into
	 * the array.
	 */
	MD_ASR_DEBUG(("ASR Exit Safemode\n"));
	array_info = mddev_to_array_info(mddev);
	sched_write = 0;
	spin_lock_irqsave(&array_info->rt_lock, flags);
	if (array_info->asr_rt.dirty != ARRAY_STATE_DIRTY) {
		array_info->asr_rt.dirty = ARRAY_STATE_DIRTY;
		sched_write = 1;
		array_info->flags |= MD_ASR_MD_SYNC_ALLOWED|MD_ASR_MD_NEED_SYNC;
	}
	spin_unlock_irqrestore(&array_info->rt_lock, flags);
	if (sched_write)
		md_asr_schedule_sync(array_info, /*wait4complete*/1);
	return (0);
}

/*
 * Handle notification of our array transitioning to Read-Only
 * mode or, in MD parlance, "entering safemode".
 */
static int
md_asr_member_op_enter_safemode(mddev_t *mddev)
{
	struct md_asr_array_info *array_info;
	u_long flags;
	int sched_write;

	/*
	 * Update the sequence number and mark the
	 * array meta-data as closed.  We increment
	 * the sequence number so that we are not
	 * confused by any members that for some
	 * reason are not written before we lose
	 * power or are incorrectly shutdown.
	 */
	MD_ASR_DEBUG(("ASR Enter Safemode\n"));
	array_info = mddev_to_array_info(mddev);
	sched_write = 0;
	spin_lock_irqsave(&array_info->rt_lock, flags);
	if (array_info->asr_rt.dirty == ARRAY_STATE_DIRTY
	 && array_info->asr_rcl->curAppBlock == 0) {
		array_info->asr_rt.dirty = ARRAY_STATE_OK;
		array_info->flags |= MD_ASR_MD_SYNC_ALLOWED|MD_ASR_MD_NEED_SYNC;
		sched_write = 1;
	}
	spin_unlock_irqrestore(&array_info->rt_lock, flags);
	if (sched_write)
		md_asr_schedule_sync(array_info, /*wait*/1);
	return (0);
}

static int
md_asr_member_op_get_checkpoint(mddev_t *mddev, mdk_checkpoint_arg_t *chkpt_arg)
{
	struct md_asr_array_info *array_info;
	mdk_member_t *array;
	u_long flags;

	array = to_member(mddev);
	array_info = mddev_to_array_info(mddev);
	spin_lock_irqsave(&array->member_lock, flags);
	chkpt_arg->checkpoint = ntohl(array_info->asr_rcl->curAppBlock);
	spin_unlock_irqrestore(&array->member_lock, flags);
	return (0);
}

static int
md_asr_member_op_set_checkpoint(mddev_t *mddev, mdk_checkpoint_arg_t *chkpt_arg)
{
	struct md_asr_array_info *array_info;
	mdk_member_t *array;
	int build_complete;
	int sync_md;

	array = to_member(mddev);
	array_info = mddev_to_array_info(mddev);
	sync_md = 0;
	spin_lock_irq(&array_info->rt_lock);
	if (array_info->asr_rt.dirty != ARRAY_STATE_DIRTY) {
		array_info->asr_rt.dirty = ARRAY_STATE_DIRTY;
		array_info->flags |= MD_ASR_MD_SYNC_ALLOWED;
		sync_md = 1;
	}
	if (chkpt_arg->checkpoint < array->data_size) {
		array_info->asr_rcl->curAppBlock = htonl(chkpt_arg->checkpoint);
		build_complete = 0;
	} else {
		array_info->asr_rcl->curAppBlock = 0;
		build_complete = 1;
		sync_md = 1;
	}
	array_info->flags |= MD_ASR_MD_DIRTY;
	spin_unlock_irq(&array_info->rt_lock);

	if (sync_md)
		md_asr_schedule_sync(array_info, /*wait*/0);

	if (build_complete) {
		mdk_member_t *child;
		struct list_head *tmp;
		u_long cflags;

		/*
		 * Mark any build target children as optimal.
		 * Do this without holding any locks in case
		 * these update result in updates in our own
		 * state.
		 */
		ITERATE_CHILDREN(array, child, tmp, cflags) {
			u_long cflags;

			spin_lock_irqsave(&child->member_lock, cflags);
			if (MDK_WRITABLE_MEMBER(child)
			 && (child->state & MDK_MSTATE_REBUILD_TARGET) != 0) {
				mdk_set_state_arg_t sstate;
				mdk_metapersonality_t *mpers;

				sstate.oldstate = child->state;
				spin_unlock_irqrestore(&child->member_lock,
						       cflags);
				sstate.newstate = child->state
						& ~MDK_MSTATE_REBUILD_TARGET;
				mpers = child->metapers;
				mpers->member_op(child,
						 MDK_MEMBER_OP_SET_STATE,
						 &sstate);
			} else {
				spin_unlock_irqrestore(&child->member_lock,
						       cflags);
			}
		}
	}
	return (0);
}

static int
md_asr_member_op_set_state(mdk_member_t *member, mdk_set_state_arg_t *sstate)
{
	mdk_member_t *toplevel;
	struct md_asr_array_info *array_info;
	spinlock_t *rt_lock;
	u_long rtflags;
	u_long mflags;
	uint32_t raidmagic;
	int asr_state;
	int error;
	int i;
	int sync_md;
	int put_member;

	toplevel = member;
	while (toplevel->subtype != MDK_MSUBTYPE_TOPLEVEL)
		toplevel = toplevel->parent;
	array_info = mddev_to_array_info(to_mddev(toplevel));
	rt_lock = &array_info->rt_lock;

	put_member = 0;
	switch (member->type) {
	case MDK_MTYPE_RDEV:
		raidmagic = get_rb(to_rdev(member))->drivemagic;
		error = md_asr_member_to_disk_state(sstate->newstate,
						    &asr_state);
		/*
		 * When RDEVs fail, schedule them for desctruction
		 * by dropping their sentinal reference.
		 */
		if (sstate->newstate == MDK_MSTATE_FAILED
		 && sstate->oldstate != MDK_MSTATE_FAILED)
			put_member = 1;
		break;
	case MDK_MTYPE_MDDEV:
	{
		mddev_t *mddev;

		mddev = to_mddev(member);
		raidmagic = mddev_to_array_info(mddev)->asr_rcl->raidmagic;
		error = md_asr_member_to_array_state(mddev, sstate->newstate,
						     &asr_state);
		break;
	}
	default:
		BUG();
		return (-EINVAL);
	}

	if (error != 0) {
		printk("member state %#x can't be mapped to ASR MD!\n",
		       sstate->newstate);
		return (-EINVAL);
	}

	spin_lock_irqsave(rt_lock, rtflags);
	spin_lock_irqsave(&member->member_lock, mflags);
	if (sstate->oldstate != member->state) {
		spin_unlock_irqrestore(&member->member_lock, mflags);
		spin_unlock_irqrestore(rt_lock, rtflags);
		return (-EINVAL);
	}
	member->state = sstate->newstate;
	for (i = 0; i < ntohs(array_info->asr_rt.elmcnt); i++) {
		if (array_info->asr_rt.ent[i].raidmagic != raidmagic)
			continue;
		array_info->asr_rt.ent[i].raidstate = asr_state;
	}
	array_info->flags |= MD_ASR_MD_NEED_SYNC|MD_ASR_MD_DIRTY;
	spin_unlock_irqrestore(&member->member_lock, mflags);
	spin_unlock_irqrestore(rt_lock, rtflags);
	sync_md = 1;
	if (member->subtype == MDK_MSUBTYPE_TOPLEVEL
	 || md_asr_update_array_state(to_mddev(member->parent)) == 0)
		md_asr_schedule_sync(array_info, /*wait*/0);

	if (put_member)
		member_put(member);
	return (0);
}

static int
md_asr_member_op_alloc_array_info(mdk_member_t *member,
				  mdu_array_info_t *array_info)
{
	struct	md_asr_meta_info *minfo;
	struct	asr_reservedblock *asr_rb;
	struct	asr_raidtable *asr_rt;
	struct	asr_raid_configline *toplevel_rcl;
	mddev_t	*array;
	int retval;
	int toplevel;
	uint32_t raidmagic;
	struct timeval md_timestamp;


	retval = md_asr_minfo_alloc(0, &minfo);
	
	if (retval != 0)
		return (retval);
	
	asr_rb = &minfo->asr_rb;
	asr_rt = &minfo->asr_rt;

	/* dump info about a phoney array */
	asr_rt->ent[0].raidlevel = FWL;
	asr_rt->sparedrivemagic = 0;
	toplevel = md_asr_find_toplevel(asr_rt);
	toplevel_rcl = &asr_rt->ent[toplevel];
	toplevel_rcl->raidtype = array_info->level;
	
	get_random_bytes(&raidmagic, sizeof(raidmagic));
	toplevel_rcl->raidmagic = htonl(raidmagic);

	do_gettimeofday(&md_timestamp);
	toplevel_rcl->raidseq = htons(md_timestamp.tv_sec);
	
	(toplevel_rcl+1)->strpsize = array_info->chunk_size >> 9;
	toplevel_rcl->lcapcty = 0;
	toplevel_rcl->raidcnt = 0;
	toplevel_rcl->raidstate = LS_DEGRADED_M;
	
	array = md_asr_mddev_alloc(&toplevel_arrays, asr_rt,
				   toplevel_rcl,
				   MDK_MSUBTYPE_TOPLEVEL,
				   MDK_PRIORITY_UNKNOWN);
	if (array == NULL) {
		md_asr_minfo_free(minfo);
		retval = -1;
		return retval;
	}

	return retval;
}

/*
 * Add a device, indicated by dev_t, with the metadata in minfo,
 * to the toplevel array tree rooted by "array".
 */
static int
md_asr_add_member(mddev_t *array, uint32_t raidmagic,
		  struct md_asr_meta_info *minfo)
{
	struct md_asr_array_info *array_info;
	struct asr_reservedblock *asr_rb;
	struct asr_raidtable *asr_rt;
	struct asr_raid_configline *asr_rcl;
	mdk_rdev_t *rdev;
	int rt_index;
	int parent_index;
	int component_index;

	array_info = mddev_to_array_info(array);
	asr_rb = &minfo->asr_rb;
	asr_rt = &minfo->asr_rt;

	/*
	 * Find this member in the raid table.
	 */
	rt_index = md_asr_find_member(asr_rt, ntohl(asr_rb->drivemagic),
				      FWP, &parent_index, &component_index);
	MD_ASR_DEBUG(("md_asr_add_member called: rti %d, pi %d, ci %d\n",
		      rt_index, parent_index, component_index));
	if (rt_index < 0
	 || parent_index < 0
	 || component_index < 0) {
		MD_ASR_DEBUG(("Invalid Raid Table\n"));
		kfree(minfo);
		return (-EINVAL);
	}
	asr_rcl = &asr_rt->ent[rt_index];

	/*
	 * Array is the top-level array.  If it heads
	 * a multi-level array, find or allocate our
	 * direct parent.
	 */
	if (array_info->num_levels == 2) {
		mddev_t *child;
		struct asr_raid_configline *parent_rcl;

		MD_ASR_DEBUG(("Member of Dual Level Array\n"));
		parent_rcl = &asr_rt->ent[parent_index];
		child = md_asr_find_mddev(&array->member,
					  ntohl(parent_rcl->raidmagic));
		if (child == NULL) {
			int parent_component_index;

			/*
			 * Determine parents component index
			 * in the array.
			 */

			md_asr_find_member(asr_rt, ntohl(parent_rcl->raidmagic),
					   FWL, NULL, &parent_component_index);
			if (parent_component_index < 0) {
				MD_ASR_DEBUG(("Invalid Raid Table "
					      "Parent Offset\n"));
				kfree(minfo);
				return (-EINVAL);
			}
			array = md_asr_mddev_alloc(to_member(array), asr_rt,
						   parent_rcl,
						   MDK_MSUBTYPE_CHILD,
						   parent_component_index);
			if (array == NULL) {
				kfree(minfo);
				return (-ENOMEM);
			}
			member_get(to_member(array));
		}
	} else {
		member_get(to_member(array));
	}

	rdev = md_asr_rdev_alloc(array, asr_rb, asr_rcl,
				 minfo->blkdev, MDK_MSUBTYPE_CHILD,
				 component_index);
	member_put(to_member(array));
	md_asr_minfo_free(minfo);
	if (rdev == NULL)
		return (-ENOMEM);
	return (0);
}

/*
 * Add a spare to the spare pool array "array".
 */
static int
md_asr_add_spare(mddev_t *array, uint32_t raidmagic,
		 struct md_asr_meta_info *minfo)
{
	mdk_msubtype_t subtype;
	mdk_rdev_t *rdev;
	int rt_index;

	subtype = MDK_MSUBTYPE_POOL_SPARE;
	if (minfo->asr_rt.raidmagic != 0)
		subtype = MDK_MSUBTYPE_DEDICATED_SPARE;
	rt_index = md_asr_find_member(&minfo->asr_rt,
				      ntohl(minfo->asr_rb.drivemagic),
				      FWP, NULL, NULL);
	rdev = md_asr_rdev_alloc(array, &minfo->asr_rb,
				 &minfo->asr_rt.ent[rt_index],
				 minfo->blkdev, subtype, /*priority*/0);
	md_asr_minfo_free(minfo);
	if (rdev == NULL)
		return (-ENOMEM);
	return (0);
}

/*
 * Find the index in the ASR RaidTable of the top
 * level array to which this member belongs.
 */
static int
md_asr_find_toplevel(struct asr_raidtable *asr_rt)
{
	int i, toplevel, physlevel, found;

	found = 0;
	toplevel = 0;
	physlevel = 0;

	for (i = 0; i < ntohs(asr_rt->elmcnt); i++) {
		if (asr_rt->ent[i].raidlevel == FWL_2) {
			toplevel = i;
			found = 1;
			break;
		} else if (asr_rt->ent[i].raidlevel == FWL) {
			physlevel = i;
		}
	}

	if (found == 0)
		toplevel = physlevel;

	MD_ASR_DEBUG(("md_asr_find_toplevel returning %d\n", toplevel));
	return (toplevel);
}

/*
 * Find a member of type "type" with magic "magic" in
 * the passed in raid-table.  Returns raid table index
 * on success, -1 on search failure.  Optionally returns
 * both the raid table index of the parent and the component
 * index relative to the parent of the found entry (e.g.
 * component 0 or 1 of the array parented by the entry at
 * parent index).  A top-level array with have a parent
 * and component index of -1.
 */
static int
md_asr_find_member(struct asr_raidtable *asr_rt, uint32_t magic, int type,
		   int *p_parent_index, int *p_component_index)
{
	int i;
	int member_index;
	int parent_index;
	int component_index;
	int toplevel;

	member_index = -1;
	parent_index = -1;
	component_index = -1;
	for (i = 0; i < ntohs(asr_rt->elmcnt); i++) {

		if (asr_rt->ent[i].raidlevel == type
		 && ntohl(asr_rt->ent[i].raidmagic) == magic) {
			member_index = i;
			break;
		}

		if (asr_rt->ent[i].raidlevel != FWP) {
			/*
			 * Anytime we encounter an array object,
			 * reset our array object relative component
			 * index.  We also set the parent index so
			 * that this new array is treated as the parent
			 * if we find the member we are looking for
			 * before seeing another array.
			 */
			parent_index = i;
			component_index = 0;
		} else
			component_index++;
	}

	toplevel = md_asr_find_toplevel(asr_rt);
	if (member_index == toplevel) {
		/*
		 * Toplevel Arrays have no parent or component index,
		 * but our initial loop may have set while passing
		 * through child arrays.  Clear them again now.
		 */
		parent_index = -1;
		component_index = -1;
	} else if (type == FWL
		&& member_index != -1) {
		int i;

		/*
		 * For child arrays of a toplevel array, we must
		 * determine the component index by finding the
		 * second FWL instance for this array under the
		 * top-level array entry.
		 */
		parent_index = toplevel;
		component_index = 0;
		for (i = toplevel + 1; i < ntohs(asr_rt->elmcnt); i++) {
			if (asr_rt->ent[i].raidlevel == FWL
			 && ntohl(asr_rt->ent[i].raidmagic) == magic)
				break;
			component_index++;
		}
	}

	if (p_parent_index != NULL)
		*p_parent_index = parent_index;

	if (p_component_index != NULL)
		*p_component_index = component_index;

	return (member_index);
}

static mdk_rdev_t *
md_asr_find_spare_by_magic(uint32_t sparemagic)
{
	mdk_member_t *spare_array;
	struct list_head *tmp;
	u_long aflags;

	ITERATE_CHILDREN(&spare_arrays, spare_array, tmp, aflags) {
		mdk_member_t *child;
		struct list_head *tmp2;
		u_long sflags;

		ITERATE_CHILDREN(spare_array, child, tmp2, sflags) {
			struct asr_reservedblock *rb;
			mdk_rdev_t *spare;

			spare = to_rdev(child);
			rb = get_rb(spare);
			if (sparemagic == ntohl(rb->drivemagic)) {
				/*
				 * Drop extra refcount on parent
				 * array object.  The child already
				 * holds a refcount on it.
				 */
				member_put(spare_array);
				return (spare);
			}
				
		}
	}
	return (NULL);
}

static mddev_t *
md_asr_mddev_alloc(mdk_member_t *parent, struct asr_raidtable *asr_rt,
		   struct asr_raid_configline *asr_rcl,
		   mdk_msubtype_t subtype, int priority)
{
	mddev_t *array;
	mdk_member_t *member;
	struct md_asr_array_info *array_info;
	int toplevel;

	array = md_mddev_alloc(NULL, subtype, priority);
	if (array == NULL)
		return (NULL);

	member = to_member(array);
	array_info = kmalloc(sizeof(*array_info), GFP_KERNEL);
	if (array_info == NULL) {
		member_put(member);
		return (NULL);
	}

	memset(array_info, 0, sizeof(*array_info));
	spin_lock_init(&array_info->rt_lock);

	/*
	 * We allocate a workitem in all top-level
	 * arrays to sync out metadata.
	 */
	if (subtype == MDK_MSUBTYPE_TOPLEVEL
	 && asr_rcl->raidtype != ASR_RAIDSPR) {
		array_info->sync_md_workitem =
		    kmalloc(sizeof(mdk_workitem_t), GFP_KERNEL);
		if (array_info == NULL) {
			member_put(member);
			kfree(array_info);
			return (NULL);
		}
		md_init_workitem(array_info->sync_md_workitem,
				 MD_ASR_WORKOP_SYNC, md_asr_work_control,
				 md_asr_sync_handler, array);
	}

	toplevel = md_asr_find_toplevel(asr_rt);
	array_info->raidmagic = ntohl(asr_rcl->raidmagic);
	array_info->raidseq = ntohs(asr_rt->ent[toplevel].raidseq);
	array_info->toplevel = toplevel;
	if (asr_rcl->raidlevel == FWL_2)
		array_info->num_levels = 2;
	else
		array_info->num_levels = 1;

	MD_ASR_DEBUG(("md_asr_mddev_alloc: %p Level %d priority %#x, "
		      "subtype %d\n", array, array_info->num_levels,
		      priority, subtype));

	memcpy(&array_info->asr_rt, asr_rt, sizeof(*asr_rt));
	array_info->asr_rcl = &array_info->asr_rt.ent[asr_rcl - asr_rt->ent];

	/*
	 * Chunksize is the portion of the stripe on each child, so
	 * use the stripe size on our first child to export this
	 * information to MD.
	 */
	array->chunk_size = ntohs((asr_rcl+1)->strpsize) << 9;
	array->preferred_unit = -1;
	member->metadata_private = array_info;
	member->metapers = &md_asr_pers;
	member->state = MDK_MSTATE_OPTIMAL;
	if (asr_rcl->raidtype != ASR_RAIDSPR) {
		int error;

		/*
		 * An array without any members must be in the
		 * failed state unless it is a spare array or
		 * it is marked for initialization. This will be
		 * updated as children are added.
		 */
		error = md_asr_array_to_member_state(asr_rcl->raidstate,
						     &member->state);
		member->state &= ~MDK_MSTATE_OPTIMAL;
		member->state |= MDK_MSTATE_FAILED;

		member->total_size = ntohl(asr_rcl->lcapcty);
		member->data_size = member->total_size;
/* XXX Verify component count for multi-level arrays. */
		array->max_disks = ntohs(asr_rcl->raidcnt);
		array->level = asr_rcl->raidtype;

		/*
		 * If we are a toplevel array and our array info indicates
		 * a dedicated spare pool, try to find it now in case the
		 * spare pool was enumerated before this array.
		 */
		if (subtype == MDK_MSUBTYPE_TOPLEVEL
		 && asr_rt->sparedrivemagic != 0) {
			mdk_rdev_t *spare;
			uint32_t sparemagic;

			sparemagic = ntohl(asr_rt->sparedrivemagic);
			spare = md_asr_find_spare_by_magic(sparemagic);
			if (spare != NULL) {
				array_info->spare_pool =
				    to_mddev(spare->member.parent);
				member_get(to_member(array_info->spare_pool));
				/*
				 * Drop refcount from the lookup of
				 * spare.  We will lookup a valid
				 * spare from the pool when one is
				 * needed.
				 */
				member_put(to_member(spare));
				MD_ASR_DEBUG(("attached spare %#x to array "
					      "%#x\n", sparemagic,
					      array_info->raidmagic));
			}
		}
		member->pers = md_find_pers(array->level);
	} else if (subtype == MDK_MSUBTYPE_DEDICATED_SPARE) {
		mddev_t *assoc_array;
		
		/*
		 * Associate this dedicated spare pool with
		 * the array it is dedicated to if it has
		 * already been discovered.  This associated
		 * array holds a reference on the spare pool.
		 */
		assoc_array = md_asr_find_mddev(&toplevel_arrays,
						ntohl(asr_rt->raidmagic));
		if (assoc_array != NULL) {
			member_get(member);
			mddev_to_array_info(assoc_array)->spare_pool = array;
			MD_ASR_DEBUG(("attached spare pool %#x to array %#x\n",
				  array_info->raidmagic,
				  mddev_to_array_info(assoc_array)->raidmagic));
		}
	}

	if (member->pers != NULL)
		member->pers->raid_op(member, MDK_RAID_OP_INIT, NULL);

	if (parent != NULL) {
		if (parent->pers != NULL)
			parent->pers->raid_op(parent, MDK_RAID_OP_INIT_MEMBER,
					      member);
		MD_ASR_DEBUG(("md_asr_mddev_alloc: Adding array %p to %p\n",
			      array, parent));
		md_member_add_child(parent, member);
		if (subtype == MDK_MSUBTYPE_CHILD) {
			md_asr_update_array_state(to_mddev(parent));
			if (parent->pers != NULL) {
				parent->pers->raid_op(member,
						      MDK_RAID_OP_ADD_MEMBER,
						      NULL);
			}
		}
	}
		
	return (array);
}

static mdk_rdev_t *
md_asr_rdev_alloc(mddev_t *array, struct asr_reservedblock *asr_rb,
		  struct asr_raid_configline *asr_rcl,
		  struct block_device *blkdev,
		  mdk_msubtype_t subtype, int priority)
{
	struct md_asr_array_info *array_info;
	struct asr_reservedblock *new_rb;
	mdk_rdev_t *rdev;
	mdk_member_t *parent;
	mdk_member_t *member;
	int error;

	new_rb = kmalloc(sizeof(*new_rb), GFP_KERNEL);
	if (new_rb == NULL)
		return (NULL);

	memcpy(new_rb, asr_rb, sizeof(*new_rb));

	rdev = md_rdev_alloc(NULL, subtype, blkdev, priority);
	if (rdev == NULL) {
		kfree(new_rb);
		return (NULL);
	}

	array_info = mddev_to_array_info(array);
	member = to_member(rdev);
	member->metadata_private = new_rb;
	member->metapers = &md_asr_pers;
	error = md_asr_disk_to_member_state(array, asr_rcl->raidstate,
					    &member->state);
	if (error) {
		member_put(member);
		return (NULL);
	}
	member->total_size = ntohl(asr_rcl->lcapcty);
	member->data_size = member->total_size;
	parent = to_member(array);
	if (parent->pers != NULL)
		parent->pers->raid_op(parent, MDK_RAID_OP_INIT_MEMBER, member);
	md_member_add_child(parent, member);
	MD_ASR_DEBUG(("Added drive %#x to array %#x\n",
	       ntohl(asr_rb->drivemagic),
	       mddev_to_array_info(array)->raidmagic));
	/*
	 * Update our parents state to reflect the
	 * addition of this member.
	 */
	md_asr_update_array_state(array);
	if (parent->pers != NULL)
		parent->pers->raid_op(member, MDK_RAID_OP_ADD_MEMBER, NULL);

	return (rdev);
}

static void
md_asr_get_arrray_stats(mddev_t *array, int *optimal, int *degraded,
			int *failed, int *build_targets)
{
	struct list_head *list_item;
	u_int optimal_cnt;
	u_int degraded_cnt;
	u_int failed_cnt;
	u_int build_target_cnt;

	optimal_cnt = 0;
	degraded_cnt = 0;
	failed_cnt = 0;
	build_target_cnt = 0;

	/*
	 * Walk our member list and record the
	 * component state that effects our array
	 * state.
	 */
	list_item = array->member.child_set.next;
	while (list_item != &array->member.child_set) {
		mdk_member_t *member;

		member = list_entry(list_item, mdk_member_t, peer_links);
		list_item = list_item->next;
		switch (member->state) {
		case MDK_MSTATE_OPTIMAL:
			optimal_cnt++;
			break;
		case MDK_MSTATE_OPTIMAL|MDK_MSTATE_REBUILD_TARGET:
		case MDK_MSTATE_DEGRADED|MDK_MSTATE_REBUILD_TARGET:
			build_target_cnt++;
			break;
		case MDK_MSTATE_DEGRADED:
			degraded_cnt++;
			break;
		case MDK_MSTATE_FAILED:
		default:
			failed_cnt++;
			break;
		}
	}

	if (optimal != NULL)
		*optimal = optimal_cnt;
	if (degraded != NULL)
		*degraded = degraded_cnt;
	if (failed != NULL)
		*failed = failed_cnt;
	if (build_targets != NULL)
		*build_targets = build_target_cnt;
}

/*
 * Revalidate our array state.   This routine is called whenever
 * a member is added, removed, or changes state.
 */
static int
md_asr_update_array_state(mddev_t *array)
{
	mdk_set_state_arg_t sstate;
	struct md_asr_array_info *array_info;
	u_long flags;
	u_int configured_cnt;
	u_int degraded_need_cnt;
	u_int optimal_cnt;
	u_int degraded_cnt;
	u_int failed_cnt;
	u_int build_targets;

	array_info = mddev_to_array_info(array);
	MD_ASR_DEBUG(("md_asr_update_array_state: %p curstate = %#x\n",
		      array, array->member.state));
	
	/*
	 * Prevent state from changing during this update.
	 */
	spin_lock_irqsave(&array->member.member_lock, flags);
	sstate.oldstate = array->member.state;
	configured_cnt = ntohs(array_info->asr_rcl->raidcnt);

	md_asr_get_arrray_stats(array, &optimal_cnt, &degraded_cnt,
				&failed_cnt, &build_targets);
	MD_ASR_DEBUG(("md_asr_update_array_state: o %d d %d f %d b %d\n",
		      optimal_cnt, degraded_cnt, failed_cnt, build_targets));

	/*
	 * All of array types save RAID0 can operate
	 * with one failed/missing member.
	 */
	degraded_need_cnt = configured_cnt - 1;
	switch (array_info->asr_rcl->raidtype) {
	case ASR_RAIDSPR:
		/*
		 * Spare arrays are always optimal.
		 */
		sstate.newstate = MDK_MSTATE_OPTIMAL;
		break;
	case ASR_RAID0:
		/*
		 * RAID0's are marked failed unless
		 * all members are present and optimal
		 * or degraded.
		 */
		degraded_need_cnt = configured_cnt;
		/* FALLTHROUGH */
	case ASR_RAID1:
	case ASR_RAID4:
	case ASR_RAID5:
		if (optimal_cnt == configured_cnt) 
			sstate.newstate = MDK_MSTATE_OPTIMAL;
		else if (optimal_cnt + degraded_cnt >= degraded_need_cnt)
			sstate.newstate = MDK_MSTATE_DEGRADED;
		else if (sstate.oldstate & MDK_MSTATE_INITIALIZING) {
			sstate.newstate = MDK_MSTATE_INITIALIZING;
			/*
			 * If we have sufficent build targets,
			 * update our state.
			 */
			if (optimal_cnt + build_targets == configured_cnt) 
				sstate.newstate |= MDK_MSTATE_OPTIMAL;
			else if (optimal_cnt + build_targets
			      >= degraded_need_cnt)
				sstate.newstate |= MDK_MSTATE_DEGRADED;
		} else
			 sstate.newstate = MDK_MSTATE_FAILED;
		break;
	default:
		sstate.newstate = sstate.oldstate;
		break;
	}
	spin_unlock_irqrestore(&array->member.member_lock, flags);

	if (sstate.newstate != sstate.oldstate) {

		md_asr_member_op_set_state(&array->member, &sstate);
		MD_ASR_DEBUG(("Array %p transitions from %#x to %#x\n",
			      array, sstate.oldstate, sstate.newstate));

		if (array->member.pers != NULL) {
			array->member.pers->raid_op(to_member(array),
						    MDK_RAID_OP_MSTATE_CHANGED,
						    to_member(array));
		}
		if (array->member.subtype == MDK_MSUBTYPE_CHILD) {
			mddev_t *parent;

			parent  = to_mddev(array->member.parent);
			md_asr_update_array_state(parent);
		}
		return (1);
	}
	return (0);
}

/*
 * Update an existing array hierarchy to match the raid-table
 * pointed to by asr_rt.  This routine is typically called when
 * a member with more recent meta-data arrives after we've seen
 * at least one member with an out-of-date raid-table.
 */
static void
md_asr_apply_rt(mddev_t *toplevel_array, struct asr_raidtable *asr_rt)
{
	/* XXX Implement */
}

static int
md_asr_member_to_array_state(mddev_t *mddev, mdk_mstate_t md_state,
			     int *adaptec_state)
{
	struct md_asr_array_info *array_info;
	mdk_member_t *member;
	u_long flags;
	int building;

	array_info = mddev_to_array_info(mddev);
	member = to_member(mddev);
	spin_lock_irqsave(&member->member_lock, flags);
	building = array_info->asr_rcl->curAppBlock != 0;
	spin_unlock_irqrestore(&member->member_lock, flags);
	switch (md_state) {
	case MDK_MSTATE_OPTIMAL:
		*adaptec_state = LS_OPTIMAL_M;
		break;
	case MDK_MSTATE_OPTIMAL|MDK_MSTATE_INITIALIZING:
		*adaptec_state = LS_BUILDING_M;
		break;
	case MDK_MSTATE_DEGRADED:
	{
		if (building != 0)
			*adaptec_state = LS_REBUILDING_M;
		else
			*adaptec_state = LS_DEGRADED_M;
		break;
	}
	case MDK_MSTATE_FAILED:
		*adaptec_state = LS_DEAD_M;
		break;
	default:
		return (1);
		break; /* NOTREACHED */
	}

	return (0);
}

static int
md_asr_array_to_member_state(int adaptec_state, mdk_mstate_t *md_state)
{

	switch (adaptec_state) {
	case LS_OPTIMAL_M:
	case LS_VERIFYING_M:
		*md_state = MDK_MSTATE_OPTIMAL;
		break;
	case LS_BUILDING_M:
		*md_state = MDK_MSTATE_OPTIMAL|MDK_MSTATE_INITIALIZING;
		break;
	case LS_REBUILDING_M:
	case LS_DEGRADED_M:
		*md_state = MDK_MSTATE_DEGRADED;
		break;
	case LS_DEAD_M:
		*md_state = MDK_MSTATE_FAILED;
		break;
	default:
		return (1);
		break;
	}

	return (0);
}

static int
md_asr_member_to_disk_state(mdk_mstate_t md_state, int *adaptec_state)
{


	switch (md_state) {
	case MDK_MSTATE_OPTIMAL:
		*adaptec_state = LSU_COMPONENT_STATE_OPTIMAL;
		break;
	case MDK_MSTATE_OPTIMAL|MDK_MSTATE_INITIALIZING:
		*adaptec_state = LSU_COMPONENT_STATE_UNINITIALIZED
			       | LSU_COMPONENT_SUBSTATE_BUILDING;
		break;
	case MDK_MSTATE_OPTIMAL|MDK_MSTATE_REBUILD_TARGET:
		*adaptec_state = LSU_COMPONENT_STATE_REPLACED
			       | LSU_COMPONENT_SUBSTATE_REBUILDING;
		break;
	case MDK_MSTATE_FAILED:
		*adaptec_state = LSU_COMPONENT_STATE_FAILED;
		break;
	default:
		return (1);
		break; /* NOTREACHED */
	}

	return (0);
}

static int
md_asr_disk_to_member_state(mddev_t *array, int adaptec_state,
			    mdk_mstate_t *md_state)
{
	MD_ASR_DEBUG(("adaptec_disk_state_to_md: adaptec_state = %#x\n",
		     adaptec_state));
	switch (adaptec_state & MAIN_STATE) {
	case LSU_COMPONENT_STATE_OPTIMAL:
		*md_state = MDK_MSTATE_OPTIMAL;
		if (array != NULL
		 && (to_member(array)->state & MDK_MSTATE_INITIALIZING) != 0)
			*md_state |= MDK_MSTATE_REBUILD_TARGET;
		break;
	case LSU_COMPONENT_STATE_DEGRADED:
	case LSU_COMPONENT_STATE_FAILED:
		*md_state = MDK_MSTATE_FAILED;
		break;
	case LSU_COMPONENT_STATE_REPLACED:
		*md_state = MDK_MSTATE_REBUILD_TARGET|MDK_MSTATE_OPTIMAL;
		break;
	case LSU_COMPONENT_STATE_UNINITIALIZED:
		*md_state = MDK_MSTATE_INITIALIZING|MDK_MSTATE_OPTIMAL;
		break;
	case LSU_ARRAY_SUBSTATE_AWAIT_FORMAT:
	case LSU_COMPONENT_STATE_UNCONFIGURED:
	default:
		/*
		 * In general these states don't come up,
		 */
		MD_ASR_DEBUG(("adaptec_disk_state_to_md: unexpected "
			      "adaptec_state = %#x\n", adaptec_state));
		*md_state = MDK_MSTATE_FAILED;
		return (1);
	}

	return (0);
}

static int
md_asr_to_md_raidtype(uint8_t asr_raidtype)
{
	int raidtype;

	switch (asr_raidtype) {
	case ASR_RAID0:
		raidtype = RAID0;
		break;
	case ASR_RAID1:
		raidtype = RAID1;
		break;
	case ASR_RAID5:
		raidtype = RAID5;
		break;
	default:
		raidtype = MD_RESERVED;
		break;
	}

	return (raidtype);
}
static const char *
md_asr_raidlevel_string(int raidlevel)
{
	const char *string;

	switch (raidlevel) {
	case FWP:
		string = "FWP";
		break;
	case FWL:
		string = "FWL";
		break;
	case OSI:
		string = "OSI";
		break;
	case FWL_2:
		string = "FWL_2";
		break;
	default:
		string = "Level UNKNOWN";
		break;
	}

	return (string);
}

static const char *
md_asr_raidtype_string(int raidtype)
{
	const char *string;

	switch (raidtype) {
	case ASR_RAID0:
		string = "RAID0";
		break;
	case ASR_RAID1:
		string = "RAID1";
		break;
	case ASR_RAID5:
		string = "RAID5";
		break;
	case ASR_RAIDRED:
		string = "RAIDRED";
		break;
	case ASR_RAIDSPR:
		string = "RAIDSPR";
		break;
	default:
		string = "Type UNKNOWN";
		break;
	}

	return (string);
}

static void
md_asr_dump_rb(struct asr_reservedblock *asr_rb)
{
#if MD_DEBUG
	printk("ASR Reserved Block\n");
	printk("    b0idcode = 0x%x, sb0flags = 0x%x, jbodEnable = 0x%x\n",
	       ntohl(asr_rb->b0idcode), asr_rb->sb0flags, asr_rb->jbodEnable);
	printk("    biosInfo = 0x%x, fstrsvrb = 0x%x, "
	       "svBlockStorageTid = 0x%x\n",
	       ntohs(asr_rb->biosInfo), ntohl(asr_rb->fstrsvrb),
	       ntohs(asr_rb->svBlockStorageTid));
	printk("    svtid = 0x%x, drivemagic = 0x%x, raidtbl = 0x%x\n",
	       ntohs(asr_rb->svtid), ntohl(asr_rb->drivemagic),
	       ntohs(asr_rb->raidtbl));
#endif
}

static void
md_asr_dump_rcl(struct asr_raid_configline *rcl, int index)
{
#if MD_DEBUG
	printk("RCL Entry [%d] %s %s Magic %#x LSU %#x\n",
	       index,  md_asr_raidlevel_string(rcl->raidlevel),
	       md_asr_raidtype_string(rcl->raidtype),
	       ntohl(rcl->raidmagic), ntohl(rcl->lsu));
	printk("    raidcnt = %d, raidseq = %d\n",
	       ntohs(rcl->raidcnt), ntohs(rcl->raidseq));
	printk("    raidstate = %#x, flags = %#x, refcnt = %d\n",
	       rcl->raidstate, rcl->flags, rcl->refcnt);
	printk("    hba = %d, channel = %d, TID = %d, LUN = %d\n",
	       rcl->raidhba, rcl->raidchnl, ntohl(rcl->raidid), rcl->raidlun);
	printk("    loffset = %d, lcapcty = %d, strpsize = %d\n",
	       ntohl(rcl->loffset), ntohl(rcl->lcapcty),
	       ntohs(rcl->strpsize));
#endif
}

static void
md_asr_dump_rt(struct asr_raidtable *asr_rt)
{
#if MD_DEBUG
	u_int elmcnt;
	u_int i;

	elmcnt = ntohs(asr_rt->elmcnt);
	printk("Raid Table RID %#x Version %#x NumElements %d\n",
	       ntohl(asr_rt->ridcode), ntohl(asr_rt->rversion), elmcnt);
	printk("    timestamp = %#x, rchksum = %#x, raidFlags = %#x\n",
	       ntohs(asr_rt->timestamp), ntohl(asr_rt->rchksum),
	       ntohl(asr_rt->raidFlags));
	printk("    irocFlags = %#x, dirty = %#x, actionPriority = %#x\n",
	       asr_rt->irocFlags, asr_rt->dirty, asr_rt->actionPriority);
	printk("    raidmagic = %#x, sparemagic = %#x, spareid = %#x\n",
	       ntohl(asr_rt->raidmagic), ntohl(asr_rt->sparedrivemagic),
	       asr_rt->spareid);
	printk("    verifyDate = %#x, recreateDate = %#x\n",
	       ntohl(asr_rt->verifyDate), ntohl(asr_rt->recreateDate));
	for (i = 0; i < elmcnt; i++)
		md_asr_dump_rcl(&asr_rt->ent[i], i);
#endif
}

static void
md_asr_dump_minfo( struct md_asr_meta_info *minfo)
{
	md_asr_dump_rb(&minfo->asr_rb);
	md_asr_dump_rt(&minfo->asr_rt);
}

static int __init
md_asr_init(void)
{
	int error;

	error = md_init_workitem_support(NULL, &md_asr_syncer, "asr_syncer");
	if (error != 0)
		return (error);
	error = register_md_metapersonality(&md_asr_pers);
	if (error != 0)
		md_teardown_workitem_support(&md_asr_syncer);
	return (error);
}

static void md_asr_exit (void)
{
	unregister_md_metapersonality(&md_asr_pers);
	md_teardown_workitem_support(&md_asr_syncer);
}

module_init(md_asr_init);
module_exit(md_asr_exit);
MODULE_LICENSE("Dual BSD/GPL");
