/*
 * raid1.c : Multiple Devices driver for Linux
 *
 * Copyright (C) 1999, 2000, 2001 Ingo Molnar, Red Hat
 *
 * Copyright (C) 1996, 1997, 1998 Ingo Molnar, Miguel de Icaza, Gadi Oxman
 *
 * RAID-1 management functions.
 *
 * Better read-balancing code written by Mika Kuoppala <miku@iki.fi>, 2000
 *
 * Fixes to reconstruction by Jakob )Bstergaard" <jakob@ostenfeld.dk>
 * Various fixes by Neil Brown <neilb@cse.unsw.edu.au>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * You should have received a copy of the GNU General Public License
 * (for example /usr/src/linux/COPYING); if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/raid/raid1.h>

#define R1_READABLE_MEMBER(conf, member, below_sync_mark)	\
    (MDK_READABLE_MEMBER(member)				\
  || (below_sync_mark && MDK_WRITABLE_MEMBER(member)))

#define R1_WRITABLE_MEMBER(member) MDK_WRITABLE_MEMBER(member)
    
#define R1_REBUILDABLE(array)					\
 (((array)->state & ~(MDK_MSTATE_DEGRADED|MDK_MSTATE_REBUILD_TARGET)) == 0)

#define R1_INITIALIZABLE(array)					\
   (((array)->state & MDK_MSTATE_INITIALIZING)			\
 && ((array)->state & (MDK_MSTATE_DEGRADED|MDK_MSTATE_OPTIMAL)))

/*
 * Number of guaranteed r1bios in case of extreme VM load:
 */
#define	NR_RAID1_BIOS 256

static int raid1_workop(mdk_member_t *member, int op, void *arg);
static void raid1_initialize_array(mddev_t *mddev);
static void raid1_attempt_rebuild(mddev_t *mddev);
static mdk_personality_t raid1_personality;
static spinlock_t retry_list_lock = SPIN_LOCK_UNLOCKED;
static LIST_HEAD(retry_list_head);

static void * r1bio_pool_alloc(int gfp_flags, void *data)
{
	mddev_t *mddev = data;
	r1bio_t *r1_bio;
	size_t	alloc_size;
	u_int	num_bios;
	u_int	i;

	/*
	 * allocate a r1bio with room to support this
	 * array if all members were optimal.  Should
	 * this module every support adding new mirror
	 * members beyond the initial configuration,
	 * some method for "updating" pre-allocated
	 * r1bios in the mempool to include room for
	 * the new member must be added.
	 */
	num_bios = mddev->max_disks;
	alloc_size = sizeof(r1bio_t) + sizeof(r1bio_holder_t) * num_bios;
	r1_bio = kmalloc(alloc_size, gfp_flags);
	if (r1_bio) {
		memset(r1_bio, 0, alloc_size);
		r1_bio->mddev = mddev;

		r1_bio->num_bios = num_bios;
		for (i = 0; i < num_bios; i++)
			r1_bio->bios[i].r1bio = r1_bio;
	}
	return r1_bio;
}

static void r1bio_pool_free(void *r1_bio, void *data)
{
	kfree(r1_bio);
}

#define RESYNC_BLOCK_SIZE (64*1024)
#define RESYNC_SECTORS (RESYNC_BLOCK_SIZE >> 9)
#define RESYNC_PAGES ((RESYNC_BLOCK_SIZE + PAGE_SIZE-1) / PAGE_SIZE)
#define RESYNC_WINDOW (1024*1024)

static void * r1buf_pool_alloc(int gfp_flags, void *data)
{
	conf_t *conf = data;
	struct page *page;
	r1bio_t *r1_bio;
	struct bio *bio;
	int i, j;

	r1_bio = r1bio_pool_alloc(gfp_flags, conf->mddev);
	if (!r1_bio)
		return NULL;

	/*
	 * Allocate bios : 1 for reading, n-1 for writing
	 */
	for (j = r1_bio->num_bios ; j-- ; ) {
		bio = bio_alloc(gfp_flags, RESYNC_PAGES);
		if (!bio)
			goto out_free_bio;
		r1_bio->bios[j].bio = bio;
	}
	/*
	 * Allocate RESYNC_PAGES data pages and attach them to
	 * the first bio;
	 */
	bio = r1_bio->bios[0].bio;
	for (i = 0; i < RESYNC_PAGES; i++) {
		uint8_t *vaddr;

		page = alloc_page(gfp_flags);
		if (unlikely(!page))
			goto out_free_pages;

		bio->bi_io_vec[i].bv_page = page;
		/*
		 * Pre-zero the page if this buffer is
		 * being used for INIT operations.
		 */
		/*
		 * Pre-zero the pages so that if
		 * this buffer is being used for
		 * INIT operations, the sync code
		 * doesn't have to zero pages on
		 * every buf allocation.
		 */
		if (to_member(conf->mddev)->state & MDK_MSTATE_INITIALIZING) {
			vaddr = page_address(page);
			memset(vaddr, 0, PAGE_SIZE);
		}
	}

	r1_bio->master_bio = bio;

	return r1_bio;

out_free_pages:
	for ( ; i > 0 ; i--)
		__free_page(bio->bi_io_vec[i-1].bv_page);
out_free_bio:
	while ( ++j < r1_bio->num_bios )
		bio_put(r1_bio->bios[j].bio);
	r1bio_pool_free(r1_bio, conf->mddev);
	return NULL;
}

static void r1buf_pool_free(void *__r1_bio, void *data)
{
	int i;
	conf_t *conf = data;
	r1bio_t *r1bio = __r1_bio;
	struct bio *bio = r1bio->bios[0].bio;

	for (i = 0; i < RESYNC_PAGES; i++) {
		__free_page(bio->bi_io_vec[i].bv_page);
		bio->bi_io_vec[i].bv_page = NULL;
	}
	for (i=0 ; i < r1bio->num_bios; i++)
		bio_put(r1bio->bios[i].bio);

	r1bio_pool_free(r1bio, conf->mddev);
}

static void put_all_bios(conf_t *conf, r1bio_t *r1_bio)
{
	int i;

	for (i = 0; i < r1_bio->num_bios; i++) {
		r1bio_holder_t *bioh = &r1_bio->bios[i];
		if (bioh->bio)
			bio_put(bioh->bio);
		bioh->bio = NULL;
	}
}

static inline void put_io_r1bio(r1bio_t *r1_bio, int rw)
{
	conf_t *conf = mddev_to_conf(r1_bio->mddev);

	/*
	 * Wake up any possible resync thread that waits for the device
	 * to go idle.
	 */
	syncable_vol_io_complete(&conf->syncinfo, rw, r1_bio->phase);

	put_all_bios(conf, r1_bio);
	mempool_free(r1_bio, conf->r1bio_pool);
}

static inline void put_sync_r1bio(r1bio_t *r1_bio)
{
	conf_t *conf = mddev_to_conf(r1_bio->mddev);
	u_int phase;

	phase = r1_bio->phase;
	mempool_free(r1_bio, conf->r1buf_pool);
	syncable_vol_syncio_complete(&conf->syncinfo, phase);
}

/*
 * Find a readable member for this array.
 */
static int map(mddev_t *mddev, mdk_member_t **member)
{
	mdk_member_t *array = to_member(mddev);
	mdk_member_t *child;
	struct list_head *tmp;

	/*
	 * Later we do read balancing on the read side
	 * now we use the first available disk.
	 */
	spin_lock_irq(&array->member_lock);
	ITERATE_CHILDREN_LOCKED(array, child, tmp) {

		if (MDK_READABLE_MEMBER(child)) {
			*member = child;
			member_get(child);
			spin_unlock_irq(&array->member_lock);
			return 0;
		}
	}
	spin_unlock_irq(&array->member_lock);

	printk(KERN_ERR "raid1_map(): huh, no more operational devices?\n");
	return -1;
}

static void reschedule_retry(r1bio_t *r1_bio)
{
	unsigned long flags;
	mddev_t *mddev = r1_bio->mddev;

	spin_lock_irqsave(&retry_list_lock, flags);
	list_add(&r1_bio->retry_list, &retry_list_head);
	spin_unlock_irqrestore(&retry_list_lock, flags);

	md_wakeup_thread(mddev->thread);
}

/*
 * raid_end_bio_io() is called when we have finished servicing a mirrored
 * operation and are ready to return a success/failure code to the buffer
 * cache layer.
 */
static void raid_end_bio_io(r1bio_t *r1_bio)
{
	struct bio *bio = r1_bio->master_bio;
	int rw;

	rw = bio_data_dir(bio);
	bio_endio(bio, bio->bi_size,
		  test_bit(R1BIO_Uptodate, &r1_bio->state) ? 0 : -EIO);
	put_io_r1bio(r1_bio, rw);
}

/*
 * Update disk head position estimator based on IRQ completion info.
 */
static inline void update_head_pos(mdk_member_t *member, r1bio_t *r1_bio)
{
	mconf_t *mconf;

	mconf = member_to_mconf(member);
	mconf->head_position = r1_bio->sector + r1_bio->sectors;
}

static int raid1_end_read_request(struct bio *bio, unsigned int bytes_done,
				  int error)
{
	int uptodate;
	r1bio_holder_t *bioh;
	r1bio_t *r1_bio;
	mdk_member_t *member;

	if (bio->bi_size)
		return 1;
	
	uptodate = test_bit(BIO_UPTODATE, &bio->bi_flags);
	bioh = (r1bio_holder_t *)bio->bi_private;
	r1_bio = bioh->r1bio;
	member = bioh->member;

	/*
	 * this branch is our 'one mirror IO has finished' event handler:
	 */
	if (!uptodate) {

		/*
		 * oops, read error:
		 */
		md_set_member_state(member, MDK_MSTATE_FAILED);
		printk(KERN_ERR "raid1: %s: rescheduling sector %llu\n",
		       mdname(member), (unsigned long long)r1_bio->sector);
		reschedule_retry(r1_bio);
	} else {
		/*
		 * Set R1BIO_Uptodate in our master bio, so that
		 * we will return a good error code for to the higher
		 * levels even if IO on some other mirrored buffer fails.
		 *
		 * The 'master' represents the composite IO operation to
		 * user-side. So if something waits for IO, then it will
		 * wait for the 'master' bio.
		 */
		set_bit(R1BIO_Uptodate, &r1_bio->state);
		raid_end_bio_io(r1_bio);
	}

	member_put(member);
	return 0;
}

static int raid1_end_write_request(struct bio *bio, u_int bytes_done, int error)
{
	int uptodate;
	r1bio_holder_t *r1bioh;
	r1bio_t *r1_bio;

	if (bio->bi_size)
		return 1;

	uptodate = test_bit(BIO_UPTODATE, &bio->bi_flags);
	r1bioh = (r1bio_holder_t *)(bio->bi_private);
	r1_bio = r1bioh->r1bio;

	/*
	 * this branch is our 'one mirror IO has finished' event handler:
	 */
	if (!uptodate) {
		md_set_member_state(r1bioh->member, MDK_MSTATE_FAILED);
	} else {
		/*
		 * Set R1BIO_Uptodate in our master bio, so that
		 * we will return a good error code for to the higher
		 * levels even if IO on some other mirrored buffer fails.
		 *
		 * The 'master' represents the composite IO operation to
		 * user-side. So if something waits for IO, then it will
		 * wait for the 'master' bio.
		 */
		set_bit(R1BIO_Uptodate, &r1_bio->state);
	}

	/*
	 *
	 * Let's see if all mirrored write operations have finished
	 * already.
	 */
	if (atomic_dec_and_test(&r1_bio->remaining)) {
		md_write_end(r1_bio->mddev);
		raid_end_bio_io(r1_bio);
	}

	member_put(r1bioh->member);
	return 0;
}


/*
 * This routine returns the disk from which the requested read should
 * be done. There is a per-array 'next expected sequential IO' sector
 * number - if this matches on the next IO then we use the last disk.
 * There is also a per-disk 'last know head position' sector that is
 * maintained from IRQ contexts, both the normal and the resync IO
 * completion handlers update this position correctly. If there is no
 * perfect sequential match then we pick the disk whose head is closest.
 *
 * If there are 2 mirrors in the same 2 devices, performance degrades
 * because position is mirror, not device based.
 *
 * The rdev for the device selected will have nr_pending incremented.
 */
static mdk_member_t *read_balance(mddev_t *mddev, struct bio *bio,
				  r1bio_t *r1_bio, int below_sync_mark)
{
	mdk_member_t *array = to_member(mddev);
	struct list_head *list_item;
	conf_t *conf = mddev_to_conf(mddev);
	mdk_member_t *new_disk;
	mdk_member_t *disk;
	mconf_t *mconf;
	sector_t new_distance, current_distance;

	/*
	 * Traverse the list of members looking for the one who's
	 * head position is closest to this request.  Optimal or
	 * degraded members (meaning the member is another array
	 * that is degraded) can handle reads to any sector.  Rebuild
	 * targets can only handle requests below the last rebuild
	 * checkpoint.
	 */
	if (list_empty(&array->child_set))
		return (NULL);

	new_disk = conf->last_used;
	if (new_disk == NULL) {
		new_disk = list_entry(array->child_set.next,
				      mdk_member_t, peer_links);
	}
		
	/*
	 * If our last disk is not optimal, find some disk that is.
	 */
	if (!R1_READABLE_MEMBER(conf, new_disk, below_sync_mark)) {
		new_disk = NULL;
		ITERATE_CHILDREN_LOCKED(array, disk, list_item) {
			if (R1_READABLE_MEMBER(conf, disk, below_sync_mark)) {
				new_disk = disk;
				break;
			}
		}
	}
	if (new_disk == NULL)
		return NULL;

	/*
	 * Don't change to another disk for sequential reads:
	 */
	mconf = member_to_mconf(new_disk);
	if (mconf->head_position == r1_bio->sector)
		return (new_disk);

	/*
	 * Find the disk whose head is closest.
	 *
	 * Update last_used so we will always terminate the
	 * loop on a member in optimal state even if no other
	 * valid member is found.  new_disk is guaranteed to
	 * by optimal by the loops above.
	 */
	conf->last_used = new_disk;
	disk = new_disk;
	current_distance = abs(r1_bio->sector - new_disk->head_position);

	do {
		struct list_head *next;

		next = disk->peer_links.next;
		if (next == &array->child_set)
			next = array->child_set.next;
		disk = list_entry(next, mdk_member_t, peer_links);
		if (!R1_READABLE_MEMBER(conf, disk, below_sync_mark))
			continue;

		mconf = member_to_mconf(disk);
		new_distance = abs(r1_bio->sector - mconf->head_position);
		if (new_distance < current_distance) {
			current_distance = new_distance;
			new_disk = disk;
		}
	} while (disk != conf->last_used);

	update_head_pos(disk, r1_bio);
	conf->last_used = new_disk;

	return new_disk;
}

static int make_request(request_queue_t *q, struct bio * bio)
{
	mddev_t *mddev = q->queuedata;
	mdk_member_t *array = to_member(mddev);
	conf_t *conf = mddev_to_conf(mddev);
	mdk_member_t *member;
	struct list_head *list_item;
	r1bio_t *r1_bio;
	struct bio *read_bio;
	u_long flags;
	u_int num_bios;
	u_int i;
	int cur_phase;
	int below_sync_mark;

	/* Block on the sync window is necessary. */
	cur_phase = syncable_vol_io_start(&conf->syncinfo, bio,
					  &below_sync_mark);

	if (bio_data_dir(bio) == WRITE) {
		disk_stat_inc(mddev->gendisk, writes);
		disk_stat_add(mddev->gendisk, write_sectors, bio_sectors(bio));
	} else {
		disk_stat_inc(mddev->gendisk, reads);
		disk_stat_add(mddev->gendisk, read_sectors, bio_sectors(bio));
	}

	/*
	 * make_request() can abort the operation when READA is being
	 * used and no empty request is available.
	 *
	 */
	r1_bio = mempool_alloc(conf->r1bio_pool, GFP_NOIO);

	r1_bio->master_bio = bio;
	r1_bio->sectors = bio->bi_size >> 9;

	r1_bio->sector = bio->bi_sector;
	r1_bio->phase = cur_phase;

	if (bio_data_dir(bio) == READ) {
		/*
		 * read balancing logic:
		 */
		spin_lock_irqsave(&array->member_lock, flags);
		member = read_balance(mddev, bio, r1_bio, below_sync_mark);
		if (member == NULL) {
/* XXX abort request??? */
			spin_unlock_irqrestore(&array->member_lock, flags);
			return 0;
		}
		member_get(member);
		spin_unlock_irqrestore(&array->member_lock, flags);

		read_bio = bio_clone(bio, GFP_NOIO);

		r1_bio->bios[0].member = member;
		r1_bio->bios[0].bio = read_bio;

		read_bio->bi_sector = r1_bio->sector + member->data_offset;
		read_bio->bi_end_io = raid1_end_read_request;
		read_bio->bi_rw = READ;
		read_bio->bi_private = &r1_bio->bios[0];

		md_make_request(member, read_bio);
		return 0;
	}

	/*
	 * WRITE:
	 * 
	 * First select target members under spinlock and
	 * inc their refcount so that they cannot be deleted.
	 */
	num_bios = 0;
	spin_lock_irqsave(&array->member_lock, flags);
	ITERATE_CHILDREN_LOCKED(array, member, list_item) {

		if (R1_WRITABLE_MEMBER(member)) {

			member_get(member);
			r1_bio->bios[num_bios].member = member;
			num_bios++;
		}
	}
	spin_unlock_irqrestore(&array->member_lock, flags);

	atomic_set(&r1_bio->remaining, num_bios);
	for (i = 0; i < num_bios; i++) {
		struct bio *wbio;

		member = r1_bio->bios[i].member;
		wbio = bio_clone(bio, GFP_NOIO);
		wbio->bi_sector	= r1_bio->sector + member->data_offset;
		wbio->bi_end_io	= raid1_end_write_request;
		wbio->bi_rw = WRITE;
		wbio->bi_private = &r1_bio->bios[i];
		r1_bio->bios[i].bio = wbio;
		md_make_request(member, wbio);
	}

	return 0;
}

struct r1_status_info {
	u_int	member_num;
	char	state_code;
};

static void status(struct seq_file *seq, mddev_t *mddev)
{
	struct r1_status_info	*status;
	conf_t *conf = mddev_to_conf(mddev);
	mdk_member_t *array;
	mdk_member_t *member;
	struct list_head *tmp;
	u_int num_members;
	u_int working_members;
	u_int i;

	status = kmalloc(sizeof(*status) * conf->raid_disks, GFP_KERNEL);
	if (status == NULL)
		return;
	array = to_member(mddev);
	num_members = 0;
	working_members = 0;
	spin_lock_irq(&array->member_lock);
	ITERATE_CHILDREN_LOCKED(array, member, tmp) {
		char state_code;

		status[num_members].member_num = member->priority;
		if (member->state & MDK_MSTATE_REBUILD_TARGET)
			state_code = 'R';
		else if (member->state & MDK_MSTATE_INITIALIZING)
			state_code = 'I';
		else {
			switch (member->state) {
			case MDK_MSTATE_OPTIMAL:
				state_code = 'O';
				working_members++;
				break;
			case MDK_MSTATE_DEGRADED:
				state_code = 'D';
				working_members++;
				break;
			case MDK_MSTATE_FAILED:
				state_code = 'F';
				break;
			default:
				state_code = 'U';
				break;
			}
		}
		status[num_members].state_code = state_code;
		num_members++;
	}
	spin_unlock_irq(&array->member_lock);

	seq_printf(seq, " [%d/%d] [", conf->raid_disks, working_members);
	for (i = 0; i < num_members; i++) {

		seq_printf(seq, "%d:%c ", status[i].member_num,
			   status[i].state_code);
	}
	seq_printf(seq, "]");
}

static void error(mddev_t *mddev, mdk_member_t *member)
{
	mdk_mstate_t  newstate;
	mdk_mstate_t *pstate;

#if 0
	printk(KERN_ALERT "raid1: Disk failure on %s, disabling device. \n"
		"	Operation continuing on %d devices\n",
		mdname(member), conf->working_disks);
#endif
	newstate = MDK_MSTATE_FAILED;
	pstate = &newstate;
	to_member(mddev)->metapers->member_op(member, MDK_MEMBER_OP_SET_STATE,
					      (void **)&pstate);
}

static void print_conf(conf_t *conf)
{
#if 0
	int i;
	mirror_info_t *tmp;

	printk("RAID1 conf printout:\n");
	if (!conf) {
		printk("(!conf)\n");
		return;
	}
	printk(" --- wd:%d rd:%d\n", conf->working_disks,
		conf->raid_disks);

	for (i = 0; i < conf->raid_disks; i++) {
		char b[BDEVNAME_SIZE];
		tmp = conf->mirrors + i;
		if (tmp->rdev)
			printk(" disk %d, wo:%d, o:%d, dev:%s\n",
				i, !tmp->rdev->in_sync, !tmp->rdev->faulty,
				bdevname(tmp->rdev->bdev,b));
	}
#endif
}

static int end_sync_read(struct bio *bio, unsigned int bytes_done, int error)
{
	int uptodate = test_bit(BIO_UPTODATE, &bio->bi_flags);
	r1bio_holder_t *bioh = (r1bio_holder_t *)(bio->bi_private);
	r1bio_t * r1_bio = bioh->r1bio;
	mdk_member_t *member = bioh->member;

	if (bio->bi_size)
		return 1;

	if (bioh->bio != bio)
		BUG();
	/*
	 * we have read a block, now it needs to be re-written,
	 * or re-read if the read failed.
	 * We don't do much here, just schedule handling by raid1d
	 */
	if (!uptodate)
		md_set_member_state(member, MDK_MSTATE_FAILED);
	else
		set_bit(R1BIO_Uptodate, &r1_bio->state);
	member_put(member);
	reschedule_retry(r1_bio);
	return 0;
}

static int end_sync_write(struct bio *bio, unsigned int bytes_done, int error)
{
	int uptodate = test_bit(BIO_UPTODATE, &bio->bi_flags);
	r1bio_holder_t *bioh = (r1bio_holder_t *)(bio->bi_private);
	r1bio_t *r1_bio = bioh->r1bio;
	mdk_member_t *member = bioh->member;

	if (bio->bi_size)
		return 1;

	if (!uptodate)
		md_set_member_state(member, MDK_MSTATE_FAILED);

	if (atomic_dec_and_test(&r1_bio->remaining))
		put_sync_r1bio(r1_bio);
	member_put(member);
	return 0;
}

static int raid1_sync_write(mddev_t *mddev, r1bio_t *r1_bio)
{
	r1bio_holder_t *bioh;
	struct bio *wbio;
	int i;

	/*
	 * The write data is always placed in
	 * the first bio.
	 */
	bioh = &r1_bio->bios[0];

	/*
	 * schedule writes
	 */
	if (!test_bit(R1BIO_Uptodate, &r1_bio->state)) {
		/*
		 * There is no point trying a read-for-reconstruct as
		 * reconstruct is about to be aborted
		 */
		printk(KERN_ALERT "raid1: %s: unrecoverable I/O read error"
			" for block %llu\n", mdname(bioh->member), 
			(unsigned long long)r1_bio->sector);
		put_sync_r1bio(r1_bio);
		return 0;
	}

	for (i = 0; i < r1_bio->num_bios; i++) {
		if (r1_bio->bios[i].member == NULL)
			break;
		wbio = r1_bio->bios[i].bio;
		if (wbio->bi_end_io != end_sync_write)
			continue;
		md_sync_acct(r1_bio->bios[i].member, wbio->bi_size >> 9);
		md_make_request(r1_bio->bios[i].member, wbio);
	}
	return 0;
}

/*
 * perform a "sync" on one "block"
 *
 * We need to make sure that no normal I/O request - particularly write
 * requests - conflict with active sync requests.
 *
 * This is achieved by tracking pending requests and a 'barrier' concept
 * that can be installed to exclude normal IO requests.
 */
static int raid1_sync_request(mddev_t *mddev, mdk_workitem_t *workitem)
{
	mdk_member_t *array;
	mdk_member_t *child;
	struct list_head *tmp;
	conf_t *conf;
	mdk_sync_support_t *syncinfo;
	md_sector_range_t *syncrange;
	r1bio_t *r1_bio;
	r1bio_holder_t *bioh;
	struct bio *first_bio;
	struct bio *bio;
	sector_t nr_sectors;
	u_long flags;
	int num_bios;
	int i;

	array = to_member(mddev);
	conf = mddev_to_conf(mddev);
	syncinfo = &conf->syncinfo;
	syncrange = &syncinfo->sync_range[syncinfo->phase];
	
	if (syncrange->start >= array->data_size)
		return -EINVAL;

	/*
	 * Allocate an r1_bio for the sync operation.
	 */
	r1_bio = mempool_alloc(conf->r1buf_pool, GFP_NOIO);
	r1_bio->mddev = mddev;
	r1_bio->sector = syncrange->start;
	set_bit(R1BIO_IsSync, &r1_bio->state);
	/*
	 * Loop through our disks assigning bios to
	 * them.  If we are initializing the array, no
	 * read is necessary.  Just schedule a write to
	 * all writable members.
	 */
	num_bios = 0;
	if (workitem->work_opcode == MD_SYNCOP_INIT)
		bioh = &r1_bio->bios[0];
	else
		bioh = &r1_bio->bios[1];
	first_bio = r1_bio->bios[0].bio;
	r1_bio->bios[0].member = NULL;
	spin_lock_irqsave(&array->member_lock, flags);
	ITERATE_CHILDREN_LOCKED(array, child, tmp) {
		if (MDK_READABLE_MEMBER(child)) {
			if (workitem->work_opcode != MD_SYNCOP_INIT
			 && r1_bio->bios[0].member == NULL) {
				/*
				 * Found a suitable member for
				 * reading.  Assign to buf 0.
				 */
				member_get(child);
				r1_bio->bios[0].member = child;
				r1_bio->bios[0].bio->bi_rw = READ;
				r1_bio->bios[0].bio->bi_end_io = end_sync_read;
				num_bios++;
			}
			/*
			 * No need to write an optimal or
			 * degraded member...
			 */
		} else if (MDK_WRITABLE_MEMBER(child)) {
			member_get(child);
			bioh->member = child;
			bioh->bio->bi_rw = WRITE;
			bioh->bio->bi_end_io = end_sync_write;
			bioh++;
			num_bios++;
		}
	}
	spin_unlock_irqrestore(&array->member_lock, flags);
	if (r1_bio->bios[0].member == NULL) {
		/*
		 * No valid member to read from.
		 */
		printk("raid1_sync_request: No Valid Sync Devices\n");
		for (i = 0; i < num_bios; i++) {
			if (r1_bio->bios[i].member != NULL)
				member_put(r1_bio->bios[i].member);
		}
		mempool_free(r1_bio, conf->r1buf_pool);
		return -EIO;
	}
	r1_bio->phase = syncable_vol_syncio_start(syncinfo);
	for (i = 0; i < num_bios; i++) {
		bioh = &r1_bio->bios[i];
		bio = bioh->bio;

		/* take from bio_init */
		bio->bi_next = NULL;
		bio->bi_flags |= 1 << BIO_UPTODATE;
		bio->bi_vcnt = 0;
		bio->bi_idx = 0;
		bio->bi_phys_segments = 0;
		bio->bi_hw_segments = 0;
		bio->bi_size = 0;
		bio->bi_sector = syncrange->start + bioh->member->data_offset;
		bio->bi_private = bioh;
		/*
		 * XXX Assumes member is an RDEV.
		 * Won't work for multi-level arrays.
		 * Need __bio_add_page(q, bio,...) interface.
		 */
		bio->bi_bdev = to_rdev(bioh->member)->bdev;
	}
	/*
	 * Set a marker so our write routine knows where to stop.
	 */
	if (num_bios < r1_bio->num_bios)
		r1_bio->bios[num_bios].member = NULL;
	nr_sectors = 0;
	do {
		struct page *page;
		int len = PAGE_SIZE;

		if (syncrange->start + (len >> 9) > array->data_size)
			len = (array->data_size - syncrange->start) << 9;
		if (syncrange->start + (len >> 9) > syncrange->end)
			len = (syncrange->end - syncrange->start) << 9;
		if (len == 0)
			break;
		for (i = 0; i < num_bios; i++) {
			bioh = &r1_bio->bios[i];
			bio = bioh->bio;

			page = first_bio->bi_io_vec[bio->bi_vcnt].bv_page;
			if (bio_add_page(bio, page, len, 0) == 0) {
				/* stop here */
				first_bio->bi_io_vec[bio->bi_vcnt].bv_page =
				    page;
				while (i > 0) {
					i--;
					bio = r1_bio->bios[i].bio;
					/* remove last page from this bio */
					bio->bi_vcnt--;
					bio->bi_size -= len;
					bio->bi_flags &= ~(1<< BIO_SEG_VALID);
				}
				goto bio_full;
			}
		}
		nr_sectors += len>>9;
		syncrange->start += len>>9;
	} while (first_bio->bi_vcnt < RESYNC_PAGES);
 bio_full:
	r1_bio->sectors = nr_sectors;

	if (workitem->work_opcode == MD_SYNCOP_INIT) {
		atomic_set(&r1_bio->remaining, num_bios);
		set_bit(R1BIO_Uptodate, &r1_bio->state);
		raid1_sync_write(mddev, r1_bio);
		return 0;
	}

	/*
	 * Don't count the read in the writes
	 * remaining count.
	 */
	atomic_set(&r1_bio->remaining, num_bios - 1);

	md_sync_acct(r1_bio->bios[0].member, nr_sectors);

	md_make_request(r1_bio->bios[0].member, first_bio);

	return 0;
}

/*
 * This is a kernel thread which:
 *
 *	1.	Retries failed read operations on working mirrors.
 *	2.	Updates the raid superblock when problems encounter.
 *	3.	Performs writes following reads for array syncronising.
 */
static void raid1d(mddev_t *mddev, void *arg)
{
	char b[BDEVNAME_SIZE];
	struct list_head *head = &retry_list_head;
	r1bio_t *r1_bio;
	struct bio *bio;
	unsigned long flags;
	conf_t *conf = mddev_to_conf(mddev);
	mdk_member_t *member;

	for (;;) {
		spin_lock_irqsave(&retry_list_lock, flags);
		if (list_empty(head))
			break;
		r1_bio = list_entry(head->prev, r1bio_t, retry_list);
		list_del(head->prev);
		spin_unlock_irqrestore(&retry_list_lock, flags);

		mddev = r1_bio->mddev;
		conf = mddev_to_conf(mddev);
		bio = r1_bio->master_bio;
		if (test_bit(R1BIO_IsSync, &r1_bio->state)) {
			raid1_sync_write(mddev, r1_bio);
		} else {
			if (map(mddev, &member) == -1) {
				printk(KERN_ALERT "raid1: %s: unrecoverable I/O"
				       " read error for block %llu\n",
				       bdevname(bio->bi_bdev,b),
				       (unsigned long long)r1_bio->sector);
				raid_end_bio_io(r1_bio);
/* XXX What about failing the member??? */
			} else {
				printk(KERN_ERR "raid1: %s: redirecting sector "
				       "%llu to another mirror\n",
				       mdname(member),
				       (unsigned long long)r1_bio->sector);
				bio->bi_sector =
				    r1_bio->sector + member->data_offset;
				bio->bi_rw = READ;
				md_make_request(member, bio);
			}
		}
	}
	spin_unlock_irqrestore(&retry_list_lock, flags);
}


static int raid1_stop(mddev_t *mddev, int ro)
{
	conf_t *conf;
	mdk_member_t *member;
	mdk_metapersonality_t *metapers;
	mdk_sync_support_t *syncinfo;
	int lastphase;

	conf = mddev_to_conf(mddev);
	member = to_member(mddev);
	metapers = member->metapers;
	syncinfo = &conf->syncinfo;

	/*
	 * Cancel any rebuild operations, wait for all writes to
	 * complete, and enter safe mode.  The normal I/O path is
	 * optimized to only do wake-ups when writes cease for the
	 * an old phase.  This avoids wakeups except when doing a
	 * reconstruction.  So, we must artificially toggle the
	 * phase to ensure that both phases are clear of writes.
	 */
	md_cancel_syncops(syncinfo);
	lastphase = syncinfo->phase ^ 1;
	wait_event_lock_irq(syncinfo->sync_wait,
			    (syncinfo->writes_pending[lastphase] == 0),
			    syncinfo->sync_lock);
	lastphase = syncinfo->phase;
	syncinfo->phase ^= 1;
	wait_event_lock_irq(syncinfo->sync_wait,
			    (syncinfo->writes_pending[lastphase] == 0),
			    syncinfo->sync_lock);
	metapers->member_op(member, MDK_MEMBER_OP_ENTER_SAFEMODE, NULL);
	if (ro == 0) {
		if (mddev->thread != NULL) {
			md_teardown_sync_support(mddev, &conf->syncinfo);
			md_unregister_thread(mddev->thread);
		}
		mddev->thread = NULL;
		if (conf->r1bio_pool)
			mempool_destroy(conf->r1bio_pool);
		kfree(conf);
		mddev->member.pers_private = NULL;
	}
	return 0;
}

/*
 * Initialize an mddev that will soon become
 * a RAID1 managed object.
 */
static int
raid1_raid_init(mddev_t *mddev)
{
	mdk_member_t *member;
	conf_t *conf;
	int error;
	
	printk("raid1_raid_init(%p)\n", mddev);
	member = to_member(mddev);
	conf = kmalloc(sizeof(conf_t), GFP_KERNEL);
	if (!conf) {
		printk(KERN_ERR "raid1: couldn't allocate conf memory\n");
		return -ENOMEM;
	}
	memset(conf, 0, sizeof(*conf));
	member->pers_private = conf;
	conf->mddev = mddev;
	conf->last_used = NULL;
	conf->raid_disks = mddev->max_disks;

	conf->r1bio_pool = mempool_create(NR_RAID1_BIOS, r1bio_pool_alloc,
					  r1bio_pool_free, mddev);
	if (!conf->r1bio_pool) {
		printk(KERN_ERR "raid1: couldn't allocate bio pool memory\n");
		kfree(conf);
		return -ENOMEM;
	}

	error = md_init_sync_support(mddev, &conf->syncinfo, NULL);
	if (error != 0) {
		printk(KERN_ERR "raid1: couldn't initialize sync support\n");
		mempool_destroy(conf->r1bio_pool);
		kfree(conf);
		return error;
	}
	conf->syncinfo.window_size = RESYNC_WINDOW >> 9;
	mddev->thread = md_register_thread(raid1d, mddev, NULL, "%s_raid1");
	if (!mddev->thread) {
		printk(KERN_ERR 
			"raid1: couldn't allocate thread for %s\n", 
			mdname(to_member(mddev)));
		md_teardown_sync_support(mddev, &conf->syncinfo);
		mempool_destroy(conf->r1bio_pool);
		kfree(conf);
		return -ENOMEM;
	}
	if ((member->state & MDK_MSTATE_INITIALIZING) != 0)
		raid1_initialize_array(mddev);

	return 0;
}

static int raid1_member_init(mddev_t *mddev, mdk_member_t *member)
{
	request_queue_t *q;
	mconf_t *mconf;

	printk("raid1_member_init(%p) called for object %p\n",
	       mddev, member);
	mconf = kmalloc(sizeof(*mconf), GFP_KERNEL);
	if (mconf == NULL)
		return (-ENOMEM);
	
	mconf->head_position = 0;
	member->parent_pers_private = mconf;

	q = md_member_get_queue(member);
	blk_queue_stack_limits(mddev->queue, q);
	/* as we don't honour merge_bvec_fn, we must never risk
	 * violating it, so limit ->max_sector to one PAGE, as
	 * a one page request is never in violation.
	 */
	spin_lock_irq(&member->member_lock);
	if (q->merge_bvec_fn && mddev->queue->max_sectors > (PAGE_SIZE>>9))
		mddev->queue->max_sectors = (PAGE_SIZE>>9);

	spin_unlock_irq(&member->member_lock);

	print_conf(mddev_to_conf(mddev));
	return 0;
}

/*
 * Returns the number of items added or removed
 * from the sync set.
 */
static int raid1_update_sync_set(mddev_t *mddev, int *padded, int *ptargets)
{
	mdk_member_t *array;
	mdk_member_t *member;
	struct list_head *tmp;
	int changed;
	int added;
	int targets;

	printk("raid1_update_sync_set called on %p\n", mddev);
	array = to_member(mddev);
	changed = 0;
	added = 0;
	targets = 0;
	ITERATE_CHILDREN_LOCKED(array, member, tmp) {
		mconf_t *mconf;
		int	old_sync;

		mconf = member_to_mconf(member);
		old_sync = mconf->sync_member;
		/*
		 * All writable members can participate
		 * in a sync operation.  We don't care
		 * here about which one will be the source
		 * for a rebuild.
		 */
		if (R1_WRITABLE_MEMBER(member)) {
			mconf->sync_member = 1;
			if (member->state & MDK_MSTATE_REBUILD_TARGET)
				targets++;
		} else
			mconf->sync_member = 0;
		if (mconf->sync_member != old_sync) {
			changed++;
			if (mconf->sync_member == 1)
				added++;
		}
	}
	printk("raid1_update_sync_set %p: changed %d added %d targets %d\n",
	       mddev, changed, added, targets);

	if (padded != NULL)
		*padded = added;
	if (ptargets != NULL)
		*ptargets = targets;
	return (changed);
}

static int raid1_run(mddev_t *mddev)
{
	conf_t *conf;
	mdk_member_t *member;

	conf = mddev_to_conf(mddev);
	member = to_member(mddev);
	md_update_thread_string(mddev->thread, "%s_raid1");
	md_update_sync_support_string(&conf->syncinfo, NULL);
	if ((member->state & MDK_MSTATE_DEGRADED) != 0)
		raid1_attempt_rebuild(mddev);
	return 0;
}

/*
 * Initialize all writable members of an array by
 * writing zeros to them.
 */
static void raid1_initialize_array(mddev_t *mddev)
{
	mdk_member_t *array;
	conf_t *conf;
	mdk_workitem_t *workitem;
	u_long flags;
	int error;

	array = to_member(mddev);
	conf = mddev_to_conf(mddev);

	printk("raid1_initialize array on %p\n", mddev);

	/*
	 * We can begin an initialization only if the array
	 * is degraded or optimal (e.g. writable).  Don't
	 * bother restarting an init operation if the sync
	 * set has not changed and we have an initializtion
	 * already in progress.
	 */
	flags = md_find_workitem(&conf->syncinfo.work_state,
				 MD_SYNCOP_INIT, NULL);
	spin_lock_irqsave(&array->member_lock, flags);
	if (!R1_INITIALIZABLE(array)
	 || (raid1_update_sync_set(mddev, NULL, NULL) == 0
	  && flags & (MD_WF_QUEUED|MD_WF_RUNNING))) {
		spin_unlock_irqrestore(&array->member_lock, flags);
		return;
	}
	spin_unlock_irqrestore(&array->member_lock, flags);

	workitem = kmalloc(sizeof(*workitem), GFP_KERNEL);
	if (workitem == NULL) {
		printk(KERN_ERR "raid1_intialize_array(): "
		       "Cannot allocate workitem\n");
		return;
	}
	md_init_workitem(workitem, MD_SYNCOP_INIT,
			 raid1_workop, md_sync_handler, &conf->syncinfo);
	error = md_queue_workitem(&conf->syncinfo.work_state,
				  workitem, /*head_of_queue*/1);
	if (error != 0) {
		kfree(workitem);
		printk(KERN_ERR "raid1_initialize_array(): "
		       "Cannot queue workitem\n");
	}
}

/*
 * Initialize all writable members of an array by
 * writing zeros to them.
 */
static void raid1_attempt_rebuild(mddev_t *mddev)
{
	mdk_member_t *array;
	conf_t *conf;
	mdk_workitem_t *workitem;
	u_long flags;
	int added;
	int targets;
	int changed;
	int error;

	conf = mddev_to_conf(mddev);
	array = to_member(mddev);

	printk("raid1_attempt_rebuild on %p\n", mddev);
	/*
	 * We can begin a rebuild only if the array is degraded
	 * and we can add a spare, or there is an active rebuild
	 * target that needs to be synced.  Don't bother restarting
	 * a rebuild operation if the sync set has not changed and
	 * we have a rebuild already in progress.
	 */
	flags = md_find_workitem(&conf->syncinfo.work_state,
				 MD_SYNCOP_REBUILD, NULL);
	spin_lock_irqsave(&array->member_lock, flags);
	changed = raid1_update_sync_set(mddev, &added, &targets);
	if (!R1_REBUILDABLE(array)
	 || ((changed == 0 || added == 0)
	  && flags & (MD_WF_QUEUED|MD_WF_RUNNING))) {
		spin_unlock_irqrestore(&array->member_lock, flags);
		return;
	}
	spin_unlock_irqrestore(&array->member_lock, flags);

	workitem = kmalloc(sizeof(*workitem), GFP_KERNEL);
	if (workitem == NULL) {
		printk(KERN_ERR "raid1_attempt_rebuild(): "
		       "Cannot allocate workitem\n");
		return;
	}
	md_init_workitem(workitem, MD_SYNCOP_REBUILD,
			 raid1_workop, md_sync_handler, &conf->syncinfo);
	error = md_queue_workitem(&conf->syncinfo.work_state,
				  workitem, /*head_of_queue*/1);
	if (error != 0) {
		kfree(workitem);
		printk(KERN_ERR "raid1_attempt_rebuild(): "
		       "Cannot queue workitem\n");
	}
}

static int raid1_state_change(mddev_t *mddev, mdk_member_t *member)
{
	conf_t *conf;
	mdk_member_t *array;
	mdk_mstate_t new_state;
	u_long flags;

	conf = mddev_to_conf(mddev);
	array = to_member(mddev);
	printk("raid1_state_change called on %p with state %#x\n",
	       mddev, member->state);

	/*
	 * Use the workitem support framework's client
	 * semaphore to single step configuration changes.
	 * This is convenient since many state changes
	 * include the creation or removal of workitems.
	 */
	down(&conf->syncinfo.work_state.work_config_sem);

	/*
	 * Snapshot the current state so that we do
	 * not have to hold the member_lock while
	 * we decide our course of action.  Any state
	 * change that occurs during our handling of
	 * the last state will block on the semaphore
	 * above.
	 */
	spin_lock_irqsave(&member->member_lock, flags);
	new_state = member->state;
	spin_unlock_irqrestore(&member->member_lock, flags);

	if (member == array) {
		/*
		 * Handle status changes that have propogated
		 * to our array object.
		 */
		switch (new_state & ~MDK_MSTATE_MODIFIER_MASK) {
		case MDK_MSTATE_OPTIMAL:
		case MDK_MSTATE_DEGRADED:
			if ((new_state & MDK_MSTATE_INITIALIZING) != 0) {
				raid1_initialize_array(mddev);
				break;
			}
			md_runnable(mddev);
			if ((new_state & MDK_MSTATE_DEGRADED) != 0)
				raid1_attempt_rebuild(mddev);
			break;
		case MDK_MSTATE_FAILED:
			md_cancel_syncops(&conf->syncinfo);
			break;
		default:
			break;
		}
	} else if ((new_state & MDK_MSTATE_REBUILD_TARGET) != 0) {
		mdk_mstate_t parent_state;

		/*
		 * Handle status changes that have occurred in a
		 * child object that is either eligible for rebuild
		 * or we have been rebuilding to in the past.  There
		 * are three cases: a spare we are rebuilding to transitions
		 * from optimal to degraded, a spare we are rebuilding to
		 * transitions from optimal or degraded to failed, or a
		 * spare that was failed has transitioned to either
		 * optimal or degraded.  The first is a noop.  The latter
		 * two require our rebuild operation to be restarted.
		 * The first is requires no action, but this is detected
		 * by the rebuild or initialization routine seeing that
		 * all members that need to be synced are already part
		 * of the active sync set.
		 */
		spin_lock_irqsave(&array->member_lock, flags);
		parent_state = array->state;
		spin_unlock_irqrestore(&array->member_lock, flags);
		if ((parent_state & MDK_MSTATE_INITIALIZING) != 0)
			raid1_initialize_array(mddev);
		else
			raid1_attempt_rebuild(mddev);
	}
	up(&conf->syncinfo.work_state.work_config_sem);
	return 0;
}

int raid1_add_spare(mddev_t *mddev)
{
	conf_t *conf;
	mdk_member_t *array;
	mdk_mstate_t state;
	u_long flags;

	conf = mddev_to_conf(mddev);
	array = to_member(mddev);
	printk("raid1_add_spare called on %p\n", mddev);

	/*
	 * Use the workitem support framework's client
	 * semaphore to single step configuration changes.
	 * This is convenient since many state changes
	 * include the creation or removal of workitems.
	 */
	down(&conf->syncinfo.work_state.work_config_sem);
	spin_lock_irqsave(&array->member_lock, flags);
	state = array->state;
	spin_unlock_irqrestore(&array->member_lock, flags);
	if ((state & MDK_MSTATE_INITIALIZING) != 0)
		raid1_initialize_array(mddev);
	else
		raid1_attempt_rebuild(mddev);
	up(&conf->syncinfo.work_state.work_config_sem);
	return 0;
}

static int raid1_open_workitem(mddev_t *mddev, mdk_workitem_t *workitem)
{
	mdk_member_t *array;
	mdk_mstate_t array_state;
	conf_t *conf;
	u_long flags;
	int added;
	int targets;
	int changed;
	int buffs;

	/*
	 * If this is a rebuild request and we have no
	 * target devices to rebuild to, try allocating
	 * a spare.  If the spare allocation fails, fail
	 * the workitem.
	 */
	array = to_member(mddev);
	spin_lock_irqsave(&array->member_lock, flags);
	changed = raid1_update_sync_set(mddev, &added, &targets);
	array_state = array->state;
	spin_unlock_irqrestore(&array->member_lock, flags);
	if (targets == 0
	 && (array_state & MDK_MSTATE_DEGRADED) != 0) {
		mdk_metapersonality_t *metapers;

		metapers = array->metapers;
		metapers->member_op(array, MDK_MEMBER_OP_ALLOC_SPARE, NULL);
		spin_lock_irqsave(&array->member_lock, flags);
		changed = raid1_update_sync_set(mddev, &added, &targets);
		spin_unlock_irqrestore(&array->member_lock, flags);
		if (targets == 0)
			return -EIO;
	}

	conf = mddev_to_conf(mddev);
	buffs = 2 * RESYNC_WINDOW / RESYNC_BLOCK_SIZE;
	if (conf->r1buf_pool)
		BUG();
	conf->r1buf_pool = mempool_create(buffs, r1buf_pool_alloc,
					  r1buf_pool_free, conf);
	if (!conf->r1buf_pool)
		return -ENOMEM;
	return 0;
}

static int raid1_close_workitem(mddev_t *mddev, mdk_workitem_t *workitem)
{
	conf_t *conf;

	conf = mddev_to_conf(mddev);
	if (conf->r1buf_pool != NULL) {
		mempool_destroy(conf->r1buf_pool);
		conf->r1buf_pool = NULL;
	}
	return 0;
}

static int raid1_free_workitem(mddev_t *mddev, mdk_workitem_t *workitem)
{
	kfree(workitem);
	return 0;
}

static int raid1_workop(mdk_member_t *member, int op, void *arg)
{
	switch (op) {
	case MDK_WORKITEM_OP_OPEN:
		return raid1_open_workitem(to_mddev(member),
					   (mdk_workitem_t *)arg);
		break;
	case MDK_WORKITEM_OP_CLOSE:
		return raid1_close_workitem(to_mddev(member),
					    (mdk_workitem_t *)arg);
		break;
	case MDK_WORKITEM_OP_FREE:
		return raid1_free_workitem(to_mddev(member),
					   (mdk_workitem_t *)arg);
	default:
		return (-EINVAL);
	}
}

static int raid1_raidop(mdk_member_t *member, int op, void *arg)
{

	switch (op) {
	case MDK_RAID_OP_INIT:
		if (member->type != MDK_MTYPE_MDDEV)
			return (0);
		return raid1_raid_init(to_mddev(member));
	case MDK_RAID_OP_INIT_MEMBER:
	{
		mdk_member_t *child;

		child = (mdk_member_t *)arg;
		return raid1_member_init(to_mddev(member), child);
	}
	case MDK_RAID_OP_MSTATE_CHANGED:
		return raid1_state_change(to_mddev(member),
					  (mdk_member_t *)arg);
	case MDK_RAID_OP_SPARE_ADDED:
		return raid1_add_spare(to_mddev(member));
	case MDK_RAID_OP_SYNC_REQUEST:
		return raid1_sync_request(to_mddev(member),
					  (mdk_workitem_t *)arg);
	case MDK_RAID_OP_RUN: 
		return raid1_run(to_mddev(member));
	case MDK_RAID_OP_RUNNABLE:
	{
		if (MDK_MSTATE_READABLE(member->state)) 
			return 0;
		return -ENODEV;
	}
	case MDK_RAID_OP_STOP:
		return raid1_stop(to_mddev(member), (int)arg);
	case MDK_RAID_OP_ENTER_SAFEMODE:
	{
		mdk_metapersonality_t *metapers;

		metapers = member->metapers;
		return (metapers->member_op(member,
					    MDK_MEMBER_OP_ENTER_SAFEMODE,
					    NULL));
	}
	case MDK_RAID_OP_EXIT_SAFEMODE:
	{
		mdk_metapersonality_t *metapers;

		metapers = member->metapers;
		return (metapers->member_op(member,
					    MDK_MEMBER_OP_EXIT_SAFEMODE,
					    NULL));
	}
	case MDK_RAID_OP_REQUEST_VERIFY:
	case MDK_RAID_OP_REQUEST_RESYNC:
	case MDK_RAID_OP_PAUSE_MEMBER:
	case MDK_RAID_OP_ADD_MEMBER:
	case MDK_RAID_OP_REMOVE_MEMBER:
	default:
		return (-EINVAL);
	}

	return 0;
}

static mdk_personality_t raid1_personality =
{
	.name		= "raid1",
	.owner		= THIS_MODULE,
	.make_request	= make_request,
	.raid_op	= raid1_raidop,
};

static int __init raid_init(void)
{
	return register_md_personality(RAID1, &raid1_personality);
}

static void raid_exit(void)
{
	unregister_md_personality(RAID1);
}

module_init(raid_init);
module_exit(raid_exit);
MODULE_LICENSE("GPL");
MODULE_ALIAS("md-personality-3"); /* RAID1 */
