/*
   md.h : Multiple Devices driver for Linux
          Copyright (C) 1996-98 Ingo Molnar, Gadi Oxman
          Copyright (C) 1994-96 Marc ZYNGIER
	  <zyngier@ufr-info-p7.ibp.fr> or
	  <maz@gloups.fdn.fr>
	  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
   
   You should have received a copy of the GNU General Public License
   (for example /usr/src/linux/COPYING); if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  
*/

#ifndef _MD_H
#define _MD_H

#include <linux/blkdev.h>
#include <asm/semaphore.h>
#include <linux/major.h>
#include <linux/ioctl.h>
#include <linux/types.h>
#include <linux/bitops.h>
#include <linux/module.h>
#include <linux/hdreg.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/smp_lock.h>
#include <linux/delay.h>
#include <net/checksum.h>
#include <linux/random.h>
#include <linux/kernel_stat.h>
#include <asm/io.h>
#include <linux/completion.h>
#include <linux/mempool.h>
#include <linux/list.h>
#include <linux/reboot.h>
#include <linux/vmalloc.h>
#include <linux/blkpg.h>
#include <linux/bio.h>

/*
 * 'md_p.h' holds the 'physical' layout of RAID devices
 * 'md_u.h' holds the user <=> kernel API
 *
 * 'md_k.h' holds kernel internal definitions
 */

#include <linux/raid/md_p.h>
#include <linux/raid/md_k.h>
#include <linux/raid/md_u.h>

/*
 * Different major versions are not compatible.
 * Different minor versions are only downward compatible.
 * Different patchlevel versions are downward and upward compatible.
 */
#define MD_MAJOR_VERSION                0
#define MD_MINOR_VERSION                90
#define MD_PATCHLEVEL_VERSION           0

extern int register_md_personality (int p_num, mdk_personality_t *p);
extern int unregister_md_personality (int p_num);
extern int register_md_metapersonality (mdk_metapersonality_t *p);
extern int unregister_md_metapersonality (mdk_metapersonality_t *p);
extern mdk_thread_t * md_register_thread (mdk_thread_handler_t *run,
				mddev_t *mddev, void *arg, const char *name);
extern void md_unregister_thread (mdk_thread_t *thread);
extern void md_update_thread_string(mdk_thread_t *thread, const char *fmt);
extern void md_wakeup_thread(mdk_thread_t *thread);
extern void md_check_recovery(mddev_t *mddev);
extern void md_interrupt_thread (mdk_thread_t *thread);
extern void md_write_start(mddev_t *mddev);
extern void md_write_end(mddev_t *mddev);
extern void md_handle_safemode(mddev_t *mddev);
extern void md_done_sync(mddev_t *mddev, int blocks, int ok);
extern void md_sync_acct(mdk_member_t *member, unsigned long nr_sectors);

extern void md_print_devices (void);
extern int md_read_blocks (struct block_device *bdev, sector_t start_sector,
			   int num_sectors, uint8_t *dest_buf);
extern int md_write_blocks (struct block_device *bdev, sector_t start_sector,
			    int num_sectors, uint8_t *src_buf);
extern mdk_rdev_t *md_rdev_alloc(mdk_member_t *parent, mdk_msubtype_t subtype,
				 struct block_device *newdev, int priority);
extern mddev_t *md_mddev_alloc(mdk_member_t *parent, mdk_msubtype_t subtype,
			       int priority);
extern void md_member_add_child(mdk_member_t *child, mdk_member_t *parent);
extern void md_member_orphan(mdk_member_t *member);
extern mdk_personality_t *md_find_pers(int level);
extern struct block_device *md_get_blkdev(dev_t dev);
extern void md_release_blkdev(struct block_device *blkdev);
extern void kick_rdev_from_array(mdk_rdev_t * rdev);
extern int md_check_chunksize(mddev_t *mddev);
extern int md_runnable(mddev_t *mddev);
extern int md_init_workitem_support(mddev_t *mddev,
				    mdk_workitem_support_t *,
				    const char *);
extern void md_teardown_workitem_support(mdk_workitem_support_t *workinfo);
extern int md_init_workitem(mdk_workitem_t *work, u_int opcode,
			    mdk_member_callback_t *control,
			    mdk_work_handler_t *handler, void *data);

extern void md_cancel_workitems(mdk_workitem_support_t *workinfo);
extern void md_wait_workitem(mdk_workitem_support_t *, mdk_workitem_t *);
extern mdk_workitem_flags_t md_find_workitem(mdk_workitem_support_t *workinfo,
					     u_int opcode, int *found);
extern int md_queue_workitem(mdk_workitem_support_t *workinfo,
			     mdk_workitem_t *work, int head_of_queue);
extern void md_exit_safemode(mddev_t *mddev, mdk_sync_support_t *syncinfo);
extern int md_init_sync_support(mddev_t *mddev, mdk_sync_support_t *syncinfo,
				const char *thread_string);
extern void md_teardown_sync_support(mddev_t *mddev,
				     mdk_sync_support_t *syncinfo);
extern mdk_work_handler_t md_sync_handler;

static inline void md_cancel_syncops(mdk_sync_support_t *syncinfo)
{
	md_cancel_workitems(&syncinfo->work_state);
}

static inline mdk_member_t *member_get(mdk_member_t *member)
{
        atomic_inc(&member->refcount);
	return (member);
}

static inline void member_put(mdk_member_t *member)
{
	if (!atomic_dec_and_test(&member->refcount))
		return;
	member->destructor(member);
}

static inline void member_reparent(mdk_member_t *member,
				   mdk_member_t *new_parent)
{
	md_member_orphan(member);

        if (new_parent)
		md_member_add_child(new_parent, member);
}

static inline void md_make_request(mdk_member_t *member, struct bio *bio)
{
	if (member->type == MDK_MTYPE_RDEV) {
		bio->bi_bdev = to_rdev(member)->bdev;
		generic_make_request(bio);
	} else {
		mddev_t *mddev;

		mddev = to_mddev(member);
		(void)member->pers->make_request(mddev->queue, bio);
	}
}

static inline void md_update_work_support_string(mdk_workitem_support_t *work,
						 const char *fmt)
{
	if (fmt == NULL)
		fmt = "%s_work";
	md_update_thread_string(work->work_thread, fmt);
}

static inline void md_update_sync_support_string(mdk_sync_support_t *syncinfo,
						 const char *fmt)
{
	md_update_work_support_string(&syncinfo->work_state, fmt);
}

static inline request_queue_t *md_member_get_queue(mdk_member_t *member)
{
	if (member->type == MDK_MTYPE_RDEV)
		return to_rdev(member)->bdev->bd_disk->queue;
	else
		return to_mddev(member)->queue;
}

#define MD_BUG(x...) { printk("md: bug in file %s, line %d\n", __FILE__, __LINE__); md_print_devices(); }

static inline void md_clear_safemode_timer(mdk_sync_support_t *syncinfo)
{
	if (syncinfo->safemode == MDK_SAFEMODE_IDLE_WAIT) {
		del_timer(&syncinfo->safemode_timer);
		syncinfo->safemode = MDK_SAFEMODE_OFF;
	}
}

static inline void md_update_safemode_timer(mdk_sync_support_t *syncinfo)
{
	mod_timer(&syncinfo->safemode_timer,
		  jiffies + syncinfo->safemode_delay);
	syncinfo->safemode = MDK_SAFEMODE_IDLE_WAIT;
}

/*
 * Called before submitting normal I/O to a an array that
 * supports synchronization operations.  Blocks any write
 * requests that are inside the current sync lock range.
 * Returns the phase that should be assigned to this bio.
 *
 * XXX Prehaps too big to be inlined???
 */
static inline int syncable_vol_io_start(mdk_sync_support_t *syncinfo,
					struct bio *bio, int *pbelow_sync_mark)
{
	sector_t bio_end;
	int phase;
	int below_sync_mark;

	bio_end = bio->bi_sector + (bio->bi_size >> 9);
	spin_lock_irq(&syncinfo->sync_lock);
	if (bio_data_dir(bio) != WRITE) {
		/*
		 * In the write case, the phase is not used.
		 * Return a value way out of range to catch
		 * improper use of this value.
		 */
		atomic_inc(&syncinfo->reads_pending);
		phase = INT_MAX;
	} else {
		wait_event_lock_irq(syncinfo->io_wait,
			    (bio->bi_sector >= syncinfo->lock_range.end
			  || bio_end < syncinfo->lock_range.start
			  || syncinfo->safemode == MDK_SAFEMODE_ACTIVE),
			    syncinfo->sync_lock);
		phase = syncinfo->phase;
		if (syncinfo->safemode == MDK_SAFEMODE_ACTIVE) {
			mddev_t *mddev;

			mddev = syncinfo->work_state.work_thread->mddev;
			md_exit_safemode(mddev, syncinfo);
		} else
			md_clear_safemode_timer(syncinfo);
		syncinfo->writes_pending[phase]++;
	}
	if (bio_end < syncinfo->lock_range.start)
		below_sync_mark = 1;
	else
		below_sync_mark = 0;
	spin_unlock_irq(&syncinfo->sync_lock);
	if (pbelow_sync_mark)
		*pbelow_sync_mark = below_sync_mark;
	return phase;
}

/*
 * Called on completion of normal I/O to a an array that
 * supports synchronization operations.  Notifies any operations
 * sleeping waiting for the current lock_range to clear of
 * pending requests.
 */
static inline void syncable_vol_io_complete(mdk_sync_support_t *syncinfo,
					    int rw, u_int bio_phase)
{
	if (rw != WRITE) {
		atomic_dec(&syncinfo->reads_pending);
		return;
	}
	spin_lock_irq(&syncinfo->sync_lock);
	syncinfo->writes_pending[bio_phase]--;
	if (syncinfo->writes_pending[bio_phase] == 0) {
		if (bio_phase != syncinfo->phase)
			wake_up(&syncinfo->sync_wait);
		if (syncinfo->writes_pending[syncinfo->phase] == 0)
			md_update_safemode_timer(syncinfo);
	}
	spin_unlock_irq(&syncinfo->sync_lock);
}

/*
 * Called when completing a sync I/O request.  If this is the
 * last I/O of a sync range, shrink the lock_range by the
 * sectors that are now synced out and wake any threads blocked
 * on that range.
 */
static inline int syncable_vol_syncio_start(mdk_sync_support_t *syncinfo)
{
	int phase;

	spin_lock_irq(&syncinfo->sync_lock);
	phase = syncinfo->phase;
	syncinfo->sync_pending[phase]++;
	spin_unlock_irq(&syncinfo->sync_lock);

	return phase;
}

/*
 * Called when completing a sync I/O request.  If this is the
 * last I/O of a sync range, shrink the lock_range by the
 * sectors that are now synced out and wake any threads blocked
 * on that range.
 */
static inline void syncable_vol_syncio_complete(mdk_sync_support_t *syncinfo,
						u_int phase)
{
	md_sector_range_t *sync_range;

	sync_range = &syncinfo->sync_range[phase];
	spin_lock_irq(&syncinfo->sync_lock);
	syncinfo->sync_pending[phase]--;
	if (syncinfo->sync_pending[phase] == 0
	 && (sync_range->start == sync_range->end)) {
		if (syncinfo->lock_range.start < sync_range->end)
			syncinfo->lock_range.start = sync_range->end;
		wake_up(&syncinfo->io_wait);
	} 
	spin_unlock_irq(&syncinfo->sync_lock);
}

/*
 * Unconditionally set a member's
 * state to the given state.
 */
static inline void md_set_member_state(mdk_member_t *member, mdk_mstate_t state)
{
	mdk_set_state_arg_t sstate;
	u_long flags;
	
	spin_lock_irqsave(&member->member_lock, flags);
	sstate.oldstate = member->state;
	sstate.newstate = state;
	spin_unlock_irqrestore(&member->member_lock, flags);
	member->metapers->member_op(member, MDK_MEMBER_OP_SET_STATE, &sstate);
}
#endif 
