/*
   md_k.h : kernel internal structure of the Linux MD driver
          Copyright (C) 1996-98 Ingo Molnar, Gadi Oxman
	  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
   
   You should have received a copy of the GNU General Public License
   (for example /usr/src/linux/COPYING); if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  
*/

#ifndef _MD_K_H
#define _MD_K_H

#define MD_RESERVED       0UL
#define LINEAR            1UL
#define RAID0             2UL
#define RAID1             3UL
#define RAID5             4UL
#define TRANSLUCENT       5UL
#define HSM               6UL
#define MULTIPATH         7UL
#define RAID6		  8UL
#define MAX_PERSONALITY   9UL

#define	LEVEL_MULTIPATH		(-4)
#define	LEVEL_LINEAR		(-1)

#define MaxSector (~(sector_t)0)
#define MD_THREAD_NAME_MAX 14

#define __wait_event_lock_irq(wq, condition, lock) 			\
do {									\
	wait_queue_t __wait;						\
	init_waitqueue_entry(&__wait, current);				\
									\
	add_wait_queue(&wq, &__wait);					\
	for (;;) {							\
		set_current_state(TASK_UNINTERRUPTIBLE);		\
		if (condition)						\
			break;						\
		spin_unlock_irq(&lock);					\
		blk_run_queues();					\
		schedule();						\
		spin_lock_irq(&lock);					\
	}								\
	current->state = TASK_RUNNING;					\
	remove_wait_queue(&wq, &__wait);				\
} while (0)

#define wait_event_lock_irq(wq, condition, lock) 			\
do {									\
	if (condition)	 						\
		break;							\
	__wait_event_lock_irq(wq, condition, lock);			\
} while (0)


#define __wait_disk_event(wq, condition) 				\
do {									\
	wait_queue_t __wait;						\
	init_waitqueue_entry(&__wait, current);				\
									\
	add_wait_queue(&wq, &__wait);					\
	for (;;) {							\
		set_current_state(TASK_UNINTERRUPTIBLE);		\
		if (condition)						\
			break;						\
		blk_run_queues();					\
		schedule();						\
	}								\
	current->state = TASK_RUNNING;					\
	remove_wait_queue(&wq, &__wait);				\
} while (0)

#define wait_disk_event(wq, condition) 					\
do {									\
	if (condition)	 						\
		break;							\
	__wait_disk_event(wq, condition);				\
} while (0)

static inline int pers_to_level (int pers)
{
	switch (pers) {
		case MULTIPATH:		return LEVEL_MULTIPATH;
		case HSM:		return -3;
		case TRANSLUCENT:	return -2;
		case LINEAR:		return LEVEL_LINEAR;
		case RAID0:		return 0;
		case RAID1:		return 1;
		case RAID5:		return 5;
		case RAID6:		return 6;
	}
	BUG();
	return MD_RESERVED;
}

static inline int level_to_pers (int level)
{
	switch (level) {
		case LEVEL_MULTIPATH: return MULTIPATH;
		case -3: return HSM;
		case -2: return TRANSLUCENT;
		case LEVEL_LINEAR: return LINEAR;
		case 0: return RAID0;
		case 1: return RAID1;
		case 4:
		case 5: return RAID5;
		case 6: return RAID6;
	}
	return MD_RESERVED;
}

typedef struct mddev_s mddev_t;
typedef struct mdk_rdev_s mdk_rdev_t;
typedef struct mdk_personality_s mdk_personality_t;
typedef struct mdk_member_s mdk_member_t;
typedef struct mdk_metapersonality_s mdk_metapersonality_t;
typedef void (mdk_thread_handler_t) (mddev_t *, void *);
typedef int (mdk_member_callback_t) (mdk_member_t *, int, void *);
#define to_mddev(obj) container_of((obj), mddev_t, member)
#define to_rdev(obj) container_of(obj, mdk_rdev_t, member)
#define to_member(obj) (&obj->member)

typedef struct mdk_thread_s {
	mdk_thread_handler_t	*run;
	void			*arg;
	mddev_t			*mddev;
	wait_queue_head_t	wqueue;
	unsigned long           flags;
	struct completion	*event;
	struct task_struct	*tsk;
	const char		*name;
} mdk_thread_t;

#define THREAD_WAKEUP  0

#define MAX_MD_DEVS  256	/* Max number of md dev */

/*
 * options passed in raidrun:
 */

#define MAX_CHUNK_SIZE (4096*1024)

/*
 * default readahead
 */

static inline int disk_faulty(mdp_disk_t * d)
{
	return d->state & (1 << MD_DISK_FAULTY);
}

static inline int disk_active(mdp_disk_t * d)
{
	return d->state & (1 << MD_DISK_ACTIVE);
}

static inline int disk_sync(mdp_disk_t * d)
{
	return d->state & (1 << MD_DISK_SYNC);
}

static inline int disk_spare(mdp_disk_t * d)
{
	return !disk_sync(d) && !disk_active(d) && !disk_faulty(d);
}

static inline int disk_removed(mdp_disk_t * d)
{
	return d->state & (1 << MD_DISK_REMOVED);
}

static inline void mark_disk_faulty(mdp_disk_t * d)
{
	d->state |= (1 << MD_DISK_FAULTY);
}

static inline void mark_disk_active(mdp_disk_t * d)
{
	d->state |= (1 << MD_DISK_ACTIVE);
}

static inline void mark_disk_sync(mdp_disk_t * d)
{
	d->state |= (1 << MD_DISK_SYNC);
}

static inline void mark_disk_spare(mdp_disk_t * d)
{
	d->state = 0;
}

static inline void mark_disk_removed(mdp_disk_t * d)
{
	d->state = (1 << MD_DISK_FAULTY) | (1 << MD_DISK_REMOVED);
}

static inline void mark_disk_inactive(mdp_disk_t * d)
{
	d->state &= ~(1 << MD_DISK_ACTIVE);
}

static inline void mark_disk_nonsync(mdp_disk_t * d)
{
	d->state &= ~(1 << MD_DISK_SYNC);
}

enum mdk_mtype_e
{
	MDK_MTYPE_NONE,
	MDK_MTYPE_RDEV,
	MDK_MTYPE_MDDEV,
	MDK_MTYPE_MASTER
};
typedef enum mdk_mtype_e mdk_mtype_t;

enum mdk_msubtype_e
{
	MDK_MSUBTYPE_NONE,
	MDK_MSUBTYPE_TOPLEVEL,
	MDK_MSUBTYPE_CHILD,
	MDK_MSUBTYPE_DEDICATED_SPARE,
	MDK_MSUBTYPE_POOL_SPARE,
	MDK_MSUBTYPE_ORPHAN,
	MDK_MSUBTYPE_AVAILABLE
};
typedef enum mdk_msubtype_e mdk_msubtype_t;

/*
 * The states are defined as independant bits
 * to facilitate checking for multiple states
 * with a simple mask.
 */
enum mdk_mstate_e
{
	MDK_MSTATE_UNKNOWN	  = 0x00,
	MDK_MSTATE_OPTIMAL	  = 0x01,
	MDK_MSTATE_DEGRADED	  = 0x02,
	MDK_MSTATE_FAILED	  = 0x04,/* No increments to refcount allowed.*/
	MDK_MSTATE_TEARDOWN	  = 0x08,
	MDK_MSTATE_MODIFIER_MASK  = 0xF0,
	MDK_MSTATE_REBUILD_TARGET = 0x10,
	MDK_MSTATE_INITIALIZING	  = 0x20
};
typedef enum mdk_mstate_e mdk_mstate_t;

/*
 * A disk is readable if it is optimal or degraded.
 */
#define MDK_MSTATE_READABLE(state)					\
    (((state) & ~(MDK_MSTATE_OPTIMAL|MDK_MSTATE_DEGRADED)) == 0)
#define MDK_READABLE_MEMBER(member) MDK_MSTATE_READABLE((member)->state)
  
/*
 * A disk it writable if it is readable or a rebuild target.
 */
#define MDK_MSTATE_WRITABLE(state)					\
    (((state) &								\
     ~(MDK_MSTATE_OPTIMAL|MDK_MSTATE_DEGRADED|MDK_MSTATE_REBUILD_TARGET)) == 0)
#define MDK_WRITABLE_MEMBER(member) MDK_MSTATE_WRITABLE((member)->state)

#define MDK_PRIORITY_UNKNOWN -1

struct mdk_member_s
{
	spinlock_t		member_lock;	/* Protection of all fields */
	mdk_mtype_t		type;
	mdk_msubtype_t		subtype;
	mdk_mstate_t		state;

	u_int			flags;

	sector_t		total_size;
	sector_t		data_size;
	sector_t		data_offset;
	sector_t		sb_size;
	sector_t		sb_offset;

	sector_t		head_position;

	mdk_member_t		*parent;
	struct list_head	child_set;	/* Children of the member */
	u_int			child_set_gencount;
	struct list_head	peer_links;	/* member of the same parent */
	mdk_personality_t	*pers;
	mdk_metapersonality_t	*metapers;

	void			*metadata_private;
	void			*pers_private;
	void			*parent_pers_private;

	int			priority;
	atomic_t		refcount;
	void			(*destructor)(mdk_member_t *);
	char			name[BDEVNAME_SIZE];
};

extern mdk_member_t toplevel_arrays;

/*
 * MD's 'extended' device
 */
struct mdk_rdev_s
{
	mdk_member_t	member;
	struct block_device *bdev;	/* block device handle */
	int		preferred_minor;	/* autorun support */
};

typedef struct mdk_workitem_s mdk_workitem_t;

#define MD_SYNCOP_REBUILD		0x10
#define MD_SYNCOP_INIT			0x20
#define MD_SYNCOP_VERIFY		0x30
#define MD_SYNCOP_VERIFY_FIX		0x31
#define MD_SYNCOP_ENTER_SAFEMODE	0x40

enum mdk_workitem_flags_e {
	MD_WF_NONE		= 0x00,
	MD_WF_QUEUED		= 0x01,
	MD_WF_RUNNING		= 0x02,
	MD_WF_CANCELED		= 0x04
};

typedef enum mdk_workitem_flags_e mdk_workitem_flags_t;

typedef int mdk_work_handler_t(mddev_t *, mdk_workitem_t *);
        
struct mdk_workitem_s {
	u_int			work_opcode;
	mdk_workitem_flags_t	work_flags;
	mdk_member_callback_t	*work_control;
	mdk_work_handler_t	*work_handler;
	void			*work_data;
	struct list_head	links;
};

/*
 * Opcodes for the work_control handler.
 */
enum mdk_woritem_op_e {
	MDK_WORKITEM_OP_OPEN,
	MDK_WORKITEM_OP_CLOSE,
	MDK_WORKITEM_OP_FREE,
	MDK_WORKITEM_OP_LAST
};
typedef enum mdk_workitem_op_e mdk_workitem_op_t;

/*
 * Structure embedded in RAID personality's private
 * data structure when using the generic code's workitem
 * framework.  The work_config_sem provides a convenient
 * syncronization point for competing attempts to respond
 * to changing array state, build, and schedule a workitem.
 * It is intended for clients of the workitem infrastructure.
 * The core workitem code does not use it.
 */
struct mdk_workitem_support_s {
	mdk_thread_t		*work_thread;
	spinlock_t		 work_state_lock;
	struct list_head	 work_items;
	struct semaphore	 work_config_sem;
	wait_queue_head_t	 work_wait;
};

typedef struct mdk_workitem_support_s mdk_workitem_support_t;

struct md_sector_range_s {
	sector_t	start;	/* inclusive */
	sector_t	end;	/* exclusive */
};

typedef struct md_sector_range_s md_sector_range_t;

/*
 * Structure embedded in RAID personality's private
 * data structure when using the generic sync framework.
 *
 * During resync operations, we must prevent writes
 * from conflicting with a resync.  We do this by declaring
 * a range of sectors (lock_range) "offlimits" to write
 * requests, allowing the resync operation to work safely
 * in that area.  The complications to this algorithm occur
 * in two places:
 *
 * 1) Changing the "lock_range" of sectors used to prevent
 *    normal I/O on the range.
 *
 *    When a lock_range is grown, there may be normal I/O
 *    requests outstanding in the new range.  To simplify
 *    accounting for these requests, we assume that all
 *    outstanding I/O is within this exclusion range.  Once
 *    these transactions complete, it is safe to schedule
 *    resync operations within the new bounds of the lock
 *    range.
 *
 *    Of course, not all pending transactions or future
 *    transactions are for sectors within the locked region.
 *    To allow this non-conflicting I/O to continue even
 *    while we are waiting for pending transactions to complete,
 *    we split our transaction acounting into two values:
 *    those scheduled before the lock_range was grown, and
 *    those scheduled afterwards.  When growing the lock range,
 *    we also toggle the "phase" so that new requests
 *    use an alternate "writes_pending" counter.  Once the counter
 *    of the oposite phase goes to 0, the sectors in the newly
 *    grown range are safe to access.
 *
 *    For this scheme to work, the lock_range can only be grown
 *    if the pending count in the "non-current" phase is 0.  In
 *    otherwords, when we change the lock range, the newly used
 *    counter always starts at 0.
 *
 * 2) Keeping sync requests streaming between lock_range changes.
 *
 *    A simple algorithm would create a lock range, sync it, move
 *    the lock range, sync the next range, etc.  When trying to
 *    sync an array with other even a minimal amount of normal
 *    I/O active, our sync speed will be limited by the latency
 *    inclurred waiting for these pending I/Os to complete before
 *    we can access the updated range.
 *
 *    To alieviate this delay,  we split our sync range in half.
 *    When the lower range (the one closest to sector 0) completes
 *    its sync operation, the lock_range is moved to encompase the
 *    higher, active sync region, as well as a new regions of sectors
 *    to be synced.  This allows sync operations in one region to
 *    continue/complete while we are waiting for the new lock_range
 *    to take effect on future requests.
 *    
 *    The two ranges are managed using a sync_phase variable held
 *    in the r1buf.
 */
typedef enum {
	MDK_SAFEMODE_OFF,
	MDK_SAFEMODE_ACTIVE,
	MDK_SAFEMODE_IDLE_WAIT,
	MDK_SAFEMODE_EXIT
} mdk_safemode_state_t;

struct mdk_sync_support_s {
	mdk_workitem_support_t	work_state;
	spinlock_t		sync_lock;
	int			phase;
	atomic_t		reads_pending;
	u_int			writes_pending[2];
	u_int			sync_pending[2];
	md_sector_range_t	lock_range;
	md_sector_range_t	sync_range[2];
	wait_queue_head_t	io_wait;
	wait_queue_head_t	sync_wait;
	mdk_workitem_t		safemode_workitem;
	mdk_safemode_state_t	safemode;
	struct timer_list	safemode_timer;
	u_int			safemode_delay;
	u_int			window_size;
	u_int			io_limit_min;
	u_int			io_limit_max;
	/*
	 * Resync statistics.  We average our
	 * speed over a sliding ~16 second window
	 * updated every 1/2 second.  The values
	 * below give the number of sectors processed
	 * over the given stat duration.
	 */
	u_long			stat_sectors;
	u_long			stat_time;
};

typedef struct mdk_sync_support_s mdk_sync_support_t;

struct mddev_s
{
	mdk_member_t			member;

	dev_t				unit;
	int				preferred_unit;
	int				sb_dirty;
	int				ro;

	struct gendisk			*gendisk;

	mdk_member_t			*dedicated_spares;

	/* Superblock information */
	int				major_version,
					minor_version,
					patch_version;
	int				persistent;
	int				chunk_size;
	time_t				ctime, utime;
	int				level, layout;
	int				raid_disks;	/* Number of active
							 * disks in the array
							 */
	int				max_disks;	/* Number of total disks
							 * in the array.
							 */

	/* XXX Is this field useful?  It seems to shadow tha size of the smallest member. */
	sector_t			size; /* used size of component devices */
	__u64				events;

	char				uuid[16];

	struct mdk_thread_s		*thread;	/* management thread */
	struct mdk_thread_s		*sync_thread;	/* doing resync or reconstruct */

	/* recovery/resync flags 
	 * NEEDED:   we might need to start a resync/recover
	 * RUNNING:  a thread is running, or about to be started
	 * SYNC:     actually doing a resync, not a recovery
	 * ERR:      and IO error was detected - abort the resync/recovery
	 * INTR:     someone requested a (clean) early abort.
	 * DONE:     thread is done and is waiting to be reaped
	 */
#define	MD_RECOVERY_RUNNING	0
#define	MD_RECOVERY_SYNC	1
#define	MD_RECOVERY_ERR		2
#define	MD_RECOVERY_INTR	3
#define	MD_RECOVERY_DONE	4
#define	MD_RECOVERY_NEEDED	5
	unsigned long			recovery;

	//int				in_sync;	/* know to not need resync */
	struct semaphore		reconfig_sem;

	int				changed;	/* true if we might need to reread partition info */
	int				degraded;	/* whether md should consider
							 * adding a spare
							 */

	atomic_t			recovery_active; /* blocks scheduled, but not written */
	wait_queue_head_t		recovery_wait;
	sector_t			recovery_cp;
	unsigned int			safemode;	/* if set, update "clean" superblock
							 * when no writes pending.
							 */ 
	unsigned int			safemode_delay;
	struct timer_list		safemode_timer;
	atomic_t			writes_pending; 
	request_queue_t			*queue;	/* for plugging ... */
};

enum mdk_raid_op_e
{
	MDK_RAID_OP_MSTATE_CHANGED,
	MDK_RAID_OP_SPARE_ADDED,
	MDK_RAID_OP_REQUEST_VERIFY,
	MDK_RAID_OP_REQUEST_RESYNC,
	MDK_RAID_OP_PAUSE_MEMBER,
	MDK_RAID_OP_REMOVE_MEMBER,
	MDK_RAID_OP_ADD_MEMBER,
	MDK_RAID_OP_INIT,
	MDK_RAID_OP_INIT_MEMBER,
	MDK_RAID_OP_SYNC_REQUEST,
	MDK_RAID_OP_SYNC_STATS,
	MDK_RAID_OP_ENTER_SAFEMODE,
	MDK_RAID_OP_EXIT_SAFEMODE,
	MDK_RAID_OP_RUN,
	MDK_RAID_OP_RUNNABLE,
	MDK_RAID_OP_STOP
};
typedef enum mdk_raid_op_e mdk_raid_op_t;

struct mdk_syncstatsop_arg_s {
	u_long	 stat_sectors;
	u_long	 stat_time;
	u_int	 syncop;
};
typedef struct mdk_syncstatsop_arg_s mdk_syncstatsop_arg_t;

struct mdk_personality_s
{
	char *name;
	struct module *owner;
	int (*make_request)(request_queue_t *q, struct bio *bio);
	int (*run)(mddev_t *mddev);
	int (*stop)(mddev_t *mddev);
	/* error_handler must set ->faulty and clear ->in_sync
	 * if appropriate, and should abort recovery if needed 
	 */
	int (*sync_request)(mddev_t *mddev, sector_t sector_nr, int go_faster);
	mdk_member_callback_t *raid_op;

	/* XXX These will go away */
	void (*error_handler)(mddev_t *mddev, mdk_rdev_t *rdev);
	int (*hot_add_disk) (mddev_t *mddev, mdk_rdev_t *rdev);
	int (*hot_remove_disk) (mddev_t *mddev, int number);
	int (*spare_active) (mddev_t *mddev);
};

enum mdk_dev_op_e
{
	MDK_DEVOP_INITIALIZE,
	MDK_DEVOP_WIPE_METADATA,
	MDK_DEVOP_CLAIM,
	MDK_DEVOP_ATTACH,
	MDK_DEVOP_GET_FREESIZE
};
typedef enum mdk_dev_op_e mdk_dev_op_t;

enum mdk_member_op_e
{
	MDK_MEMBER_OP_SET_STATE,
	MDK_MEMBER_OP_SET_ARRAY_ACTION,
	MDK_MEMBER_OP_SET_USER_ACTION,
	MDK_MEMBER_OP_GET_ARRAY_ACTION,
	MDK_MEMBER_OP_GET_USER_ACTION,
	MDK_MEMBER_OP_ALLOC_SPARE,
	MDK_MEMBER_OP_ENTER_SAFEMODE,
	MDK_MEMBER_OP_EXIT_SAFEMODE,
	MDK_MEMBER_OP_GET_CHECKPOINT,
	MDK_MEMBER_OP_SET_CHECKPOINT,
	MDK_MEMBER_OP_ALLOC_ARRAY_INFO
	/* XXX Create Spares? */
	/* XXX Array creation? */
};
typedef enum mdk_member_op_e mdk_member_op_t;

struct mdk_set_state_arg_s {
	mdk_mstate_t	oldstate;
	mdk_mstate_t	newstate;
};
typedef struct mdk_set_state_arg_s mdk_set_state_arg_t;

struct mdk_checkpoint_arg_s {
	u_int		checkpoint_type;
	sector_t	checkpoint;
};
typedef struct mdk_checkpoint_arg_s mdk_checkpoint_arg_t;

struct mdk_metapersonality_s
{
	char *name;
	struct module *owner;
	struct list_head links;
	int (*dev_op)(dev_t dev, mdk_dev_op_t, void *arg);
	mdk_member_callback_t *member_op;
	atomic_t	refcount;
};

enum mdk_action_e
{
	MDK_ACTION_NONE,
	MDK_ACTION_ZERO,
	MDK_ACTION_RESYNC,
	MDK_ACTION_VERIFY,
	MDK_ACTION_VERIFY_FIX
};
typedef enum mdk_action_e mdk_action_t;

static inline char * mdname (mdk_member_t *member)
{
	return ((member->name[0] == 0) ? "mdX" : member->name);
}

extern mdk_rdev_t * find_rdev_nr(mddev_t *mddev, int nr);

/*
 * iterates through some rdev ringlist. It's safe to remove the
 * current 'rdev'. Dont touch 'tmp' though.
 */
#define ITERATE_RDEV_GENERIC(head,rdev,tmp)				\
									\
	for ((tmp) = (head).next;					\
		(rdev) = (list_entry((tmp), mdk_rdev_t,			\
				     member.peer_links)),		\
			(tmp) = (tmp)->next, (tmp)->prev != &(head)	\
		; )
/*
 * iterates through the 'same array disks' ringlist
 */
#define ITERATE_RDEV(mddev,rdev,tmp)					\
	ITERATE_RDEV_GENERIC((mddev)->member.child_set,rdev,tmp)

/*
 * Iterates through 'pending RAID disks'
 */
#define ITERATE_RDEV_PENDING(rdev,tmp)					\
	ITERATE_RDEV_GENERIC(pending_raid_disks,rdev,tmp)

/*
 * Handle superblock details.
 * We want to be able to handle multiple superblock formats
 * so we have a common interface to them all, and an array of
 * different handlers.
 * We rely on user-space to write the initial superblock, and support
 * reading and updating of superblocks.
 * Interface methods are:
 *   int load_super(mdk_rdev_t *dev, mdk_rdev_t *refdev, int minor_version)
 *      loads and validates a superblock on dev.
 *      if refdev != NULL, compare superblocks on both devices
 *    Return:
 *      0 - dev has a superblock that is compatible with refdev
 *      1 - dev has a superblock that is compatible and newer than refdev
 *          so dev should be used as the refdev in future
 *     -EINVAL superblock incompatible or invalid
 *     -othererror e.g. -EIO
 *
 *   int validate_super(mddev_t *mddev, mdk_rdev_t *dev)
 *      Verify that dev is acceptable into mddev.
 *       The first time, mddev->raid_disks will be 0, and data from
 *       dev should be merged in.  Subsequent calls check that dev
 *       is new enough.  Return 0 or -EINVAL
 *
 *   void sync_super(mddev_t *mddev, mdk_rdev_t *dev)
 *     Update the superblock for rdev with data in mddev
 *     This does not write to disc.
 *
 */

/* XXX XXX XXX This is all temporary while the metadata abstraction work is
 * in progress.
 */
struct super_type  {
	char 		*name;
	struct module	*owner;
	int		(*load_super)(mdk_rdev_t *rdev, mdk_rdev_t *refdev, int minor_version);
	int		(*validate_super)(mddev_t *mddev, mdk_rdev_t *rdev);
	void		(*sync_super)(mddev_t *mddev, mdk_rdev_t *rdev);
};

/*
 * Iterates through all the children of a member.  We take
 * care to grab the child set list lock whenever navigating
 * the list, and to always hold a refcount when unlocked.
 * Any code which breaks out of this loop will own a reference
 * to the object pointed to by "member" that must evenutally be
 * released.
 */
#define ITERATE_CHILDREN(parent, child, tmp, flags)			    \
									    \
for (({ spin_lock_irqsave(&(parent)->member_lock, flags);		    \
	tmp = (parent)->child_set.next;					    \
	child = NULL;});						    \
     ({ if (tmp != &(parent)->child_set)				    \
		member_get(list_entry(tmp, mdk_member_t, peer_links));	    \
	spin_unlock_irqrestore(&(parent)->member_lock, flags);	    	    \
	if (child) member_put(child);					    \
	child = list_entry(tmp, mdk_member_t, peer_links);		    \
	tmp != &(parent)->child_set;});					    \
     ({ spin_lock_irqsave(&(parent)->member_lock, flags);		    \
	tmp = tmp->next;})						    \
	)

/*
 * Same as above, but assumes the member_lock is already held
 * on the parent.  No reference counts are taken on the children.
 */
#define ITERATE_CHILDREN_LOCKED(parent, child, tmp)			    \
for (({tmp = (parent)->child_set.next, child = NULL;});			    \
     ({ child = list_entry(tmp, mdk_member_t, peer_links);		    \
	tmp != &(parent)->child_set;});					    \
     ({tmp = tmp->next;}))

int read_disk_sb(mdk_rdev_t *rdev);
sector_t calc_dev_size(mdk_rdev_t *rdev, unsigned chunk_size);
int super_90_load(mdk_rdev_t *rdev, mdk_rdev_t *refdev, int minor_version);

#endif
